//camera.cpp

#include "main.h"

extern FRAME frame;
extern LOGGER err_log;

#define pi 3.141592654
double rad_deg = pi/180;

void angle2xyz(XYZ *v,float theta, float phi);

CAMERA::CAMERA()
{
//	at.x = 3.1f;
//	at.y = 2.1f;
//	at.z = 1.5f;
	at.x = 1;
	at.y = 1;
	at.z = 1;

	up.x = 0;
	up.y = 0;
	up.z = 1;

	//Starting angles
	theta=deg90/2 + deg90*2;
	phi=-deg90/4 + .0001f;

//	theta = deg90;
//	phi = deg90/16 - deg90;
//	phi = 0;

	float_distance = 35;

	//Starting FOV
	fov_init = 90;
	fov = fov_init;

	//Clipping plane distance
	view_distance = 20;

	//Speed of camera when moving freely
	freemove_speed = 0.01f;

	type = CT_FREE;

	constrain_rotate = true;
	drag_mod_x = 1;
	drag_mod_y = 1;
	flip_drag = false;

	drag_speed = .001f;

	orth_x = .4f;
	orth_y = .3f;

	zoom_height = .5;
	zoom_speed = 1.08f;

	ortho_type = OT_NULL;

	update_camera();
}

void CAMERA::update_camera()
{

//	if (float_distance>FLOAT_DISTANCE_MAX)
//		float_distance=FLOAT_DISTANCE_MAX;
//	if (float_distance<FLOAT_DISTANCE_MIN)
//		float_distance=FLOAT_DISTANCE_MIN;

//	if (phi>PHI_MAX)
//		phi=PHI_MAX;

//	XYZ v;

	if (type==CT_FLOAT)
	{
//		angle2xyz(&at,theta,phi);

//		at.x = player.pos->x - at.x*float_distance;
//		at.y = player.pos->y - at.y*float_distance;
//		at.z = player.pos->z - at.z*float_distance;

//		look.x=player.pos->x;
//		look.y=player.pos->y;
//		look.z=player.pos->z;

		look.x = 0;
		look.y = 0;
		look.z = 0;
	}
	else if (type==CT_FREE)
	{
		angle2xyz(&look,theta,phi);

		look.x=at.x+look.x;
		look.y=at.y+look.y;
		look.z=at.z+look.z;
	}
}

void CAMERA::fix_angle()
{
	while (theta>deg360)
		theta -= deg360;
	while (theta<0)
		theta += deg360;
}

void angle2xyz(XYZ *v,float theta, float phi)
{
	while (theta>deg360)
		theta -= deg360;
	while (theta<0)
		theta += deg360;

	v->x = (float)cos(theta)*(float)cos(phi);
	v->y = (float)sin(theta)*(float)cos(phi);
	v->z = (float)sin(phi);
}

void CAMERA::set_XY()
{
	at.x = 0;	at.y = 0;	at.z = 1;
	look.x = 0;	look.y = 0;	look.z = 0;
	up.x = 0;	up.y = 1;	up.z = 0;
	theta = 0;
	phi = -deg90;
	constrain_rotate = false;
	projection = PROJECTION_ORTHO;
	drag_mod_x = 1;
	drag_mod_y = -1;
	flip_drag = true;
	ortho_type = OT_XY;
}
void CAMERA::set_YZ()
{
	at.x = 1;	at.y = 0;	at.z = 0;
	look.x = 0;	look.y = 0;	look.z = 0;
	up.x = 0;	up.y = 0;	up.z = 1;
	theta = deg180;
	phi = 0;
	constrain_rotate = false;
	projection = PROJECTION_ORTHO;
	drag_mod_x = -1;
	drag_mod_y = -1;
	ortho_type = OT_YZ;
}
void CAMERA::set_XZ()
{
	at.x = 0;	at.y = -1;	at.z = 0;
	look.x = 0;	look.y = 0;	look.z = 0;
	up.x = 0;	up.y = 0;	up.z = 1;
	theta = deg90;
	phi = 0;
	constrain_rotate = false;
	projection = PROJECTION_ORTHO;
	drag_mod_x = -1;
	drag_mod_y = -1;
	ortho_type = OT_XZ;
}

void CAMERA::set_matrices_ortho()
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-orth_x*zoom_height,orth_x*zoom_height, -orth_y*zoom_height,orth_y*zoom_height, 0.001f, 2);
//	glOrtho(0,frame.display.screen_x,0,frame.display.screen_y,0.001f,view_distance);
//	gluPerspective(fov,4.0f/3.0f,0.001f,view_distance);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
//	gluLookAt(1,0,0,0,0,0,0,0,1);
	gluLookAt(at.x,at.y,at.z,look.x,look.y,look.z,up.x,up.y,up.z);
}

void CAMERA::set_matrices()
{set_matrices_perspective();}
void CAMERA::set_matrices_perspective()
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(fov,4.0f/3.0f,0.001f,view_distance);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gluLookAt(at.x,at.y,at.z,look.x,look.y,look.z,0,0,1);
}


void CAMERA::set_overlay_matrices()
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0,frame.display.screen_x,0,frame.display.screen_y,-2000,2000);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gluLookAt(0,0,500,0,0,0,0,1,0); 
}

void CAMERA::mouse_move(POINT *cursor_pos, bool *click)
{
	float move_scale = 1;

	if (frame.config.get_bool(VAR_FAST_MOVE))
		move_scale = 50;

	float move_mod_x = frame.control->mouse_sensitivity_x*freemove_speed*move_scale;
	float move_mod_y = frame.control->mouse_sensitivity_y*freemove_speed*move_scale;

	float drag_x, drag_y;

	if (flip_drag)
	{
		drag_x = drag_speed * cursor_pos->y * drag_mod_y;
		drag_y = drag_speed * cursor_pos->x * drag_mod_x;
	}
	else
	{
		drag_x = drag_speed * cursor_pos->x * drag_mod_x;
		drag_y = drag_speed * cursor_pos->y * drag_mod_y;
	}

	drag_x *= zoom_height;
	drag_y *= zoom_height;

	if (projection==PROJECTION_ORTHO)
	{
		//Move along (viewer) XY plane
		at.x+=( (at.x+(float)cos(theta)*(float)cos((phi+rad90))) -at.x)*(drag_y);
		at.y+=( (at.y+(float)sin(theta)*(float)cos((phi+rad90))) -at.y)*(drag_y);
		at.z+=( (at.z+(float)sin((phi+rad90))) -at.z)*(drag_y);

		at.x+=((at.x+(float)cos((theta+rad90)))-at.x)*(drag_x);
		at.y+=((at.y+(float)sin((theta+rad90)))-at.y)*(drag_x);
	}
	else
	{

		if (click[0] && !click[2])
		{
			//Look around
			theta+=(cursor_pos->x)*frame.control->mouse_sensitivity_x;
			phi+=(cursor_pos->y)*frame.control->mouse_sensitivity_y;
			if (constrain_rotate)
			{
				if (phi<=-rad89) phi=-rad89;
				if (phi>=rad89) phi=rad89;
			}
		}
		else if (click[0] && click[2])
		{
			//Move along (viewer) XZ plane
			at.x+=(look.x-at.x)*(cursor_pos->y)*move_mod_y;
			at.y+=(look.y-at.y)*(cursor_pos->y)*move_mod_y;
			at.z+=(look.z-at.z)*(cursor_pos->y)*move_mod_y;

			at.x+=((at.x+(float)cos((theta+rad90)))-at.x)*(cursor_pos->x)*move_mod_x;
			at.y+=((at.y+(float)sin((theta+rad90)))-at.y)*(cursor_pos->x)*move_mod_x;
		}
		else if (!click[0] && click[2])
		{
			//Move along (viewer) XY plane
			at.x+=( (at.x+(float)cos(theta)*(float)cos((phi+rad90))) -at.x)*(cursor_pos->y)*move_mod_y;
			at.y+=( (at.y+(float)sin(theta)*(float)cos((phi+rad90))) -at.y)*(cursor_pos->y)*move_mod_y;
			at.z+=( (at.z+(float)sin((phi+rad90))) -at.z)*(cursor_pos->y)*move_mod_y;

			at.x+=((at.x+(float)cos((theta+rad90)))-at.x)*(cursor_pos->x)*move_mod_x;
			at.y+=((at.y+(float)sin((theta+rad90)))-at.y)*(cursor_pos->x)*move_mod_x;
		}
	}

//	err_log << "CAMERA:   " << at.x << ", " << at.y << ", " << at.z << "   :   " << look.x << ", " << look.y << ", " << look.z << endl;
	update_camera();
//	err_log << "CAMERA`:  " << at.x << ", " << at.y << ", " << at.z << "   :   " << look.x << ", " << look.y << ", " << look.z << endl;
}

void CAMERA::zoom_out()
{
	if (projection==PROJECTION_ORTHO)
	{
		zoom_height *= zoom_speed;
//		orth_x *= zoom_speed;
//		orth_y *= zoom_speed;
	}
}

void CAMERA::zoom_in()
{
	if (projection==PROJECTION_ORTHO)
	{
		zoom_height /= zoom_speed;
//		orth_x /= zoom_speed;
//		orth_y /= zoom_speed;
//		if (orth_x <= 0) orth_x = zoom_speed;
//		if (orth_y <= 0) orth_y = zoom_speed;
	}
}


void CAMERA::draw_grid()
{
	if (!src_viewport->get_grid())
		return;

	double x,y,w,h;
	w = (orth_x * zoom_height);
	h = (orth_y * zoom_height);

	switch (ortho_type) {
	case OT_XY:	x = at.x - w;	y = at.y - h; break;
	case OT_YZ:	x = at.y - w;	y = at.z - h; break;
	case OT_XZ:	x = at.x - w;	y = at.z - h; break;
	default:	return;
	}

	w *= 2;
	h *= 2;

	grid.draw(x,y,w,h);
}

void CAMERA::grid_up()
{
	double w = 2 *(orth_x * zoom_height);
	grid.resize(1, w);
}
void CAMERA::grid_down()
{
	double w = 2 *(orth_x * zoom_height);
	grid.resize(-1, w);
}





