//config.cpp

#include "main.h"


extern FRAME frame;
extern LOGGER err_log;
extern int compare(char *a,char *b);

const char *boolstr[2] = {"OFF","ON"};

CONFIG::CONFIG()
{
	initialized = false;
	sprintf(filename, "config.dat");

	always_run = true;
	slowdown = false;

	var_list[0]			= 0;
	var_bool_list[0]	= 0;

	rebuild_var_str = true;
	var_str = NULL;

	sprintf(data_dir,"data");

//	sprintf(data_dir,"d:/~Documents/~Projects/Bumfight/data");
//	sprintf(data_dir,"C:/~~Documents/~Projects/Bumfight/data");
}
void CONFIG::destroy()
{
	if (var_str!=NULL)
	{
		for (int i=0; i<num_cols; i++)
		{
			free(var_str[i]);
		}
		free(var_str);
	}
}
VARIABLE::VARIABLE()
{
	active = false;
	type = VT_GENERIC;
	b_value = true;
	value = 0;
	force_new_list = false;
}

void CONFIG::init()
{
/*
But Doug, what does all this crap mean?

You can define important stuff about your config vars here.  You can give each a string name,
which is then used by the console.  Type "set <string> <value>", where <string> is some variable's
string name (such as "debug_step"), and the value can be whatever new value you want it to be.
You can also use functions like .set() or .toggle() to do various things, and their results should
be seemingly obvious.  Variable types of bool combined with another type (like VT_BOOL_FLOAT) work
slightly differently - they have a given value and a boolean value, and if the boolean is true, 
they return their given value.  If the boolean is false, they return zero.  This is useful for
being able to toggle something off without ruining its actual value (like, say, you defined a gravity
factor variable).

Boolean variables do not reference their float component, so it can remain zero.

init_var() notes:

	redraw:			Whenever this variable is changed, the display lists need to be recreated
	rangemin/max:	for _RANGE types, this defines their range

*/

/*
		var name					var_type		redraw		var string					initial float
													 rangemin								 initial bool
													  rangemax
*/
init_var(VAR_DRAW_FPS,				VT_BOOL,		false,0,0,	"d_fps",					0,false);
init_var(VAR_DEBUG_STEP,			VT_BOOL,		false,0,0,	"debug_step",				0,false);
init_var(VAR_FAST_MOVE,				VT_BOOL,		false,0,0,	"fast_move",				0,true);
init_var(VAR_FORCE_NEW_LIST,		VT_BOOL,		false,0,0,	"force_new_list",			0,false);
init_var(VAR_DRAW_AXES,				VT_BOOL,		false,0,0,	"draw_axes",				0,true);
init_var(VAR_DRAW_EXTENTS_MODE,		VT_BOOL_RANGE,	false,0,2,	"draw_extents_mode",		0,false);
init_var(VAR_DRAW_MESSAGES,			VT_BOOL,		false,0,0,	"draw_messages",			0,true);
init_var(VAR_DRAW_VECTORS_MODE,		VT_BOOL_RANGE,	false,0,4,	"draw_vectors_mode",		0,false);
init_var(VAR_DRAW_VECTORS_MODE2,	VT_BOOL_RANGE,	false,0,4,	"draw_vectors_mode2",		2,true);
init_var(VAR_DBINS,					VT_BOOL,		false,0,0,	"dbins",					0,false);
init_var(VAR_DTRAILS,				VT_BOOL,		false,0,0,	"dtrails",					0,false);
init_var(VAR_FREEZE_PARTICLES,		VT_BOOL,		false,0,0,	"freeze_particles",			0,false);
init_var(VAR_RUN_SIMULATION,		VT_BOOL,		false,0,0,	"run_simulation",			0,true);
init_var(VAR_DEBUG_LOG,				VT_BOOL,		false,0,0,	"debug_log",				0,true);
init_var(VAR_SINGLE_VIEW,			VT_BOOL,		false,0,0,	"single_view",				0,true);
init_var(VAR_CAMERA_LOOK,			VT_BOOL,		false,0,0,	"camera_look",				0,true);
init_var(VAR_SHOW_KEYS,				VT_BOOL,		false,0,0,	"show_keys",				0,false);
init_var(VAR_AUTO_REFRESH,			VT_BOOL,		false,0,0,	"auto_refresh",				0,false);
init_var(VAR_IN_SIMULATION,			VT_BOOL,		false,0,0,	"in_simulation",			0,false);
init_var(VAR_PHOTON_ELASTICITY,		VT_FLOAT,		false,0,0,	"photon_elasticity",		.2f,false);	//.2f
init_var(VAR_PHOTON_GRAVITY,		VT_FLOAT,		false,0,0,	"photon_gravity",			1,false);
init_var(VAR_PHOTON_VF_BIAS,		VT_FLOAT,		false,0,0,	"photon_vf_bias",			5.01f,true);
init_var(VAR_PHOTON_CONSERVE,		VT_BOOL,		false,0,0,	"photon_conserve",			0,true);
init_var(VAR_FIRE_STRAIGHT,			VT_BOOL,		false,0,0,	"photon_fire_straight",		0,true);
init_var(VAR_SIM_X0,				VT_FLOAT,		true,0,0,	"sim_x0",					-.05f,true);
init_var(VAR_SIM_X1,				VT_FLOAT,		true,0,0,	"sim_x1",					.05f,true);
init_var(VAR_SIM_Y0,				VT_FLOAT,		true,0,0,	"sim_y0",					-.07f,true);
init_var(VAR_SIM_Y1,				VT_FLOAT,		true,0,0,	"sim_y1",					.09f,true);
init_var(VAR_SIM_Z0,				VT_FLOAT,		true,0,0,	"sim_z0",					.1f,true);
init_var(VAR_SIM_Z1,				VT_FLOAT,		true,0,0,	"sim_z1",					.1f,true);
init_var(VAR_SIM_WAVES,				VT_FLOAT,		false,0,0,	"sim_waves",				1,false);
init_var(VAR_APPLY_ALL,				VT_BOOL,		false,0,0,	"apply_all",				0,false);
init_var(VAR_SHOW_VARS,				VT_BOOL,		false,0,0,	"show_vars",				0,false);
init_var(VAR_SPREAD_TYPE,			VT_FLOAT,		false,0,0,	"spread_type",				ST_JITTER,false);
init_var(VAR_INCREMENTAL_SUBDIV,	VT_BOOL,		false,0,0,	"incremental_subdiv",		0,true);
init_var(VAR_GIANT_BIN,				VT_BOOL,		false,0,0,	"giant_bin",				0,false);
init_var(VAR_SUBDIV_TARGET,			VT_FLOAT,		false,0,0,	"subdiv_target",			.0012,false);	//.0012: good for horse; .004 good for box or face
init_var(VAR_SUBDIV_ONLY,			VT_BOOL,		false,0,0,	"subdiv_only",				0,false);
init_var(VAR_NORMALIZE_IBFV,		VT_BOOL,		false,0,0,	"normalize_ibfv",			0,true);

initialized = true;
}

void CONFIG::init_var(int v, int new_type, bool force_new_list, int range_min, int range_max, char *str, float new_value, bool new_b_value)
{
	sprintf(var[v].name,"%s",str);
	var[v].value_default = new_value;
	var[v].value = new_value;
	var[v].active = true;
	var[v].type = new_type;
	var[v].b_value = new_b_value;
	var[v].force_new_list = force_new_list;
	var[v].range_min = range_min;
	var[v].range_max = range_max;

	append_list(var_list,var[v].name);
	if (var[v].type==VT_BOOL || var[v].type==VT_BOOL_FLOAT)
		append_list(var_bool_list,var[v].name);
}

//Save and load do nothing.
// Yeah, this is a holdover from previous uses of these files
int save()
{
	err_log << "J00r g4m3 r s4v3d!!1" << endl;
	return 1;
}
int load()
{
	return 1;
}


//If redraw is required, it'll set the var that'll require a redraw
void CONFIG::check_force_new_list(int i)
{
	if (var[i].force_new_list)
		var[VAR_FORCE_NEW_LIST].b_value = true;
}

//For ranged variables, this increments the range by d and loops if necessary
int CONFIG::cycle(char *str, int d)
{
	int i = get_var(str);
	if (i < 0) return -1;
	return cycle(i,d);
}
int CONFIG::cycle(int i, int d)
{
	rebuild_var_str = true;
	var[i].value += d;
	if (var[i].value > var[i].range_max) var[i].value = var[i].range_min;
	if (var[i].value < var[i].range_min) var[i].value = var[i].range_max;
	var[i].b_value = (var[i].value != 0);

	check_force_new_list(i);
	return i;
}

//For bools, the boolean component is toggled
int CONFIG::toggle(char *str)
{
	int i = get_var(str);
	if (i < 0) return -1;
	return toggle(i);
}
int CONFIG::toggle(int i)
{
	rebuild_var_str = true;
	if (var[i].b_value)	var[i].b_value = false;
	else				var[i].b_value = true;
	check_force_new_list(i);

	if (i!=VAR_DRAW_MESSAGES)
		frame.messages->add("%s toggled %s",var[i].name,boolstr[(int)var[i].b_value]);
	return i;
}

//Explicitly set a variable's values
int CONFIG::set(char *str, float new_value)
{
	int i = get_var(str);
	if (i < 0) return -1;
	return set(i,new_value);
}
int CONFIG::set(int i, float new_value)
{
	rebuild_var_str = true;
	var[i].value = new_value;
	var[i].b_value = (new_value!=0);

	if (var[i].type==VT_FLOAT)
		frame.messages->add("Variable %s set to %f",var[i].name,get_float(i));
	if (var[i].type==VT_BOOL)
		frame.messages->add("Variable %s set to %i",var[i].name,get_bool(i));
	if (var[i].type==VT_BOOL_FLOAT)
		frame.messages->add("Variable %s set to %i",var[i].name,get_boolint(i));
	if (var[i].type==VT_BOOL_RANGE)
		frame.messages->add("Variable %s set to %i",var[i].name,get_boolint(i));


	check_force_new_list(i);
	return i;
}

//Reset a variable to whatever defaults were made with init_var()
int CONFIG::reset(char *str)
{
	int i = get_var(str);
	if (i < 0) return -1;
	return reset(i);
}
int CONFIG::reset(int i)
{
	rebuild_var_str = true;
	var[i].value = var[i].value_default;
	check_force_new_list(i);
	return i;
}

//Given a variable's stringname, find the variable it represents
int CONFIG::get_var(char *str)
{
	for (int i=0; i<MAX_VARIABLES; i++)
	{
		if (compare(var[i].name,str)==MATCH_TOTAL)
			return i;
	}
	return -1;
}


//WHOAH WATCH OUT --
// this pretty much invalidates the security here.  But it is
// needed for some things for performance!
VARIABLE *CONFIG::steal_var(int i)
{
	//Drop a warning, just so I'm aware of it if things go nuts
	// BUT OF COURSE warning outputs to err_log!  And err_log will crash if it's used BEFORE THIS FUNCTION RETURNS.  Oh the humanity
//	frame.warning("VARIABLE[%i] is being STOLEN, just so you know.",i);

	return &var[i];
}


//Get the expressed value of the variable, based on variable type
bool CONFIG::get_bool(int i)	{ return var[i].b_value; }
float CONFIG::get_float(int i)		{ return var[i].value; }
float CONFIG::get_boolint(int i)
{
	if (!var[i].b_value) return 0;
	return var[i].value;
}


//Get a list of variables (used by console)
char *CONFIG::get_var_list()		{ return var_list; }
char *CONFIG::get_var_bool_list()	{ return var_bool_list; }

void CONFIG::append_list(char *dst, char *src)
{
	if (dst[0]==0)	sprintf(dst,"%s",src);
	else			sprintf(dst,"%s\n%s",dst,src);
}


void CONFIG::show_vars()
{
	num_rows = 40;
	num_cols = ceil((float)MAX_VARIABLES / (float)num_rows);
	int i,j,J,n;
	char tmp[128];

	if (rebuild_var_str)
	{
		rebuild_var_str = false;

		//free mem first bastard
		if (var_str!=NULL)
		{
			for (i=0; i<num_cols; i++)
			{
				free(var_str[i]);
			}
			free(var_str);
		}

		//Allocate for each column
		var_str = (char**)calloc(num_cols,sizeof(char*));
		
		for (i=0; i<num_cols; i++)
		{
			if (i==0)	n = 18;
			else		n = 1;
			for (j=0; j<num_rows; j++)
			{
				J = j + i*num_rows;
				if (J>=MAX_VARIABLES) break;

				if (var[J].type==VT_FLOAT)	sprintf(tmp,"%f",get_float(J));
				if (var[J].type==VT_BOOL)		sprintf(tmp,"%i",get_bool(J));
				if (var[J].type==VT_BOOL_FLOAT)	sprintf(tmp,"%f",get_boolint(J));
				if (var[J].type==VT_BOOL_RANGE)	sprintf(tmp,"%i",(int)get_float(J));

				n += (int)strlen(var[J].name);
				n += (int)strlen(tmp);
				n += 3;
			}
			n++;

			var_str[i] = (char*)calloc(n,sizeof(char));
			if (i==0)	sprintf(var_str[i],"*** Variables ***\n");
			else		sprintf(var_str[i],"\n");
			for (j=0; j<num_rows; j++)
			{
				J = j + i*num_rows;
				if (J>=MAX_VARIABLES) break;

				if (var[J].type==VT_FLOAT)	sprintf(tmp,"%f",get_float(J));
				if (var[J].type==VT_BOOL)		sprintf(tmp,"%i",get_bool(J));
				if (var[J].type==VT_BOOL_FLOAT)	sprintf(tmp,"%f",get_boolint(J));
				if (var[J].type==VT_BOOL_RANGE)	sprintf(tmp,"%i",(int)get_float(J));

				sprintf(var_str[i],"%s%s: %s\n",var_str[i],var[J].name,tmp);
			}
		}
	}


	for (i=0; i<num_cols; i++)
	{
		frame.display.fonts->draw(FF_TIMES_SMALL, var_str[i],200+i*200,frame.display.screen_y-15);
	}


}


//Get a list of files with the given extension, found in the data directory
char *CONFIG::get_data_file_list(char *extension)
{
//	char list[1024];
	char *list;

	//Aww screwit -- this shouldn't be called much anyway, so
	// it shouldn't matter too terribly that it isn't cleaned up
	list = (char*)calloc(1024,sizeof(char));

	list[0] = 0;
	sprintf(list,"%s",read_dir("data",extension));

	err_log << "FILE LIST FOR [" << extension << "]: {" << list << "}" << endl;

	return list;
}
