//config.h

#ifndef _CONFIG_
#define _CONFIG_

#include "main.h"


#define MAX_VARIABLE_NAME	64

/*
A neat way for me to manage some global settings

How to create a new variable:

1. Define a variable integer (such as "#define VAR_DNORMAL 4") with a unique number
2. Update MAX_VARIABLES to reflect the number of vars defined
3. Add an init_var() statement for the new variable (see comment in config.cpp, init())

These variables can be set/toggled/reset/etc, and their current
values can be checked.  Some of their behavior depends on their
variable type.  VT_BOOL, VT_FLOAT, and VT_STRING work
as you would expect.  However:

VT_BOOL_FLOAT		- if the bool part of the variable is true, the variable's value is checked,
					else, the variable is returned as zero.
VT_BOOL_RANGE	- the variable is limited to a set of numbers, and can be cycled.  If the bool
					component is set to false, the value of the variable will return as zero


*/

#define VAR_DRAW_FPS			0
#define VAR_DEBUG_STEP			1
#define VAR_FAST_MOVE			2
#define VAR_FORCE_NEW_LIST		3
#define VAR_DRAW_AXES			4
#define VAR_DRAW_EXTENTS_MODE	5
#define VAR_DRAW_MESSAGES		6
#define VAR_DRAW_VECTORS_MODE	7
#define VAR_DBINS				8
#define VAR_DTRAILS				9
#define VAR_FREEZE_PARTICLES	10
#define VAR_RUN_SIMULATION		11
#define VAR_DEBUG_LOG			12
#define VAR_SINGLE_VIEW			13
#define VAR_CAMERA_LOOK			14
#define VAR_SHOW_KEYS			15
#define VAR_AUTO_REFRESH		16
#define VAR_IN_SIMULATION		17
#define VAR_PHOTON_ELASTICITY	18
#define VAR_PHOTON_GRAVITY		19
#define VAR_PHOTON_VF_BIAS		20
#define VAR_PHOTON_CONSERVE		21
#define VAR_FIRE_STRAIGHT		22
#define VAR_SIM_X0				23
#define VAR_SIM_Y0				24
#define VAR_SIM_Z0				25
#define VAR_SIM_X1				26
#define VAR_SIM_Y1				27
#define VAR_SIM_Z1				28
#define VAR_SIM_WAVES			29
#define VAR_APPLY_ALL			30
#define VAR_SHOW_VARS			31
#define VAR_SPREAD_TYPE			32
#define VAR_INCREMENTAL_SUBDIV	33
#define VAR_GIANT_BIN			34
#define VAR_SUBDIV_TARGET		35
#define VAR_SUBDIV_ONLY			36
#define VAR_DRAW_VECTORS_MODE2	37
#define VAR_NORMALIZE_IBFV		38

#define MAX_VARIABLES			39	//Update this whenever you add/remove vars



#define VT_GENERIC			0
#define VT_BOOL				1
#define VT_FLOAT			2
#define VT_STRING			3
#define VT_BOOL_FLOAT		4
#define VT_BOOL_RANGE		5

struct VARIABLE
{
	VARIABLE();
	char name[MAX_VARIABLE_NAME];
	float value;
	float value_default;
	bool b_value;
	bool active;
	int type;
	bool force_new_list;

	int range_max;
	int range_min;

	//floats?
};

class CONFIG
{
public:
	CONFIG();
	void destroy();

	int save();
	int load();

	void check_force_new_list(int i);
	void init();
	void init_var(int v, int new_type, bool force_new_list, int range_min, int range_max, char *str, float new_value, bool new_b_value);

	int cycle(char *str, int d);
	int cycle(int i, int d);

	int toggle(char *str);
	int toggle(int i);

	int set(char *str, float new_value);
	int set(int i, float new_value);

	int reset(char *str);
	int reset(int i);

	int get_var(char *str);

	float get_float(int i);
	bool get_bool(int i);
	float get_boolint(int i);

	char *get_var_list();
	char *get_var_bool_list();

	void show_vars();

	void append_list(char *dst, char *src);

	char *get_data_file_list(char *extension);

	//Key configuration

	//Movement configuration
	bool always_run;
	bool slowdown;
	bool initialized;

	VARIABLE *steal_var(int i);

protected:

	VARIABLE var[MAX_VARIABLES];
	char filename[32];

	char var_list[512];
	char var_bool_list[512];

	char data_dir[256];

	char **var_str;
	bool rebuild_var_str;

	int num_rows;
	int num_cols;
};









#endif

