//console.cpp

#include "main.h"


extern FRAME frame;
extern LOGGER err_log;

//Initially, commands would require a static prefix.  This has been disabled because
// there really is no need for it here.
char command_prefix[8] = "/";

void to_lower(char *str)
{
	for (int i=0; i<(int)strlen(str); i++)
		if (str[i]>=65 && str[i]<=90)
			str[i] += 32;
}
int compare(char *a,char *b);
int compare(char *a,char *b)
{
	to_lower(a);
	to_lower(b);
	int lengtha = (int)strlen(a);
	int lengthb = (int)strlen(b);
	if (lengtha<lengthb)							return MATCH_NONE;		//B is too long
	for (int i=0; i<lengthb; i++) if (a[i]!=b[i])	return MATCH_NONE;		//Up to lengthb, there's a discrepency
	if (lengtha!=lengthb)							return MATCH_PARTIAL;	//Up to lengthb, it's fine, but a is longer
													return MATCH_TOTAL;		//Up to lengthb, it's fine, and both have same length
}

COMMAND::COMMAND() { num_arg_matches = 0; }
void COMMAND::set_matches(char *str)
{
	int i;
	int m = 64;

	if (num_arg_matches>0)
	{
		dynfree<char>(match_args,num_arg_matches,m);
	}

	int k,j;
	i = 0;
	num_arg_matches = 0;
	while (str[i]!=0)
	{
		if (str[i]=='\n')
			num_arg_matches++;

		if (i>=100000)
		{
			err_log << "Argument match setup has failed horrifically." << endl;
			num_arg_matches = 0;
			return;
		}
		i++;
	}
	num_arg_matches++;

	match_args = dynarray<char>(num_arg_matches,m);

	i = 0;
	k = 0;
	for (j=0; j<TEXT_BOX_MAX_INPUT; j++)
	{
		if (str[j]=='\n' || k>=m)
		{
			match_args[i][k] = 0;
			i++;
			k = 0;


			continue;
		}


		match_args[i][k] = str[j];
		if (str[j]==0) break;
		k++;
	}

}

CONSOLE::CONSOLE()
{
	command_list[COM_NULL].set(" ");
	command_list[COM_TOGGLE].set("toggle");
	command_list[COM_SET].set("set");
	command_list[COM_RESET].set("reset");
	command_list[COM_QUIT].set("quit");
	command_list[COM_IMPORT_MODEL].set("import_model");
	command_list[COM_IMPORT_FIELD].set("import_field");

	active = false;
	num_matches = 0;
	opening = false;
	closing = false;
	drop_speed = DROP_SPEED_DEFAULT;

	match_index = -1;
	history_index = -1;
}

void CONSOLE::init()
{
	text_box.set_console();
	text_box.set_width(frame.display.screen_x-2);
	text_box.set_graphic(frame.display.textures->gui[2],2);
	text_box.set_font(FF_DOOM_SMALL_WHITE);
	text_box.set_type(TBT_CONSOLE);
	text_box.set_rows(1);
	input_height = text_box.get_pixel_height();
	text_box.set_color(1,0,0,1);
	text_box.set_color_font(0,1,0,1);
	text_box.set_color_background(.5,0,0,1);


	graphic_box.set_tex(frame.display.textures->gui[1],3);
	graphic_box.set_color(1,0,0,1);
	graphic_box.set_dims(frame.display.screen_x,DROP_HEIGHT);
	graphic_box.set_type(GBT_TEXTURED, frame.display.textures->gui[0]);


	match_box.set_width(frame.display.screen_x-2);
	match_box.set_graphic(frame.display.textures->gui[2],2);
	match_box.set_font(FF_DOOM_SMALL_WHITE);
	match_box.set_type(TBT_READONLY);
	match_box.set_color(1,0,0,1);
	match_box.set_color_background(.5,0,0,1);
	match_box.set_color_font(0,.8f,0,1);

	command_list[COM_SET].set_matches(frame.config.get_var_list());
	command_list[COM_RESET].set_matches(frame.config.get_var_list());
	command_list[COM_TOGGLE].set_matches(frame.config.get_var_bool_list());
	command_list[COM_IMPORT_MODEL].set_matches(frame.config.get_data_file_list(".ply"));
	command_list[COM_IMPORT_FIELD].set_matches(frame.config.get_data_file_list(".vf"));


	for (int i=0; i<MAX_HISTORY; i++)	history[i][0] = 0;
	input_temp[0] = 0;
}

void COMMAND::set(char *str)
{
	sprintf(string,"%s",str);
	match = MATCH_NONE;
}


void CONSOLE::submit(char *str)
{
	if (str[0]==0) return;
	augment_history(str);
	if (parse(str))
		process();
}

bool CONSOLE::parse(char *str)
{
	int i,d;
	char arg[MAX_ARGUMENT_STRING_LENGTH];
	int max = (int)strlen(str);
	if (max>MAX_PARSE_LENGTH) max = MAX_PARSE_LENGTH;

	i = 0;

	//Parse command
	d = parse_arg(arg,str);
	i += d;
	str += d;

	//If command is bad, drop out here
	if (!set_command(arg))
		return false;


	task.num_arguments = 0;
	while (i<max)
	{
		d = parse_arg(arg,str);
		i += d;
		str += d;
		//We've got an argument

		add_argument(arg);
	}

	return true;
}

void CONSOLE::add_argument(char *arg)
{
	sprintf(task.argument[task.num_arguments].string,"%s",arg);

	task.argument[task.num_arguments].value = atof(arg);
	task.argument[task.num_arguments].integer = atoi(arg);

	task.num_arguments ++;

	if (task.num_arguments>=MAX_ARGUMENTS)
	{
		task.num_arguments = MAX_ARGUMENTS - 1;
		err_log << "WARNING: Too many arguments" << endl;
	}
}

int CONSOLE::parse_arg(char *arg, char *str)
{
	int n = 0;
	while (n<MAX_ARGUMENT_STRING_LENGTH && !is_white(*str) && *str!=0)
	{
		err_log << " PARSING *STR: " << *str << endl;
		arg[n] = *str;
		n++;
		str++;
	}
	arg[n] = 0;

	//Move to next non-whitespace section or end of string
	while (n<MAX_ARGUMENT_STRING_LENGTH && is_white(*str) && *str!=0)
	{
		n++;
		str++;
	}

	err_log << "N: " << n << endl;
	return n;
}

bool CONSOLE::is_white(char c)
{
	return (c == ' ');
}


bool CONSOLE::set_command(char *str)
{
	int i;

	i = match_command(str);
	err_log << "MATCHING COMMAND: " << str << endl;

	num_matches = 0;

	if (i<0)
	{
		bad_command(str);
		return false;
	}

	task.command = i;
	return true;
}


int CONSOLE::match_command()
{
	char tmp[TEXT_BOX_MAX_INPUT];
	text_box.read_input(tmp);
	return match_command(tmp);
}
int CONSOLE::match_command(char *str)
{
	int i, m;
	char tmp[TEXT_BOX_MAX_INPUT];
	clear_browse();
	strcpy(tmp,str);

	i=0;
	while (tmp[i]!=0)
	{
		if (tmp[i]==' ' || tmp[i]=='\t' || tmp[i]=='\r' || tmp[i]=='\n')
		{
			tmp[i] = 0;
			break;
		}
		i++;
	}


	num_matches = 0;
	for (i=0; i<MAX_COMMANDS; i++)
	{
		m = compare(command_list[i].string, tmp);
		command_list[i].match = m;
		if (m==MATCH_PARTIAL)
		{
			sprintf(match_list[num_matches],"%s",command_list[i].string);
			num_matches++;
		}
		if (m==MATCH_TOTAL)
		{
			//Check here if command corresponds to one with argument list
			// and if so, create match+argument matchlist!  OMG

			if (command_list[i].num_arg_matches==0)
			{
				num_matches = 0;
				return i;
			}
			else
			{
				num_matches = command_list[i].num_arg_matches;
				for (int j=0; j<num_matches; j++)
				{
					sprintf(match_list[j],"%s %s",command_list[i].string,command_list[i].match_args[j]);
				}
				fill_match_box();

				return i;
			}
		}
	}

	fill_match_box();

	return -1;
}

void CONSOLE::fill_match_box()
{
	int i;
	char temp[TEXT_BOX_MAX_INPUT*MATCH_COMMANDS_SHOWN];
	temp[0] = 0;

	int s,e;

	s = match_index+1;
	e = s + MATCH_COMMANDS_SHOWN - 1;
	if (e>=num_matches)
	{
		//We can have two behaviors here:
		// 1 - always have SHOWN if we have enough (uncommment line)
		// 2 - don't have a min rows requirement (leave comment) *looks better anyway 

//		s -= e - num_matches + 1;
		e -= e - num_matches + 1;
		if (s<0) s = 0;
	}

	match_box.set_rows(e-s+1);

	for (i=s; i<=e; i++)
		sprintf(temp,"%s%s\n",temp,match_list[i]);


	match_box.set_text(temp);
}

void CONSOLE::bad_command(char *str)
{
	frame.messages->add("ERROR: Bad command: %s",str);
	err_log << "ERROR: Bad command: " << str << endl;
}
void CONSOLE::bad_argument(char *str)
{
	frame.messages->add("ERROR: Bad argument: %s",str);
	err_log << "ERROR: Bad argument: " << str << endl;
}

void CONSOLE::more_arguments(int given, int needs)
{
	frame.messages->add("ERROR: Insufficient arguments: %i given, %i needed",given,needs);
	err_log << "ERROR: Insufficient arguments: " << given << " given, " << needs << " needed" << endl;
}


void CONSOLE::process()
{
	err_log << "COMMAND: ";
	switch (task.command)
	{
	default:
	case COM_NULL:
		{
			err_log << "NULL" << endl;
			break;
		}
	case COM_TOGGLE:
		{
			err_log << "TOGGLE" << endl;
			if (frame.config.toggle(task.argument[0].string) < 0)
				bad_argument(task.argument[0].string);
			break;
		}
	case COM_SET:
		{
			err_log << "SET: " << task.argument[0].string << " <- " << task.argument[1].value << endl;
			if (task.num_arguments<2)
				more_arguments(task.num_arguments,2);
			else if (frame.config.set(task.argument[0].string, task.argument[1].value) < 0)
				bad_argument(task.argument[0].string);
			break;
		}
	case COM_RESET:
		{
			err_log << "RESET" << endl;
			if (frame.config.reset(task.argument[0].string) < 0)
				bad_argument(task.argument[0].string);
			break;
		}
	case COM_QUIT:
		{
			frame.engine.set_state(ES_QUIT);
			break;
		}
	case COM_IMPORT_MODEL:
		{
			err_log << "IMPORT MODEL" << endl;

			if (task.num_arguments<1)
				more_arguments(task.num_arguments,2);
			else
			{
				char tmp[128];
				sprintf(tmp,"data/%s",task.argument[0].string);
				if (frame.add_model(tmp,task.argument[0].string))
					frame.init_bins();
				else
					bad_argument(task.argument[0].string);
			}
			
			break;
		}
	case COM_IMPORT_FIELD:
		{
			err_log << "IMPORT FIELD" << endl;

			if (task.num_arguments<1)
				more_arguments(task.num_arguments,2);
			else
			{
				char tmp[128];
				sprintf(tmp,"data/%s",task.argument[0].string);
				if (!frame.add_field(tmp))
					bad_argument(task.argument[0].string);
			}


			break;
		}
	}
}


TEXT_BOX *CONSOLE::activate()
{
	active = true;
	opening = true;
	closing = false;
	return &text_box;
}
void CONSOLE::deactivate()
{
	opening = false;
	closing = true;
	active = false;
}

//Fancypants name for add_to_history
void CONSOLE::augment_history(char *str)
{
	int i;
	for (i=MAX_HISTORY-1; i>0; i--)
		strcpy(history[i], history[i-1]);
	strcpy(history[0], str);
}

void CONSOLE::send_up()
{
	if (history_index==-1 && match_index==-1)	text_box.read_input(input_temp);
	if (match_index<0 && history_index<MAX_HISTORY-1)
	{
		if (history[0][0]!=0)
		{
			num_matches = 0;	//reset matches!  We don't care here!
			history_index++;
			if (history[history_index][0]==0) history_index--;
			text_box.set_text(history[history_index]);
		}
	}
	if (match_index>=0)
	{
		match_index--;
		if (match_index==-1)
		{
			text_box.set_text(input_temp);
			match_command(input_temp);
		}
		else text_box.set_text(match_list[match_index]);
	}
	fill_match_box();
}
void CONSOLE::send_down()
{
	if (history_index==-1 && match_index==-1)	text_box.read_input(input_temp);
	if (history_index<0 && match_index<num_matches-1)
	{
		match_index++;
		text_box.set_text(match_list[match_index]);
	}
	if (history_index>=0)
	{
		history_index--;
		if (history_index==-1)
		{
			text_box.set_text(input_temp);
			match_command(input_temp);
		}
		else text_box.set_text(history[history_index]);
	}
	fill_match_box();
}
void CONSOLE::clear_browse()
{
	match_index = -1;
	history_index = -1;
}



void CONSOLE::step()
{
	if (opening && closing)
	{
		opening = false;
		closing = false;
	}

	if (opening)
	{
		drop += drop_speed * frame.engine.get_super_time_factor();
		if (drop>=DROP_HEIGHT)
		{
			drop = DROP_HEIGHT;
			opening = false;
		}
	}
	if (closing)
	{
		drop -= drop_speed * frame.engine.get_super_time_factor();
		if (drop<=0)
		{
			drop = 0;
			closing = false;
		}
	}
}

void CONSOLE::draw()
{
	//Temp hack.  It is the end of the semester, and for some ungodly reason, if I try
	// to draw this here ALWAYS, then fonts work.  Seriously!  What.
	frame.magic.try_begin(1);


	if (drop<=0) return;

	int screen_y = frame.display.screen_y;
	int height = screen_y - drop;

	text_box.set_pos(1,height+1);
//	graphic_box.set_pos(0,height+DROP_HEIGHT);

	frame.magic.set_color(1,1,1,drop/DROP_HEIGHT);

	//Graphic box disabled because....  I need to divert output to it better first
//	graphic_box.draw();


	text_box.draw();

	if (num_matches>0 && !opening && !closing)
	{
		match_box.set_pos(1,height-input_height+1);
		match_box.draw();
	}

}






