//console.h

#ifndef _CONSOLE_
#define _CONSOLE_

#include "main.h"

#define MAX_ARGUMENTS				32
#define MAX_PARSE_LENGTH			1024
#define MAX_HISTORY					32
#define MATCH_COMMANDS_SHOWN		10


// **** These are the console command identifiers!
//These will have to be sorted alphabetically by their string counterpart (defined in control.cpp)
#define COM_NULL					0
#define COM_TOGGLE					1
#define COM_SET						2
#define COM_RESET					3
#define COM_QUIT					4
#define COM_IMPORT_MODEL			5
#define COM_IMPORT_FIELD			6




#define MATCH_NONE					0
#define MATCH_PARTIAL				1
#define MATCH_TOTAL					2

#define DROP_HEIGHT					70
#define DROP_SPEED_DEFAULT			150

struct COMMAND
{
	COMMAND::COMMAND();
	void set(char *str);
	char string[MAX_ARGUMENT_STRING_LENGTH];
	int match;

	char **match_args;
	int num_arg_matches;
	void set_matches(char *str);
};

struct ARGUMENT
{
	float value;
	int integer;
	char string[MAX_ARGUMENT_STRING_LENGTH];
};

struct TASK
{
	int command;
	ARGUMENT argument[MAX_ARGUMENTS];
	int num_arguments;
};

class CONSOLE
{
public:
	CONSOLE();

	void init();

	void submit(char *str);
	void process();
	bool parse(char *str);
	bool set_command(char *str);
	int parse_arg(char *arg, char *str);
	void add_argument(char *arg);

	int match_command(char *str);
	int match_command();

	void bad_command(char *str);
	void bad_argument(char *str);
	void more_arguments(int given, int needs);

	bool is_white(char c);

	void step();
	void draw();

	void send_up();
	void send_down();
	void clear_browse();

	void augment_history(char *str);

	TEXT_BOX *activate();
	void deactivate();

	void fill_match_box();

protected:
	bool active;
	TASK task;
	COMMAND command_list[MAX_COMMANDS];
	TEXT_BOX text_box;

	GRAPHIC_BOX graphic_box;
	TEXT_BOX match_box;

	char history[MAX_HISTORY][TEXT_BOX_MAX_INPUT];
	int history_index;

	char match_list[MAX_COMMANDS][TEXT_BOX_MAX_INPUT];	//will point to command_list entries //** We will need to change this to allow ARGUMENTS
	int match_index;
	int num_matches;

	char input_temp[TEXT_BOX_MAX_INPUT];

	float drop_speed;
	float drop;
	bool opening;
	bool closing;

	int input_height;	//height in pixels of input box
};








#endif


