//control.cpp

#include "main.h"

extern FRAME frame;
extern LOGGER err_log;
extern double rnd();
extern double rnd(float range);

static double last_wait;

CONTROL::CONTROL()
{
	state = CS_GAME;
	shift = false;
	caps = false;
	input_active = true;
	mod_camera = false;
	initialized = false;
}
CONTROL::~CONTROL()
{
	if (initialized)
	{
		for (int j=0; j<num_bindings; j++)
		{
			//Don't ask me why
			err_log << ".";
			//The following line causes a failure, but I believe I am done with this program so screw it.
//			free(binding_name[j]);
		}

		if (binding_str)	free(binding_str);
		if (binding_str2)	free(binding_str2);

		frame.key_names.destroy(key_name);
	}

	err_log << "Control destroyed." << endl;
}

int CONTROL::init()
{
	int i, j;
	slow_pop_counter = 0;
	prev_cursor.x = 0; prev_cursor.y = 0;

	calculating = false;
	get_viewport = true;

	rebuild_binding_str = true;
	binding_str = binding_str2 = NULL;

	active_viewport = 0;

	//Mouse sensitivity, callously assigned here
	mouse_sensitivity=.005;
	mouse_sensitivity_x=mouse_sensitivity;
	mouse_sensitivity_y=mouse_sensitivity;

	//This will create placeholders for bindings
	for (j=0; j<num_bindings; j++)
	{
		binding_name[j] = (char*)calloc(8,sizeof(char));
		sprintf(binding_name[j],"UNNAMED");
		binding[PRIMARY_KEY][j]=0;
		binding[SECONDARY_KEY][j]=0;
		binding_type[j]=CT_INSTANT;
		show_calculating[j]=false;
	}
	//All keys initialized to off
	for (i=0; i<MAX_KEYS; i++)
	{
		keys[i] = false;
		slow[i] = false;
	}
	click[0] = false;
	click[1] = false;
	click[2] = false;
	instant_click[0] = false;
	instant_click[1] = false;
	instant_click[2] = false;


	//Create list of actual key names (as in, "Numpad +" or things like that)
	frame.key_names.build(key_name);




/*
My glorious input/binding scheme.

The meaning of set_primary/set_secondary/set_both:

	set_primary() specifies the primary key, and set_secondary() specifies the secondary key.
	set_both() just lets you specify both in one call.

Parameters:

	name:			The integer associated with the action, defined in control.h
	string-name:	The string that is displayed to represent the action
	input type:		CT_TOGGLE=>key is turned off after one press, no repeats until key is pressed again
					CT_INSTANT=>action is run every frame wherein the key is held down
					CT_SLOW=>action is performed periodically while key is held down
	wait?:			Displays a waiting overlay (use for actions that take any noticable time to complete)
	key(s):			Which key to bind to it.  Note: if using set_both(), you will specify two keys.

*/

//				name						string-name						input type	wait?	key(s)
	set_primary(DO_QUIT,					"Quit",							CT_TOGGLE,	false,	SDLK_ESCAPE);
	set_primary(TOGGLE_CONSOLE,				"Toggle console",				CT_TOGGLE,	false,	'`');
	set_primary(TOGGLE_DPOLY,				"Toggle draw polygons",			CT_TOGGLE,	true,	'1');
	set_primary(TOGGLE_DWIRE_MODE,			"Toggle draw wireframe",		CT_TOGGLE,	true,	'2');
	set_primary(TOGGLE_DNORMAL,				"Toggle draw normals",			CT_TOGGLE,	true,	'3');
	set_primary(TOGGLE_DPOINTS,				"Toggle draw vertices",			CT_TOGGLE,	true,	'4');
	set_primary(TOGGLE_FAST_MOVE,			"Toggle movement speed",		CT_TOGGLE,	false,	'm');
	set_primary(TOGGLE_HIDDEN_LINE,			"Toggle hidden line removal",	CT_TOGGLE,	true,	'h');
	set_primary(CYCLE_GRID_MODE,			"Toggle grid mode",				CT_TOGGLE,	false,	'g');
	set_primary(TOGGLE_DRAW_AXES,			"Toggle draw axes",				CT_TOGGLE,	false,	'a');
	set_primary(CYCLE_DRAW_EXTENTS_MODE,	"Toggle draw extents",			CT_TOGGLE,	false,	'e');
	set_primary(TOGGLE_MESSAGES,			"Toggle show messages",			CT_TOGGLE,	false,	'n');
	set_primary(TOGGLE_DRAW_VECTORS_MODE,	"Toggle draw vector field",		CT_TOGGLE,	false,	'v');
	set_primary(TOGGLE_DRAW_VECTORS_MODE2,	"Toggle draw vector field alternate",CT_TOGGLE,	false,	'c');
	set_primary(VECTOR_FIELD_TEST,			"Vector field test",			CT_INSTANT, true,	's');
	set_primary(STEP_PHOTONS_ONCE,			"Step photons once",			CT_TOGGLE,	false,	'o');
	set_primary(SCREEN_SHOT,				"Screen shot",					CT_TOGGLE,	false,	',');
	set_primary(TOGGLE_DBINS,				"Toggle draw bins",				CT_TOGGLE,	false,	'b');
	set_primary(TOGGLE_DTRAILS,				"Toggle draw photon trails",	CT_TOGGLE,	false,	't');
	set_primary(TOGGLE_RUN_SIMULATION,		"Toggle simulation",			CT_TOGGLE,	false,	'r');
	set_primary(TOGGLE_DEBUG_LOG,			"Toggle debug log",				CT_TOGGLE,	false,	'd');
	set_primary(SPAWN_PHOTON,				"Spawn test photon",			CT_SLOW,	false,	'x');
	set_primary(MOD_CAMERA,					"Move camera (instant)",		CT_INSTANT,	false,	SDLK_SPACE);
	set_both(ZOOM_OUT,						"Zoom out",						CT_TOGGLE,	false,	MOUSE_WHEEL_DOWN,	'-');
	set_both(ZOOM_IN,						"Zoom in",						CT_TOGGLE,	false,	MOUSE_WHEEL_UP,		'=');
	set_primary(GRID_UP,					"Increase grid",				CT_TOGGLE,	false,	']');
	set_primary(GRID_DOWN,					"Decrease grid",				CT_TOGGLE,	false,	'[');
	set_primary(TOGGLE_CAMERA_LOOK,			"Move camera (toggle)",			CT_TOGGLE,	false,	'z');
	set_primary(TOGGLE_AUTO_REFRESH,		"Toggle auto refresh",			CT_TOGGLE,	false,	'j');

	set_primary(TOGGLE_SHOW_KEYS,			"Show key bindings",			CT_TOGGLE,	false,	SDLK_F1);
	set_primary(TOGGLE_SINGLE_VIEW,			"Toggle single viewport",		CT_TOGGLE,	false,	SDLK_F2);
	set_primary(TOGGLE_SHOW_VARS,			"Toggle show vars",				CT_TOGGLE,	false,	SDLK_F3);
	set_primary(BIN_DOWN,					"Reduce bin size",				CT_TOGGLE,	false,	SDLK_F4);
	set_primary(REFRESH,					"Refresh",						CT_TOGGLE,	true,	SDLK_F5);
	set_primary(FULL_SUBDIVISION,			"Run full subdivision",			CT_TOGGLE,	true,	SDLK_F6);
	set_primary(FORCE_SUBDIVISION,			"Force some subdivision",		CT_TOGGLE,	true,	SDLK_F7);
	set_primary(INVERT_MODEL,				"Invert model",					CT_TOGGLE,	true,	SDLK_F8);
	set_primary(TOGGLE_APPLY_ALL,			"Toggle apply all",				CT_TOGGLE,	false,	SDLK_F9);
	set_primary(CYCLE_MODEL_SELECTION,		"Cycle model selection",		CT_TOGGLE,	false,	SDLK_F10);
	set_primary(TOGGLE_INCREMENTAL_SUBDIV,	"Toggle incremental subdiv",	CT_TOGGLE,	false,	SDLK_F11);

	set_primary(TOGGLE_LIGHTING,			"Toggle lighting",				CT_TOGGLE,	true,	'l');
	set_primary(RUN_SIMULATION,				"Run simulation",				CT_TOGGLE,	true,	'y');
	set_primary(TOGGLE_FIRE_STRAIGHT,		"Toggle fire straight",			CT_TOGGLE,	true,	'u');
	set_primary(TOGGLE_DSPAWN_BOX,			"Toggle draw spawn box",		CT_TOGGLE,	false,	'f');
	set_primary(CYCLE_VISUALIZATION,		"Cycle visualization",			CT_TOGGLE,	true,	'q');

	set_primary(SUBDIVIDE_FIELD,			"Subdivide vector field",		CT_TOGGLE,	true,	'.');

	set_primary(DYE_TEST,					"Dye test",						CT_TOGGLE,	true,	'i');
	set_primary(TOGGLE_NORMALIZE_IBFV,		"Toggle normalize IBFV",		CT_TOGGLE,	true,	'p');


	initialized = true;
	return 1;
}



void CONTROL::set_primary(int action, char *str, int type, bool show_calc, int key)
{
	//If name exists, free it.  Then, create new space.
	if (binding_name[action]!=NULL) free(binding_name[action]);
	binding_name[action] = (char*)calloc(strlen(str)+1,sizeof(char));
	strcpy(binding_name[action],str);

	binding[PRIMARY_KEY][action]=key;
	binding_type[action]=type;
	show_calculating[action]=show_calc;
}
void CONTROL::set_secondary(int action, char *str, int type, bool show_calc, int key)
{
	if (binding_name[action]!=NULL) free(binding_name[action]);
	binding_name[action] = (char*)calloc(strlen(str)+1,sizeof(char));
	strcpy(binding_name[action],str);

	binding[SECONDARY_KEY][action]=key;
	binding_type[action]=type;
	show_calculating[action]=show_calc;
}
void CONTROL::set_both(int action, char *str, int type, bool show_calc, int key1, int key2)
{
	if (binding_name[action]!=NULL) free(binding_name[action]);
	binding_name[action] = (char*)calloc(strlen(str)+1,sizeof(char));
	strcpy(binding_name[action],str);

	binding[PRIMARY_KEY][action]=key1;
	binding[SECONDARY_KEY][action]=key2;
	binding_type[action]=type;
	show_calculating[action]=show_calc;
}

void CONTROL::handle_events()
{
SDL_Event event;

    while( SDL_PollEvent( &event ) )
	{
        switch( event.type )
		{
		case SDL_VIDEORESIZE:
			{
				err_log << "RESIZE" << endl;

				//Nah

				break;
			}
		case SDL_ACTIVEEVENT:
			{
				//Keeps track of input focus

				if (event.active.gain==1 && event.active.state & SDL_APPACTIVE)
					input_active = true;
				if (event.active.gain==0 && event.active.state & SDL_APPINPUTFOCUS)
					input_active = false;

				break;
			}

		//And the rest of these just keep track of buttons up/down

		case SDL_KEYDOWN:
			{
				slow_pop_counter = -SLOW_POP_DELAY;
				keys[ event.key.keysym.sym ] = true;
	            break;
			}
        case SDL_KEYUP:
			{
				keys[ event.key.keysym.sym ] = false;
				slow[ event.key.keysym.sym ] = false;
	            break;
			}
        case SDL_QUIT:
			{
				frame.quit(1);
				break;
			}
		default:
			{
				if (event.type==4) break;
				err_log << "Unknown event: " << event.type << endl;
				break;
			}
		case SDL_MOUSEBUTTONDOWN:	//5:	//Mouse down
			{
				get_viewport = true;
				if (event.button.button==SDL_BUTTON_LEFT)
				{
					keys[MOUSE_LEFT] = true;
					if (!click[0]) instant_click[0] = true;
					click[0]=true;
					break;
				}
				else if (event.button.button==SDL_BUTTON_MIDDLE)
				{
					keys[MOUSE_MIDDLE] = true;
					if (!click[1]) instant_click[1] = true;
					click[1]=true;
					break;
				}
				else if (event.button.button==SDL_BUTTON_RIGHT)
				{
					keys[MOUSE_RIGHT] = true;
					if (!click[2]) instant_click[2] = true;
					click[2]=true;
					break;
				}
			}
		case SDL_MOUSEBUTTONUP:	//6:	//Mouse up
			{
				if (event.button.button==SDL_BUTTON_WHEELUP)
				{
					keys[MOUSE_WHEEL_UP] = true;
					break;
				}
				if (event.button.button==SDL_BUTTON_WHEELDOWN)
				{
					keys[MOUSE_WHEEL_DOWN] = true;
					break;
				}
				if (event.button.button==SDL_BUTTON_LEFT)
				{
					keys[MOUSE_LEFT] = false;
					click[0]=false;
					break;
				}
				else if (event.button.button==SDL_BUTTON_MIDDLE)
				{
					keys[MOUSE_MIDDLE] = false;
					click[1]=false;
					break;
				}
				else if (event.button.button==SDL_BUTTON_RIGHT)
				{
					keys[MOUSE_RIGHT] = false;
					click[2]=false;
					break;
				}
			}
		case WM_LBUTTONUP:
			{
				break;
			}
		}
	}

	//Increment slow pop and see if it has exceeded its timer
	slow_pop_counter += frame.engine.get_super_time_difference();
	if (slow_pop_counter>=SLOW_POP_REPEAT)
	{
		slow_pop_counter = 0;
		slow_pop();
	}

	if (input_active)
		get_actions();
}

/*
What is this slow pop?

CT_TOGGLE makes an action occur only once when a key is pressed.
CT_INSTANT makes an action occur continually while a key is pressed.
What if you want it to perform the action periodically while holding it down,
but less often than CT_INSTANT?
That is what this does.  Every SLOW_POP_REPEAT seconds, any CT_SLOW
key will be given the chance to occur again, if the key is still pressed.

By default, this is used for any input into text boxes, such as the console.

*/
void CONTROL::slow_pop()
{
	int i;
	for (i=0; i<MAX_KEYS; i++)
	{
		if (slow[i])
		{
			keys[i] = true;
		}
	}
}

void CONTROL::convert_input()
{
	//Check which viewport is active
	//what hax
	if (get_viewport)// || (!mod_camera && (click[0] || click[1])))
	{
		get_viewport = false;
		int new_viewport = frame.display.viewports.get_viewport(cursor_pos.x-cursor_offset.x,cursor_pos.y-cursor_offset.y);
		if (new_viewport!=-1)
			active_viewport = new_viewport;
	}
	if (active_viewport < 0) active_viewport = 0;

	for (long i=0;i<num_bindings;i++)
	{
		if (check_binding(i))
		{
			last_wait = 0;
			if (show_calculating[i])
			{
				frame.display.overlay->show_calculating(binding_name[i]);
				calculating = true;
				last_wait = frame.engine.get_time();
			}

			handle_action(i);

			if (binding_type[i]==CT_TOGGLE)
			{
				keys[binding[PRIMARY_KEY][i]]=false;
				keys[binding[SECONDARY_KEY][i]]=false;
			}
			if (binding_type[i]==CT_SLOW)
			{
				if (keys[binding[PRIMARY_KEY][i]])		slow[binding[PRIMARY_KEY][i]]=true;
				if (keys[binding[SECONDARY_KEY][i]])	slow[binding[SECONDARY_KEY][i]]=true;
				keys[binding[PRIMARY_KEY][i]]=false;
				keys[binding[SECONDARY_KEY][i]]=false;
			}
		}
	}


	POINT cursor_dif;
	update_cursor();
	if ((mod_camera || frame.config.get_bool(VAR_CAMERA_LOOK)) && (click[0] || click[2]))
	{
		mod_camera = false;
		cursor_dif.x = prev_cursor.x - cursor_pos.x;
		cursor_dif.y = prev_cursor.y - cursor_pos.y;

		//For some fairly odd reason, if I do this while in fullscreen,
		// it resets the mouse to the center of the screen, which
		// is definitely no good.
		if (!frame.display.is_fullscreen())
			SDL_ShowCursor(0);

		frame.display.viewports.vp[active_viewport].set_camera();
		frame.camera->mouse_move(&cursor_dif,click);

		revert_cursor();
	}
	else
		SDL_ShowCursor(1);

}


void CONTROL::update_cursor()
{
	int X, Y;
//*	GetCursorPos(&abs_cursor_pos);
	SDL_GetMouseState(&hax_cursor[0],&hax_cursor[1]);
	abs_cursor_pos.x = hax_cursor[0];
	abs_cursor_pos.y = hax_cursor[1];

	SDL_GetMouseState(&X,&Y);
	cursor_offset.x = abs_cursor_pos.x - X;
	cursor_offset.y = abs_cursor_pos.y - Y;

	prev_cursor.x = cursor_pos.x;
	prev_cursor.y = cursor_pos.y;
//*	GetCursorPos(&cursor_pos);
	SDL_GetMouseState(&hax_cursor[0],&hax_cursor[1]);
	cursor_pos.x = hax_cursor[0];
	cursor_pos.y = hax_cursor[1];
}

void CONTROL::center_cursor()
{
	//This should never happen anymore

//	SetCursorPos(frame.display.screen_half_x,frame.display.screen_half_y);
}
void CONTROL::revert_cursor()
{
	//Move cursor to previous position

	cursor_pos.x = prev_cursor.x;
	cursor_pos.y = prev_cursor.y;
//*	SetCursorPos(prev_cursor.x,prev_cursor.y);
	hax_cursor[0] = prev_cursor.x;
	hax_cursor[1] = prev_cursor.y;
	SDL_WarpMouse(hax_cursor[0],hax_cursor[1]);



	POINT crap;
//*	GetCursorPos(&crap);
	SDL_GetMouseState(&hax_cursor[0],&hax_cursor[1]);
	crap.x = hax_cursor[0];
	crap.y = hax_cursor[1];
}

bool CONTROL::check_binding(int a)
{
	return keys[binding[PRIMARY_KEY][a]] || keys[binding[SECONDARY_KEY][a]];
}

void CONTROL::get_actions()
{
	int i;
	char c;

	actions = 0;

	switch (state)
	{
	case CS_MENU:
		{
			//If a menu state exists
			break;
		}
	default:
	case CS_GAME:
		{
			//If a game state is active (i.e., just normal operation)
			convert_input();
			break;
		}
	case CS_TEXT_BOX:
		{
			//If a text box is active (such as the console)

			if (keys[TEXT_BOX_ESCAPE])
			{
				current_text_box->flush();
			}

			if (keys[SDLK_RSHIFT] || keys[SDLK_LSHIFT]) shift = true;
			else shift = false;

			if (check_binding(TOGGLE_CONSOLE))
			{
				convert_input();
			}
			for (i=32;i<=126;i++)
			{
				if (keys[i])
				{
					c = (char)i;
					current_text_box->send_input(c,shift,caps);
				}
			}

			if (keys[SDLK_LEFT]) current_text_box->send_left();
			if (keys[SDLK_RIGHT]) current_text_box->send_right();
			if (keys[SDLK_BACKSPACE]) current_text_box->send_backspace();
			if (keys[SDLK_DELETE]) current_text_box->send_delete();

			if (keys[SDLK_UP]) current_text_box->send_up();
			if (keys[SDLK_DOWN]) current_text_box->send_down();


			if (keys[SDLK_LEFT])		slow[SDLK_LEFT] = true;
			if (keys[SDLK_RIGHT])		slow[SDLK_RIGHT] = true;
			if (keys[SDLK_BACKSPACE])	slow[SDLK_BACKSPACE] = true;
			if (keys[SDLK_DELETE])		slow[SDLK_DELETE] = true;
			if (keys[SDLK_UP])			slow[SDLK_UP] = true;
			if (keys[SDLK_DOWN])		slow[SDLK_DOWN] = true;

			keys[SDLK_LEFT] = false;
			keys[SDLK_RIGHT] = false;
			keys[SDLK_BACKSPACE] = false;
			keys[SDLK_DELETE] = false;
			keys[SDLK_UP] = false;
			keys[SDLK_DOWN] = false;

			for (i=32;i<=126;i++)
			{
				if (i=='`')
				{
					continue;
				}

				if (keys[i]) slow[i] = true;
				keys[i] = false;
			}

			convert_input();

			break;
		}
	}
}

void CONTROL::handle_action(int which_action)
{
	mod_camera = false;
	int j;

	switch (which_action)
	{
	case DO_QUIT:
		{
			err_log << "A quit command is issued from the keyboard." << endl;
			frame.finish();
//			frame.quit(1);
			break;
		}
	case TOGGLE_CONSOLE:
		{
			if (state==CS_GAME)
			{
				current_text_box = frame.console.activate();
				state = CS_TEXT_BOX;
			}
			else
			{
				frame.console.deactivate();
				current_text_box = NULL;
				state = CS_GAME;
			}
			break;
		}

// *** Viewport-specific toggles/cycles ***
	case TOGGLE_DPOLY:			frame.display.viewports.vp[active_viewport].cycle_poly();	break;
	case TOGGLE_DWIRE_MODE:		frame.display.viewports.vp[active_viewport].cycle_wire();	break;
	case TOGGLE_DNORMAL:		frame.display.viewports.vp[active_viewport].cycle_normal();	break;
	case TOGGLE_DPOINTS:		frame.display.viewports.vp[active_viewport].cycle_points();	break;
	case TOGGLE_HIDDEN_LINE:	frame.display.viewports.vp[active_viewport].cycle_hidden();	break;
	case CYCLE_GRID_MODE:		frame.display.viewports.vp[active_viewport].cycle_grid();	break;
	case ZOOM_OUT:				frame.display.viewports.vp[active_viewport].zoom_out();		break;
	case ZOOM_IN:				frame.display.viewports.vp[active_viewport].zoom_in();		break;
	case GRID_UP:				frame.display.viewports.vp[active_viewport].grid_up();		break;
	case GRID_DOWN:				frame.display.viewports.vp[active_viewport].grid_down();	break;
	case TOGGLE_LIGHTING:		frame.display.viewports.vp[active_viewport].cycle_lighting(); break;
	case TOGGLE_DSPAWN_BOX:		frame.display.viewports.vp[active_viewport].cycle_spawn_box(); break;
	case CYCLE_VISUALIZATION:	frame.display.viewports.vp[active_viewport].cycle_visualization(); break;


// *** Config-specific toggles/cycles ***
	case TOGGLE_SINGLE_VIEW:		frame.config.toggle(VAR_SINGLE_VIEW);			break;
	case TOGGLE_DBINS:				frame.config.toggle(VAR_DBINS);					break;
	case TOGGLE_DTRAILS:			frame.config.toggle(VAR_DTRAILS);				break;
	case TOGGLE_FAST_MOVE:			frame.config.toggle(VAR_FAST_MOVE);				break;
	case TOGGLE_CAMERA_LOOK:		frame.config.toggle(VAR_CAMERA_LOOK);			break;
	case CYCLE_DRAW_EXTENTS_MODE:	frame.config.cycle(VAR_DRAW_EXTENTS_MODE,1);	break;
	case TOGGLE_DRAW_VECTORS_MODE:	frame.config.cycle(VAR_DRAW_VECTORS_MODE,1);	break;
	case TOGGLE_DRAW_VECTORS_MODE2:	frame.config.cycle(VAR_DRAW_VECTORS_MODE2,1);	break;
	case TOGGLE_DRAW_AXES:			frame.config.toggle(VAR_DRAW_AXES);				break;
	case TOGGLE_MESSAGES:			frame.config.toggle(VAR_DRAW_MESSAGES);			break;
	case TOGGLE_RUN_SIMULATION:		frame.config.toggle(VAR_RUN_SIMULATION);		break;
	case TOGGLE_DEBUG_LOG:			frame.config.toggle(VAR_DEBUG_LOG);				break;
	case TOGGLE_SHOW_KEYS:			frame.config.toggle(VAR_SHOW_KEYS);				break;
	case TOGGLE_AUTO_REFRESH:		frame.config.toggle(VAR_AUTO_REFRESH);			break;
	case TOGGLE_FIRE_STRAIGHT:		frame.config.toggle(VAR_FIRE_STRAIGHT);			break;
	case TOGGLE_APPLY_ALL:			frame.config.toggle(VAR_APPLY_ALL);				break;
	case TOGGLE_SHOW_VARS:			frame.config.toggle(VAR_SHOW_VARS);				break;
	case TOGGLE_INCREMENTAL_SUBDIV:	frame.config.toggle(VAR_INCREMENTAL_SUBDIV);	break;


// *** Actions ***
	case STEP_PHOTONS_ONCE:			frame.photon_list.step();									break;
	case SCREEN_SHOT:				frame.display.screen_shot();								break;
	case MOD_CAMERA:				mod_camera = true;											break;
	case REFRESH:
		{
			for (j=0; j<MAX_MODELS; j++)
				frame.model[j].recalculate();

			break;
		}

	case TOGGLE_NORMALIZE_IBFV:
		{
			frame.config.toggle(VAR_NORMALIZE_IBFV);
			frame.ibfv.set_faces();
			break;
		}
	case DYE_TEST:
		{
			frame.ibfv.dye_test();
			break;
		}
	case SUBDIVIDE_FIELD:
		{
			frame.vector_field->subdivide_field(1);
			break;
		}
	case RUN_SIMULATION:
		{
			frame.photon_list.run_simulation();
			break;
		}
	case FULL_SUBDIVISION:
		{
			//This should be selective -- use current_model only

			err_log << "lewlz" << endl;

			if (!frame.config.get_bool(VAR_APPLY_ALL))
			{
				err_log << "Apply to current model: " << frame.current_model << endl;
				if (frame.current_model<0) break;
				frame.model[frame.current_model].full_subdivision();
				frame.model[frame.current_model].recalculate();
			}
			else
			{
				err_log << "APPLY ALL" << endl;
				for (j=0; j<MAX_MODELS; j++)
				{
					err_log << "Use: " << j << endl;
					frame.model[j].full_subdivision();
					frame.model[j].recalculate();
				}
			}
			break;
		}
	case FORCE_SUBDIVISION:
		{
//			if (frame.current_model<0) break;
//			frame.model[frame.current_model].subdivide();
//			frame.model[frame.current_model].recalculate();

			//This performs cleanup subdivision, so it may as well happen to ALL models
			for (j=0; j<MAX_MODELS; j++)
			{
				frame.model[j].subdivide();
				frame.model[j].recalculate();
			}
			break;
		}
	case INVERT_MODEL:
		{
			//This should be selective -- use current_model only

			if (!frame.config.get_bool(VAR_APPLY_ALL))
			{
				if (frame.current_model<0) break;
				frame.model[frame.current_model].do_invert();
				frame.model[frame.current_model].recalculate();
			}
			else
			{
				for (j=0; j<MAX_MODELS; j++)
				{
					frame.model[j].do_invert();
					frame.model[j].recalculate();
				}
			}
			break;
		}
	case CYCLE_MODEL_SELECTION:
		{
			frame.inc_model();
			break;
		}

	case BIN_DOWN:
		{
			frame.bin_down(true);
			break;
		}


// *** Temporary testing actions ***
	case VECTOR_FIELD_TEST:
		{
//			for (j=0; j<MAX_MODELS; j++)
//				frame.model[j].vector_field_test();

			if (frame.current_model<0) break;
			frame.model[frame.current_model].vector_field_test();
			break;
		}

	case SPAWN_PHOTON:
		{
			//This fires a photon from the current camera,
			// in the direction of look.  The origin of the photon
			// is slightly offset to the right of the camera (relative
			// to the camera's orientation) so that the photon trail
			// is more interesting looking.
			//This is basically a temporary feature, so it doesn't
			// need its own function.

			double a[3], v[3], r[3], u[3];
			frame.camera->update_camera();
			a[0] = frame.camera->at.x;
			a[1] = frame.camera->at.y;
			a[2] = frame.camera->at.z;
			v[0] = frame.camera->look.x - a[0];
			v[1] = frame.camera->look.y - a[1];
			v[2] = frame.camera->look.z - a[2];
			u[0] = 0;
			u[1] = 0;
			u[2] = 1;

			cross(r, v, u);

			double scale=.00065;
			r[0] *= scale;
			r[1] *= scale;
			r[2] *= scale;

			a[0] += r[0];
			a[1] += r[1];
			a[2] += r[2];

			frame.photon_list.add(a[0],a[1],a[2],v[0],v[1],v[2]);
			break;
		}

	default: break;
	}


	//If a command had been issued which was marked as requiring a waiting
	// screen, then this will calculate the time spent waiting on that
	// command, and will post it as a message upon completion of the command.
	if (last_wait!=0)
	{
		frame.engine.set_time_factor();
		frame.messages->add("Waited %f seconds for %s",frame.engine.get_time() - last_wait,binding_name[which_action]);
		last_wait = 0;
	}

}


void CONTROL::set_current_text_box(TEXT_BOX *new_text_box)
{
	//Set a new text box for current input
	current_text_box = new_text_box;
}


void CONTROL::show_bindings()
{
	int row_nums = 41;	//40
	int i, n = 0;
	if (rebuild_binding_str || binding_str==NULL)
	{
		if (binding_str!=NULL) free(binding_str);
		if (binding_str2!=NULL) free(binding_str2);

		n += 13;													//A
		for (i=1; i<num_bindings; i++)
		{
			if (i>row_nums) break;
			n += (int)strlen(binding_name[i]);						//B
			n += 5;													//C
			n += (int)strlen(key_name[binding[PRIMARY_KEY][i]]);	//D
			n += 1;													//E
			n += (int)strlen(key_name[binding[SECONDARY_KEY][i]]);	//F
			n += 1;													//G
		}
		n++;	//terminating char
		binding_str = (char*)calloc(n,sizeof(char));

		sprintf(binding_str,"*** Keys ***\n");							//A
		for (i=1; i<num_bindings; i++)
		{
			if (i>row_nums) break;
			strcat(binding_str,key_name[binding[PRIMARY_KEY][i]]);		//D
			strcat(binding_str," ");								//E
			strcat(binding_str,key_name[binding[SECONDARY_KEY][i]]);	//F
			strcat(binding_str,"    :");								//C
			strcat(binding_str,binding_name[i]);						//B
			strcat(binding_str,"\n");									//G
		}

		err_log << "LENGTH: " << strlen(binding_str) << endl;
		err_log << "ALLOCATED: " << n << endl;


		n = 0;
		//Spill to second column.  I COULD make this more dynamic, allowing more
		// columns than two.  But one already contains more keys than you can keep track of.
		// So, there is no need for MORE than two -- that would be insane.
		if (num_bindings>row_nums)
		{
			n += 1;														//A
			for (i=row_nums; i<num_bindings; i++)
			{
				n += (int)strlen(binding_name[i]);						//B
				n += 5;													//C
				n += (int)strlen(key_name[binding[PRIMARY_KEY][i]]);	//D
				n += 1;													//E
				n += (int)strlen(key_name[binding[SECONDARY_KEY][i]]);	//F
				n += 1;													//G
			}
			n++;	//terminating char
			binding_str2 = (char*)calloc(n,sizeof(char));

			sprintf(binding_str2,"\n");					//A
			for (i=row_nums; i<num_bindings; i++)
			{
				strcat(binding_str2,key_name[binding[PRIMARY_KEY][i]]);		//D
				strcat(binding_str2," ");									//E
				strcat(binding_str2,key_name[binding[SECONDARY_KEY][i]]);	//F
				strcat(binding_str2,"     ");								//C
				strcat(binding_str2,binding_name[i]);						//B
				strcat(binding_str2,"\n");									//G
			}
		}

		err_log << "LENGTH: " << strlen(binding_str2) << endl;
		err_log << "ALLOCATED: " << n << endl;



	}
	rebuild_binding_str = false;

	frame.display.fonts->draw(FF_TIMES_SMALL, binding_str,400,frame.display.screen_y-15);
	if (binding_str2!=NULL)
		frame.display.fonts->draw(FF_TIMES_SMALL, binding_str2,720,frame.display.screen_y-15);


}



