//control.h
//
//Controls input states.  Actions are linked to keys/mouse actions.

#ifndef _CONTROL_
#define _CONTROL_

#include "main.h"

//#include <windows.h>

#define MAX_KEYS		500

#define PRIMARY_KEY		0
#define SECONDARY_KEY	1


//**** Virtual Keys
//keys 0-322 are DEFINITELY reserved.  More may be,
// so if we assume that 0-399 are reserved, this leaves
// us with 100 virtual keys for whatever use we'd like.
#define MOUSE_LEFT		400
#define MOUSE_MIDDLE	401
#define MOUSE_RIGHT		402
#define MOUSE_WHEEL_UP	403
#define MOUSE_WHEEL_DOWN 404



#define WM_MOUSEWHEEL 0x020A
#define WHEEL_DELTA 120

#define CT_TOGGLE		0
#define CT_INSTANT		1
#define CT_SLOW			2

#define MAX_ACTIONS		64

/*
How to handle action bindings?

1. Define an action integer (such as "#define TOGGLE_CONSOLE 2") with a unique number
2. Adjust num_bindings to reflect the new binding count (it should equal the number of defined actions)
3. Set the key binding (see set_primary()/etc comments in control.cpp, init())
4. Add a case for the action integer in control.cpp, handle_action()

*/


//Various actions.  If more are added, just adjust num_bindings to reflect it.
#define ACTION_NULL					0
#define DO_QUIT						1

#define TOGGLE_CONSOLE				2
#define TOGGLE_DPOLY				3
#define TOGGLE_DWIRE_MODE			4
#define TOGGLE_DNORMAL				5
#define TOGGLE_FAST_MOVE			6
#define TOGGLE_HIDDEN_LINE			7
#define CYCLE_GRID_MODE				8
#define TOGGLE_DRAW_AXES			9
#define CYCLE_DRAW_EXTENTS_MODE		10
#define TOGGLE_MESSAGES				11
#define TOGGLE_DRAW_VECTORS_MODE	12
#define TOGGLE_DPOINTS				13
#define TOGGLE_DBINS				14
#define TOGGLE_DTRAILS				15
#define TOGGLE_LIGHTING				16
#define TOGGLE_DSPAWN_BOX			17
#define CYCLE_VISUALIZATION			18

#define RUN_SIMULATION				19
#define STEP_PHOTONS_ONCE			20
#define TOGGLE_RUN_SIMULATION		21
#define SPAWN_PHOTON				22
#define TOGGLE_FIRE_STRAIGHT		23
#define TOGGLE_AUTO_REFRESH			24

#define ZOOM_OUT					25
#define ZOOM_IN						26
#define GRID_UP						27
#define GRID_DOWN					28
#define MOD_CAMERA					29
#define TOGGLE_CAMERA_LOOK			30

#define TOGGLE_SHOW_KEYS			31
#define TOGGLE_SINGLE_VIEW			32
#define TOGGLE_SHOW_VARS			33
#define BIN_DOWN					34
#define REFRESH						35
#define FULL_SUBDIVISION			36
#define FORCE_SUBDIVISION			37
#define INVERT_MODEL				38
#define TOGGLE_APPLY_ALL			39
#define CYCLE_MODEL_SELECTION		40
#define TOGGLE_INCREMENTAL_SUBDIV	41

#define VECTOR_FIELD_TEST			42
#define TOGGLE_DEBUG_LOG			43
#define SCREEN_SHOT					44

#define SUBDIVIDE_FIELD				45
#define TOGGLE_DRAW_VECTORS_MODE2	46

#define DYE_TEST					47
#define TOGGLE_NORMALIZE_IBFV		48

const int num_bindings = 49;	//Update this whenever you add/remove actions from above defines!



const double rad90 = 1.570796327;
const double rad89 = 1.553343034;


#define CS_GAME				1
#define CS_MENU				2
#define CS_TEXT_BOX			3

#define TEXT_BOX_ESCAPE SDLK_RETURN

#define SLOW_POP_DELAY	.25f
#define SLOW_POP_REPEAT	.1f
class CONTROL
{
public:
	CONTROL();
	~CONTROL();

	int init();

	void update_cursor();

	void handle_events();
	void convert_input();
	void get_actions();
	void handle_action(int which_action);
//	void handle_actions();

	bool check_binding(int a);

	void show_bindings();

	//Periodically, this will be called.  It will go through slow[],
	// and for every true value, it will enable the appropriate key
	void slow_pop();

	void set_primary(int action, char *str, int type, bool show_calc, int key);
	void set_secondary(int action, char *str, int type, bool show_calc, int key);
	void set_both(int action, char *str, int type, bool show_calc, int key1, int key2);

	void center_cursor();
	void revert_cursor();

	char *binding_name[num_bindings];
	int binding[2][num_bindings];
	int binding_type[num_bindings];
	int show_calculating[num_bindings];


	bool keys[MAX_KEYS];
	bool click[3];
	bool instant_click[3];


	bool slow[MAX_KEYS];


	double mouse_sensitivity;
	double mouse_sensitivity_x;
	double mouse_sensitivity_y;


	void set_current_text_box(TEXT_BOX *new_text_box);

	//might want to move the other crap in here ;P

	int active_viewport;
protected:
	int state;
	TEXT_BOX *current_text_box;
	bool shift;
	bool caps;

	bool calculating;
	long actions;
	float slow_pop_counter;

	POINT cursor_pos;
	POINT prev_cursor;
	POINT abs_cursor_pos;
	POINT cursor_offset;
//	XYZ cursor_pos;

	SDL_Cursor *sdl_cursor;
	int hax_cursor[2];	//ugh!

	char *key_name[MAX_KEYS];

	//binding_str will contain explicit text code to draw.
	// If bindings get changed, just set rebuild_binding_str to true,
	// and binding_str will be rebuilt
	char *binding_str;
	char *binding_str2;
	bool rebuild_binding_str;

	bool get_viewport;
	bool mod_camera;
	bool input_active;

	bool initialized;
};

#endif

