//display.cpp

#include "main.h"

extern FRAME frame;
extern LOGGER err_log;

DISPLAY::DISPLAY()
{
//	background_color[0] = 0;
//	background_color[1] = 0;
//	background_color[2] = 0;

	background_color[0] = .34f;
	background_color[1] = .34f;
	background_color[2] = .15f;
	background_color[3] = 1;

	initialized = false;
	screen_bits = 16;		//16

	pworld_x = (1.0f/3.0f);		//320
	pworld_y = (1.0f/4.0f);		//240

//	set_screen(1280,1024);
	set_screen(1024,768);
	set_world(1,1);

	fullscreen = false;
	last_screen_shot = -1;

}

DISPLAY::~DISPLAY()
{
	destroy();
}

void DISPLAY::destroy()
{
	if (!initialized) return;

	GLuint t[MAX_TEXTURES];
	int i = 1;
	int n = 0;
	while (i<MAX_TEXTURES)
	{
		if (texture_data->active)
		{
			t[n] = i;
			n++;
		}
		i++;
	}
	glDeleteTextures(n, t);

	delete textures;
	delete [] texture_data;
	delete fonts;
	delete overlay;
	delete extra;

	err_log << "Display destroyed." << endl;
}

//int DISPLAY::init(FRAME *this_frame)
int DISPLAY::init()
{
	enables();

	textures = new TEXTURES;
	texture_data = new TEXTURE_DATA[MAX_TEXTURES];
	init_textures();

	fonts = new FONTS;
	fonts->build_fonts();

//*	ShowCursor(true);
	SDL_ShowCursor(true);

//	SDL_ShowCursor(SDL_DISABLE);
//	SDL_ShowCursor(SDL_ENABLE);

	overlay = new OVERLAY;
	overlay->init();

	extra = new EXTRA;


	initialized = true;
	return 1;
}

void DISPLAY::set_screen(int x, int y)
{
	screen_x = x;
	screen_y = y;
	screen_half_x = screen_x/2;
	screen_half_y = screen_y/2;
}

void DISPLAY::set_world(float fx, float fy)
{
	world_x = pworld_x * fx;
	world_y = pworld_y * fy;
}

void DISPLAY::enables()
{
	glEnable(GL_TEXTURE_2D);
	glShadeModel(GL_SMOOTH);
	glClearColor(background_color[0],background_color[1],background_color[2],background_color[3]);
	glClearDepth(1.0f);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

	glFogfv(GL_FOG_COLOR,background_color);
	glFogf(GL_FOG_DENSITY,25);

	glCullFace(GL_BACK);
	glEnable(GL_CULL_FACE);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

//	glEnable(GL_LIGHTING);

	GLfloat light_mat_diff[] = {1,1,1,1};
	GLfloat light_mat_spec[] = {1,1,1,1};
	GLfloat light_mat_shin[] = {500};
	GLfloat light_white[] = {1,1,1,0};
	GLfloat light_ambient[] = {.25f,.25f,.25f,0};

	glEnable(GL_LIGHT0);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, light_mat_diff);
	glMaterialfv(GL_FRONT, GL_SPECULAR, light_mat_spec);
	glMaterialfv(GL_FRONT, GL_SHININESS, light_mat_shin);
	glLightfv(GL_LIGHT0,GL_AMBIENT, light_ambient);
	glLightfv(GL_LIGHT0,GL_DIFFUSE, light_white);
	glLightfv(GL_LIGHT0,GL_SPECULAR, light_white);
}


SDL_Surface *DISPLAY::load_bmp(const char *Filename)
{
	FILE *File=NULL;

	if (!Filename)
		return NULL;

	File=fopen(Filename,"r");

	if (File)
	{
		fclose(File);
//		return auxDIBImageLoad(Filename);	// Load The Bitmap And Return A Pointer
		return SDL_LoadBMP(Filename);
	}

	return NULL;
}

int DISPLAY::init_texture(const char *filename)
{
	return init_texture(filename, GL_CLAMP_TO_EDGE, GL_NEAREST, GL_RGBA, false);
}
int DISPLAY::init_texture(const char *filename, GLuint wrap_param)
{
	return init_texture(filename, wrap_param, GL_NEAREST, GL_RGBA, false);
}
int DISPLAY::init_texture(const char *filename, float x, float y, float z)
{
	return init_texture(filename, GL_CLAMP_TO_EDGE, GL_NEAREST, GL_RGBA, false, x,y,z);
}
int DISPLAY::init_texture(const char *filename, GLuint wrap_param, GLuint mag_filter, GLuint internal, bool use_mipmaps)
{
	return init_texture(filename, wrap_param, mag_filter, internal, use_mipmaps, 0,1,1);
}
int DISPLAY::init_texture(const char *filename, GLuint wrap_param, GLuint mag_filter, GLuint internal, bool use_mipmaps, float x, float y, float z)
{
	int dimx, dimy;
	GLuint texture = 2;												// Texture ID

//	AUX_RGBImageRec *pImage;										// Create Storage Space For The Texture
	SDL_Surface *pImage;

	pImage = load_bmp( filename );									// Loads The Bitmap Specified By filename

	if (pImage==NULL)	frame.fail("Unable to open texture file %s!",filename);

	// Load The Bitmap, Check For Errors, If Bitmap's Not Found Quit
	if ( pImage != NULL && pImage->pixels != NULL )					// If Texture Image Exists
	{
		glGenTextures(1, &texture);
		dimx = pImage->w;
		dimy = pImage->h;

		// Typical Texture Generation Using Data From The Bitmap
		glBindTexture(GL_TEXTURE_2D, texture);

		glPixelTransferf(GL_RED_SCALE, 1);
		glPixelTransferf(GL_GREEN_SCALE, 1);
		glPixelTransferf(GL_BLUE_SCALE, 1);
		glPixelTransferf(GL_RED_BIAS, 0);
		glPixelTransferf(GL_GREEN_BIAS, 0);
		glPixelTransferf(GL_BLUE_BIAS, 0);


		//Such h4xx!
		// We just did (in paint terms) a flip-vertical
		IMG_RGB *tmp_rgb;
		IMG_RGB tmp;
		tmp_rgb = (IMG_RGB*)pImage->pixels;
		for (int i=0; i<(int)(dimy/2); i++)
		{
			for (int j=0; j<dimx; j++)
			{
				int c1 = i*dimx + j;
				int c2 = (dimy-i-1)*dimx + j;

				tmp.r = tmp_rgb[c1].r;
				tmp.g = tmp_rgb[c1].g;
				tmp.b = tmp_rgb[c1].b;

				//Note the flippage of rgb to bgr here.  Yeah, it's so amazing.
				tmp_rgb[c1].r = tmp_rgb[c2].b;
				tmp_rgb[c1].g = tmp_rgb[c2].g;
				tmp_rgb[c1].b = tmp_rgb[c2].r;
				tmp_rgb[c2].r = tmp.b;
				tmp_rgb[c2].g = tmp.g;
				tmp_rgb[c2].b = tmp.r;
			}
		}

		glTexImage2D(GL_TEXTURE_2D, 0, internal, pImage->w, pImage->h, 0, GL_RGB, GL_UNSIGNED_BYTE, pImage->pixels);

		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,mag_filter);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, wrap_param);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, wrap_param);

		SDL_FreeSurface(pImage);
//		free(pImage->data);
//		free(pImage);
	}

//	err_log << "texture: " << texture  << endl;
	if (texture >=0 && texture<MAX_TEXTURES)
	{
		texture_data[texture].active = true;
		texture_data[texture].width = dimx;
		texture_data[texture].height = dimy;
	}
	set_transparent(texture, x, y, z, dimx, dimy);

	err_log << "Texture [" << texture << "] built from ["<< filename <<"]"<< endl;
	return texture;													// Return The Status
}

void DISPLAY::set_transparent(int tex_num, float x, float y, float z, int dimx, int dimy)
{
	int i,n;
	int r,g,b,a;
	int bignum; //272144;
	float *image;

	glBindTexture(GL_TEXTURE_2D, tex_num);
	//Get texture dimensions, set to dimx, dimy
	//*OR* just see about putting this function IN init_texture(etc)

//	dimx = dimy = dim;
	bignum = dimx * dimy * 4 + 10000;

	image = (float*)calloc(bignum, sizeof(float));

//	use_transparency = true;

	glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_FLOAT, image);

	n = 0;
	for (i=0; i<bignum; i+=4)
	{
		r = i+0;
		g = i+1;
		b = i+2;
		a = i+3;

		if (image[r]==x && image[g]==y && image[b]==z)
			image[a] = 0, n++;
	}
/*	n=0;
	for (i = 0; i <= dimx; i++)
	{
		for (j = 0; j <= dimy; j++)
		{
//			for (k = 0; k < 4; k++)
//			{
				r = i*((dimx-1)*4)+(j*4)+0;
				g = i*((dimx-1)*4)+(j*4)+1;
				b = i*((dimx-1)*4)+(j*4)+2;
				a = i*((dimx-1)*4)+(j*4)+3;

				if (image[r]==x && image[g]==y && image[b]==z)
					image[a] = 0, n++;
//			}
		}
	}
*/
	glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, dimx, dimy, GL_RGBA, GL_FLOAT, image);
//	const unsigned char *str = gluErrorString(glGetError());
//	err_log << "ERROR: " << str << endl;

	free(image);
}

void DISPLAY::init_textures()
{
	//Create procedural texture as default texture.  If texture creation fails for others,
	// (if the file doesn't exist), then it'll return the index of this default texture.

	textures->set_default_texture();

	textures->gui[0] = init_texture("data/console_background.bmp", GL_REPEAT);
	textures->gui[1] = init_texture("data/text_box1.bmp");
	textures->gui[2] = init_texture("data/text_box2.bmp");
	textures->gui[3] = init_texture("data/text_box3.bmp");
	textures->gui[4] = init_texture("data/status_bar.bmp", GL_REPEAT);
	textures->gui[5] = init_texture("data/blank.bmp");


}

void TEXTURES::set_default_texture()
{
	int i;

	//This just needs to be made procedural (for another day!)
	def = frame.display.init_texture("data/default.bmp");

	for (i=0; i<D_MAX_GUI; i++)			gui[i] = def;
}

//Assuming naming scheme: ss#####.bmp
void DISPLAY::get_last_screen_shot()
{
	err_log << "Calculating last screen shot id..." << endl;

	int k;
	int padding = 5;
	char ext[5] = ".bmp";
	char *dir;
	char word[1024];
	char number[16];	//must be less than padding, no sense in making this dynamic though
	dir = read_dir("../Screens/Fall 06", ext);


	int n;
	int last = -1;
	int j = 0;
	for (int i=0; ; i++)
	{
		if (dir[i]=='\n' || dir[i]==0)
		{
			j -= 4;			//extension
			j -= padding;	//padding

			for (k=0; k<padding; k++)
			{
				number[k] = word[j];
				j++;
			}
			number[k] = 0;
			err_log << "NUMBER: " << number << endl;
			n = atoi(number);
			if (n > last) last = n;

			if (dir[i]==0) break;
			j = 0;
		}
		else
		{
			word[j] = dir[i];
			j++;
		}
	}

	last_screen_shot = last;

	err_log << "LAST: " << last_screen_shot << endl;
	err_log << "DIR: " << dir << endl;
}


//A simple screen shot function.  It's not very good.  It doesn't seem to use the latest buffer.
// There's weird colors appearing on some polygons.  What the heck.  Good job doug.  It kinda works, I guess.
void DISPLAY::screen_shot()
{
	err_log << "Saving screen shot..." << endl;
	if (last_screen_shot<0)
		get_last_screen_shot();

	int this_screen = last_screen_shot + 1;


	Uint32 rmask, gmask, bmask, amask;
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
    rmask = 0xff0000;
    gmask = 0x00ff00;
    bmask = 0x0000ff;
    amask = 0x000000;
#else
    rmask = 0x0000ff;
    gmask = 0x00ff00;
    bmask = 0xff0000;
    amask = 0x000000;
#endif
	amask = 0;


	SDL_Surface *surface;


	struct temp
	{
		unsigned int r:8;
		unsigned int g:8;
		unsigned int b:8;
	};

	temp *p;
	temp *pt;
	float *pixels;
	float *pixels2;

	p = (temp*)calloc(screen_x*screen_y,sizeof(temp));
	pixels = (float*)calloc(screen_x*screen_y*3,sizeof(float));
	pixels2 = (float*)calloc(screen_x*screen_y*3,sizeof(float));
	glReadBuffer(GL_LEFT);
	glReadPixels(0,0,screen_x,screen_y,GL_RGB,GL_FLOAT,pixels2);

	//Flip vertically, ugh
	int T0, T1, J;
	for (int j=0; j<screen_y; j++)
	{
		J = screen_y - j - 1;
		for (int i=0; i<screen_x*3; i+=3)
		{
			T0 = j*screen_x*3 + i;
			T1 = J*screen_x*3 + i;
			pixels[T1] = pixels2[T0];
			pixels[T1+1] = pixels2[T0+1];
			pixels[T1+2] = pixels2[T0+2];
		}
	}

	//Transfer pixels to image structure
	pt = p;
	for (int i=0; i<screen_x*screen_y*3; )
	{
		pt->r = (int)(pixels[i] * 256);		i++;
		pt->g = (int)(pixels[i] * 256);		i++;
		pt->b = (int)(pixels[i] * 256);		i++;
		pt++;
	}


//	screen = SDL_GetVideoSurface();
	
//	SDL_BlitSurface(screen, NULL, surface, NULL);

//	surface = SDL_CreateRGBSurfaceFrom(void *pixels, int width, int height, int depth, int pitch, Uint32 Rmask, Uint32 Gmask, Uint32 Bmask, Uint32 Amask);
//	surface = SDL_CreateRGBSurfaceFrom(p, screen_x, screen_y, 24, screen_x*screen_y, rmask, gmask, bmask, amask);





	surface = SDL_CreateRGBSurface(SDL_SWSURFACE, screen_x, screen_y, 32, rmask, gmask, bmask, amask);
	surface->pixels = p;




	char filename[256];
	char to[256];
	sprintf(filename, "ss%05i.bmp", this_screen);
	sprintf(to,"../Screens/Fall 06/%s", filename);
	last_screen_shot = this_screen;

	SDL_SaveBMP(surface, to);

	//I cannot believe I forgot to add these initially.  
	free(p);
	free(pixels);
	free(pixels2);

	frame.messages->add("Screen shot saved to %s",filename);
}




