//engine.cpp

#include "main.h"

extern FRAME frame;
extern LOGGER err_log;

ENGINE::ENGINE()
{
}

float ENGINE::get_time()
{
#ifdef BUILD_WINDOWS
    static __int64 start = 0;
    static __int64 frequency = 0;

    if (start==0)
    {
        QueryPerformanceCounter((LARGE_INTEGER*)&start);
        QueryPerformanceFrequency((LARGE_INTEGER*)&frequency);
        return 0.0f;
    }

    __int64 counter = 0;
    QueryPerformanceCounter((LARGE_INTEGER*)&counter);
    return (float) ((counter - start) / double(frequency));
#else
	//This is crap!  Better than nothing for now though.  Runs slow-motion on some computers.
	// At least it'll compile and run for testing on Linux.
	return (float)(SDL_GetTicks()/1000.0);
#endif


}

double ENGINE::get_time_factor()	{ return time.factor; }
float ENGINE::get_time_difference()	{ return time.difference; }

double ENGINE::get_super_time_factor()		{ return super_time.factor; }
float ENGINE::get_super_time_difference()	{ return super_time.difference; }


int ENGINE::init()
{
	to_list = 0;
	state = 0;
	set_state(ES_ACTIVE);
	set_state(ES_INIT);
//	gravity = true;
	pause = false;

	time.factor_denom = 17;
	time.fps = 60;
	time.fps_count = 0;
	time.fps_running = 0;

	super_time.factor_denom = 17;

	skips_i = 0;
	skips_working = 0;
	skips = 0;
	skips_max = 64;

	return 1;
}

int ENGINE::get_state() { return state; }
bool ENGINE::get_pause() { return pause; }
int ENGINE::get_skips() { return skips; }

//float ENGINE::get_time() { return time.current; }


void ENGINE::set_state(int this_state)
{
	state |= this_state;
}
void ENGINE::clear_state(int this_state)
{
	state &= ~this_state;
}
void ENGINE::toggle_state(int this_state)
{
	if (check_state(this_state)) clear_state(this_state);
	else set_state(this_state);
}
bool ENGINE::check_state(int this_state)
{
	return ((state & this_state) != 0);
}

void ENGINE::toggle_gravity()
{
	if (gravity)	gravity = false;
	else			gravity = true;
}
bool ENGINE::get_gravity()
{
	return gravity;
}
void ENGINE::toggle_pause()
{
	err_log << "TOGGLE PAUSE" << endl;
	if (pause)	pause = false;
	else		pause = true;
}
void ENGINE::set_pause(bool new_pause) { pause = new_pause; }


void ENGINE::calc_fps()
{
	if (!frame.config.get_bool(VAR_DRAW_FPS)) return;

	if (time.fps_count>=FPS_COUNT_MAX)
	{
//		time.fps = time.fps_running / FPS_COUNT_MAX;
//		time.fps_running = 0;
		time.fps_count = 0;
		time.fps = 1.0 / time.difference;
	}
//	time.fps_running += 1.0 / time.difference;
	time.fps_count+=time.difference;
}
float ENGINE::get_fps() {return time.fps;}


void ENGINE::set_time_factor()
{
//	set_time();
	time.t1 = time.t2;
	time.t2 = get_time();
	time.difference = time.t2-time.t1;
	time.total += time.difference;

	super_time.difference = time.difference;
	super_time.total += time.difference;
	super_time.factor = super_time.difference;

	calc_fps();

	if (pause)	time.difference = 0;
	else
	{
/*
		if (time.difference==0)	skips_working++;
		else					skips_i++;
		if (skips_i==skips_max)
		{
			skips = skips_working;
			err_log << "                      Skips: " << skips << "    FPS: " << time.fps << endl;
			skips_working = 0;
			skips_i = 0;
		}
*/
	}
	time.factor = time.difference;



//	err_log << "TIME DIFF: " << time.difference << endl;

/*
	if (to_list >= 128) to_list = 0;
	factor_list[to_list] = time.factor;
	to_list++;
*/
//	time.factor /= 17;

//	if (time.total % 2 == 0)
//		calc_fps();
}



