//engine.h

#ifndef _ENGINE_
#define _ENGINE_

#include "main.h"

#define ES_INIT		1
#define ES_PAUSE	2
#define ES_QUIT		4
#define ES_ACTIVE	8


#define FPS_COUNT_MAX 0.25f


class ENGINE;

struct TIME
{
	float t1;				//Previous time
	float t2;				//Current time
	double total;			//Running total time
	float difference;		//Difference between t1 and t2

	double factor;			//Factor of increment
	double factor_denom;	//Denominator - probably not needed, could be better without

	float fps;
	float fps_running;
	float fps_count;
};

class ENGINE
{
public:
	ENGINE();
	int init();

	int get_state();

	void set_time_factor();
	bool get_gravity();
	double get_time_factor();
	float get_time_difference();
	bool get_pause();
	float get_fps();
	int get_skips();

	double get_super_time_factor();
	float get_super_time_difference();


	bool check_state(int this_state);
	void set_state(int this_state);
	void clear_state(int this_state);
	void toggle_state(int this_state);

	float get_time();

	void calc_fps();

	void toggle_gravity();
	void toggle_pause();
	void set_pause(bool new_pause);

protected:

	int skips_working;
	int skips;
	int skips_i;
	int skips_max;

	int state;
	TIME time;
	TIME super_time;
	bool gravity;
	bool pause;

	double factor_list[128];
	int to_list;
};

#endif

