//font.cpp

#include "main.h"

extern FRAME frame;
extern LOGGER err_log;
extern double rnd();

FONT::FONT()
{
	max_height = 0;
	spacing_x = 1;
	spacing_y = 1;
	tex = -1;
	drop = 0;
}
FONT::~FONT()
{
	GLuint textures[1] = {tex};
	glDeleteTextures(1, textures);
}

FONTS::FONTS()
{
	color.r = 1;
	color.g = 1;
	color.b = 1;
	color.a = 1;
}

//Build a set of fonts
void FONTS::build_fonts()
{
	err_log << "Building fonts..." << endl;
	face[0].set_face("data/font1.bmp");
	face[1].set_face("data/font2.bmp",0,1,0);
	face[2].set_face("data/font3.bmp");
	face[3].set_face("data/font4.bmp",3);
}


//Used when building the font
char character_set[MAX_CHARACTERS] = {'a','b','c','d','e','f','g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y','z', '0','1','2','3','4','5','6','7','8','9','!','?','.',',',':',' ', 'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z'};


//Build a given font, given a filename and the drop height of each char
void FONT::set_face(char *filename, int new_drop)
{
	set_face(filename,1,0,0);
	drop = new_drop;
}
void FONT::set_face(char *filename)
{
	set_face(filename,1,0,0);
}
//dr,dg,db are the delimiter color.  This is used to determine what parts of the file
// are part of the font, and which parts are spacers.  The default is pure red, but
// if a font wants to use pure red, you can specify another color (but be sure to
// modify the background delimiting color for the file appropriately!)
void FONT::set_face(char *filename, float dr, float dg, float db)
{
	err_log << "Building font from ["<< filename <<"]... " << endl;

	int J,i,j,n;
	int bignum; //272144;
	float *image;
	int dimx, dimy;
	int si,sj,ei,ej;
	int w, h;
	int r,g,b,a;
	int row_max_height;

	IMG_RGBAf **img;

	IMG_RGBf delim;
	delim.r = dr;
	delim.g = dg;
	delim.b = db;

	tex = frame.display.init_texture(filename);
	dimx = frame.display.texture_data[tex].width;
	dimy = frame.display.texture_data[tex].height;

	img = dynarray<IMG_RGBAf>(dimx,dimy);

	glBindTexture(GL_TEXTURE_2D, tex);
	bignum = dimx * dimy * 4 + 10000;
	image = (float*)calloc(bignum, sizeof(float));
	glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_FLOAT, image);


	for (i = 0; i < dimx; i++)
	{
		for (j = 0; j < dimy; j++)
		{
			r = j*dimy*4 + i*4 + 0;
			g = j*dimy*4 + i*4 + 1;
			b = j*dimy*4 + i*4 + 2;
			a = j*dimy*4 + i*4 + 3;

			J = dimy-j-1;
			img[i][J].r = image[r];
			img[i][J].g = image[g];
			img[i][J].b = image[b];
			img[i][J].a = image[a];
		}
	}

	max_height = 0;
	max_width = 0;

	i=0;
	j=0;
	n = 32;
	for (; j<dimy;)
	{
		i = 0;
		//Move indices over to first non-delimiter color
		while (j<dimy)
		{
			if ( !(img[i][j].r==delim.r && img[i][j].g==delim.g && img[i][j].b==delim.b) ) break;
			i++;
			if (i>=dimx) { j++; i=0; }
		}

		row_max_height = 0;
		w = 0;
		for (; i<dimx; i++)
		{
			if (img[i][j].r==delim.r && img[i][j].g==delim.g && img[i][j].b==delim.b)
			{
				i--;
				ei = i + 1;
				ej = dimy - j;

				h = 0;
				for (;j<dimy;j++) if (img[i][j].r==delim.r && img[i][j].g==delim.g && img[i][j].b==delim.b) break; else h++;
				j--;

				si = i - (w - 1);
				sj = dimy - j - 1;

				c[n].h = h;
				c[n].w = w;
				c[n].u0 = (float)si/(float)dimx;
				c[n].v0 = (float)sj/(float)dimy;
				c[n].u1 = (float)ei/(float)dimx;
				c[n].v1 = (float)ej/(float)dimy;
				c[n].c = character_set[n];

				if (h > max_height) max_height = h;
				if (w > max_width) max_width = w;
				if (h > row_max_height) row_max_height = h;

				n++;		//Next character
				j -= h - 1;	//Knock j back to top of item
				i++;
				
				//Run until we hit the edge, or find a pixel that's not the delim
				while (i<dimx && (img[i][j].r==delim.r && img[i][j].g==delim.g && img[i][j].b==delim.b)) i++;
				i--;
				w = -1;		//Reset width (will be ++'ed after, so -1)

			}
			w++;
		}

		//Now we want to jump down to next row:
		i = 0;
		j += row_max_height;
//		while (j<dimy && !(img[i][j].r==delim.r && img[i][j].g==delim.g && img[i][j].b==delim.b)) j++;
//		while (j<dimy && (img[i][j].r==delim.r && img[i][j].g==delim.g && img[i][j].b==delim.b)) j++;
	}

	err_log << n-32 << " characters built." << endl;
	if (n-32 != 95)
		err_log << "WARNING: Font has abnormal character number!" << endl;

	//No need to have delimiting color anymore!
	frame.display.set_transparent(tex, delim.r, delim.g, delim.b, dimx, dimy);

	free(image);


	dynfree<IMG_RGBAf>(img,dimx,dimy);
}


void FONTS::set_color(float r, float g, float b, float a)
{
	color.r = r;
	color.g = g;
	color.b = b;
	color.a = a;
}

void FONT::set_drop(int new_drop)
{
	drop = new_drop;
}

int FONTS::get_max_height(int font)	{ return face[font].max_height; }
int FONTS::get_max_width(int font)	{ return face[font].max_width; }
int FONTS::get_spacing_x(int font)	{ return face[font].spacing_x; }
int FONTS::get_spacing_y(int font)	{ return face[font].spacing_y; }
int FONTS::get_width(int font, char this_c)
{
	return face[font].c[(int)this_c].w;
}


//Draw text with default font
void FONTS::draw(char *text, int x, int y)
{
	face[FF_DEFAULT].draw(text,x,y);
}
//Draw text with a given face and text
void FONTS::draw(int use_face, char *text, int x, int y)
{
	face[use_face].draw(text,x,y);
}
//Specify a color to draw at
void FONTS::draw(int use_face, char *text, int x, int y, float r, float g, float b, float a)
{
	glColor4f(r,g,b,a);
	face[use_face].draw(text,x,y);
}
//Draw the text at the given screen coordinate (lower-left corner of text aligned)
void FONT::draw(char *text, int x, int y)
{
	y -= drop;
	int sx = x;
	FONT_CHARACTER *current;
	int i;

	frame.magic.try_begin(tex);

	for (i=0; text[i]!=0; i++)
	{
		if (text[i]=='\n')
		{
			x = sx;
			y -= max_height + spacing_y;
			continue;
		}
		current = &c[text[i]];

		glTexCoord2f(current->u0,current->v0);	glVertex2i(x,				y);
		glTexCoord2f(current->u0,current->v1);	glVertex2i(x,				current->h+y);
		glTexCoord2f(current->u1,current->v1);	glVertex2i(x+current->w,	current->h+y);
		glTexCoord2f(current->u1,current->v0);	glVertex2i(x+current->w,	y);

		x += current->w + spacing_x;
	}
}


//Let's see about FLOAT TEXT.  This is excrutiatingly slow, because it's applying
// all kinds of matrix calculations for EACH BIT OF TEXT.  But this is just
// for debugging stuff, so it's ok.
void FONTS::draw(int use_face, char *text, float x, float y, float z)
{
	double X[1],Y[1],Z[1];
	int length=0;

	//Convert world coords into screen coords.  This is not blazing speed.
	glGetDoublev(GL_MODELVIEW_MATRIX, frame.display.modelMatrix);
	glGetDoublev(GL_PROJECTION_MATRIX, frame.display.projMatrix);
	glGetIntegerv(GL_VIEWPORT, frame.display.viewport);
	gluProject(x,y,z, frame.display.modelMatrix, frame.display.projMatrix, frame.display.viewport, X,Y,Z);

	if (*Z>1)
		return;

	int ix, iy;

	ix = (int)(*X);
	iy = (int)(*Y);

	frame.camera->set_overlay_matrices();
	glCullFace(GL_FRONT);

	glEnable(GL_TEXTURE_2D);
	glColor4f(1,1,1,1);
	face[use_face].draw(text,ix,iy);

	frame.camera->set_matrices();
}



