//frame.cpp

#include "main.h"

extern FRAME frame;
extern LOGGER err_log;
extern LOGGER errors;
extern double rnd();
extern double rnd(float range);


int error_count = 0;


FRAME::FRAME()
{
	initialized = false;
	done=false;
	current_model = -1;
	num_models = 0;
	absolute_longest_edge = 0;
	bins = NULL;
//	error_index = -1;
}
FRAME::~FRAME()
{
	if (!initialized) return;

	delete control;

	err_log.close();
	errors.close();
}
void FRAME::gl_errors()
{
	int error_val = glGetError();
	if (error_val==0)
	{
		return;
	}
	errors << "GL Error for [" << error_val << "]: " << gluErrorString(error_val) << endl;
}

int FRAME::init()
{
	srand(RANDOM_SEED);

	err_log.always_active = true;
	err_log.open("debug_log.txt");
	err_log << "Initializing FRAME..." << endl;

	errors.always_active = true;
	errors.open("errors.txt");
	errors << "Errors & Warnings:" << endl;
	errors << "------------------------" << endl;


	config.init();
	control = new CONTROL;

	if (engine.init()<=0)		fail("Error initializing engine.");

	photon_list.create(100000);

	const SDL_VideoInfo* info = NULL;
	int width = 0;
	int height = 0;
	int bpp = 0;
	int flags = 0;

	if( SDL_Init( SDL_INIT_VIDEO | SDL_INIT_JOYSTICK ) < 0 )
	{
		fail("Video initialization failed: %i",SDL_GetError());
//		err_log << "Video initialization failed: " << SDL_GetError() << endl;
//		quit(-1);
	}
	info = SDL_GetVideoInfo( );
	if (control->init()<=0)		fail("Error initializing control.");

	width = display.screen_x;
	height = display.screen_y;
	bpp = info->vfmt->BitsPerPixel;

	SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
	SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

	if (display.is_fullscreen())
		flags = SDL_OPENGL | SDL_FULLSCREEN;
	else
		flags = SDL_OPENGL;

	SDL_WM_SetCaption("Weathering", NULL);
	SDL_WM_SetIcon(SDL_LoadBMP("data/icon.bmp"), NULL);

	screen = SDL_SetVideoMode( width, height, bpp, flags );
	if( screen == 0 )
		fail("Video mode failed: %s", SDL_GetError());
	else
	{
		err_log << "Video initialized." << endl;
		err_log << " Resolution: " << width << " x " << height << endl;
		err_log << " Depth:      " << bpp << endl;
	}

//	err_log << "TOGGLE: " << SDL_WM_ToggleFullScreen(screen) << endl;

	if (display.init()<0)			fail("Error initializing display.");

	messages = new MESSAGES(FF_TIMES_SMALL);

	initialized = true;
	return 1;
}


bool FRAME::add_field(char *file)
{
	return vector_field->build_from_brandon(file);
}

bool FRAME::add_model(char *file, char *name)
{
	if (num_models>=MAX_MODELS)
	{
		error("Cannot add %s -- too many models.",file);
		return false;
	}
	err_log << "LOAD MODEL: " << endl;
	if (!model[num_models].load_ply(file,name))
	{
		frame.error("load_ply(%s) failed.",file);
		return false;
	}
	err_log << "DONE" << endl;
	num_models++;
	return true;
}

void FRAME::do_main_loop()
{
	err_log << "Entering main loop... " << endl;

    while(!done)
	{
		if (engine.check_state(ES_QUIT))
		{
			err_log << "Engine in quit state.  Quitting..." << endl;
			done = true;
		}
		else if (engine.check_state(ES_ACTIVE))
		{
			if (engine.check_state(ES_INIT))
			{

				cache_misses = 0;
				cache_hits = 0;

				err_log << "Engine in initialization state.  Initializing..." << endl;

				console.init();

				//Random stuff!  I need to spread this out better
				glEnable(GL_BLEND);
				glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
				glEnable(GL_ALPHA_TEST);
				glAlphaFunc(GL_GREATER,0);


				glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
				engine.clear_state(ES_INIT);

				extents.set(-1,-1,-1,1,1,1);
//				extents.set(-.5f,-0.5f,-.5f,.5f,.5f,.5f);
//				extents.set(-.25f,-0.25f,-.25f,.25f,.25f,.25f);

				//These GL commands allow the calculating message to appear on load
				display.viewports.init();
				glClearColor(0,0,0,0);
				glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
/*				display.overlay->show_calculating("LOADING");
				glClearColor(display.background_color[0],display.background_color[1],display.background_color[2],display.background_color[3]);
				camera->set_overlay_matrices();
				glCullFace(GL_FRONT);
				display.overlay->show_calculating("Loading...");
*/

				glClearColor(display.background_color[0],display.background_color[1],display.background_color[2],display.background_color[3]);


//				add_model("data/horse.ply","Horse");
//				add_model("data/bunny.ply","Bunny");
//				add_model("data/dragon.ply",false,"Dragon",0.5f);
//				add_model("data/hand.ply",false,"Hand",0.5f);
//				add_model("data/happy.ply",false,"Happy",0.5f);

//				add_model("data/face.ply","Face");
//				add_model("data/face2.ply","Face 2");
//				add_model("data/face3.ply",true,"Face 3",.75);
//				add_model("data/box.ply",true,"Box",.75);

				inc_model();	//This will select first model in list;
				display.extra->spawn_box.init();	//Visualization of spawn box set from this

				init_bins();



				//Temp vector field creation!
				vector_field = new VECTOR_FIELD();
				/*
				double o[3] = {-1,-1,-1};
				double w[3] = {.05,.05,.05};
				int dim[3] = {1000,1000,1};
				int input_length = dim[0]*dim[1]*dim[2]*3;
				double *input_field = (double*)calloc(input_length,sizeof(double));
				for (int Q=0; Q<input_length; Q+=3)
				{
//					input_field[Q] = rnd(1); input_field[Q+1] = rnd(1); input_field[Q+2] = rnd(1);
					input_field[Q] = 1; input_field[Q+1] = 0; input_field[Q+2] = 0;
				}
				*/
				//This reads in a vector field with the given parameters.
				// If you were to create the field otherwise, or read it in from
				// a file, you would do so just before this and give the parameters
				// from that here.
				//vector_field->read_field(input_field, o,w,dim);
//				vector_field->build_from_file();


				vector_field->build_from_brandon("data/vortex.vf");
//				vector_field->build_from_matlab();
//				err_log << "YES" << endl;

				ibfv.init(3);

				messages->add("Press F1 to view keylist");

				err_log << "Engine initialized. " << endl;

			}

			if (config.get_bool(VAR_DEBUG_STEP))
				engine.set_pause(true);

			control->handle_events();

			engine.set_time_factor();

			redraw();
		}
	}

	err_log << "Exiting main loop..." << endl;
}


void FRAME::redraw()
{
//	err_log << "Redraw..." << endl;
	float t1, t2, total, target;
	t1 = t2 = engine.get_time();
	int m;
	int vp;

	if (config.get_bool(VAR_RUN_SIMULATION))
		photon_list.step();


	//We want to see about FORCING NEW LISTs
	// for specific models/viewports, in addition to
	// forcing new change for ALL
	if (config.get_bool(VAR_FORCE_NEW_LIST))
	{
		display.extra->spawn_box.init();
	}

	for (vp=0; vp<4; vp++)
	{
		if (display.viewports.vp[vp].will_force_new_list())
		{
			for (m=0;m<MAX_MODELS;m++) model[m].create_list(vp);
//			config.set(VAR_FORCE_NEW_LIST,0);
		}
	}

	glEnable(GL_DEPTH_TEST);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	//Any stepping / simulation will go BEFORE the loops below

	GLfloat light_position[] = {-.1f,.1f,.1f,0};

	//This controls the views for each viewport, or just the active one
	// if you are in single viewport mode
	for (vp = 0; vp < 4; vp++)
	{
		if (config.get_bool(VAR_SINGLE_VIEW))
		{
			vp = control->active_viewport;
		}

		magic.try_end();
		display.viewports.vp[vp].draw();
		display.viewports.vp[vp].set_camera();

		display.viewports.vp[vp].set_matrices();
//		camera->set_matrices();


		glLightfv(GL_LIGHT0,GL_POSITION, light_position);


		magic.try_end();

		glDisable(GL_TEXTURE_2D);
		glColor4f(1,1,1,1);

		glCullFace(GL_BACK);

//		vertex_array.draw();

		display.viewports.vp[vp].draw_grid();
		vector_field->draw();

/*
		glColor4f(1,1,1,1);
		glBegin(GL_POLYGON);
			glVertex3f(0,0,.5f);
			glVertex3f(1,0,.5f);
			glVertex3f(1,1,.5f);
			glVertex3f(0,1,.5f);
		glEnd();
*/

		//Messy code to loop through all the models
		float i,j,k;
		float s = .25f;
		for (m=0;m<MAX_MODELS;m++)
		{
			if (m==0) {i=0;j=0;k=0;}
			if (m==1) {i=s;j=0;k=0;}
			if (m==2) {i=0;j=s;k=0;}
			if (m==3) {i=s;j=s;k=0;}
			if (m==4) {i=0;j=0;k=s;}
			if (m==5) {i=s;j=0;k=s;}
			if (m==6) {i=0;j=s;k=s;}
			if (m==7) {i=s;j=s;k=s;}

			glPushMatrix();

			model[m].draw(vp);

			glPopMatrix();
		}

		bins->draw();

		photon_list.draw();




		if (display.viewports.vp[vp].get_spawn_box())
			display.extra->spawn_box.draw();


		ibfv.draw_all();

		
		//	frame.display.fonts->draw(FF_TIMES_SMALL, "23.4234065932",0,0,0);
	frame.magic.try_end();

		//Maybe even move this after the next statement
		ibfv.generate_all();

		camera->set_overlay_matrices();
//		camera->draw_grid();


		glEnable(GL_TEXTURE_2D);
		glCullFace(GL_FRONT);


		//Loop only once if you're in single view mode
		if (config.get_bool(VAR_SINGLE_VIEW))
			break;
	}



	display.overlay->draw_global();

//	err_log << "CACHE_MISSES/HITS: " << cache_misses << ", " << cache_hits << endl;
	cache_misses = 0;
	cache_hits = 0;


	SDL_GL_SwapBuffers();

	//Frame limit, currently disabled
	if (config.slowdown) target = MIN_TIME_STEP * 8;	//,05
	else target = 0;//MIN_TIME_STEP;							//,01
	total = 0;
	while (true)
	{
		t1 = t2;
		t2 = engine.get_time();
		total += t2-t1;
		if (total>target)
			break;
	}
}

//Select the next model
void FRAME::inc_model()
{
	current_model++;
	int n = 0;
	while (!model[current_model].is_ready())
	{
		if (current_model>=MAX_MODELS)
		{
			//Cycled to end of list -- at this point, toggle selection OFF
			messages->add("Model deselected.");
			current_model = -1;
			return;
		}
		current_model++;
		n++;

		if (n>=MAX_MODELS)
		{
			//Cycled through each, and none active -- ignore!
			current_model = -1;
			return;
		}
	}
	messages->add("Model %i : %s selected.",current_model,model[current_model].name);
}

void FRAME::quit(char *reason, int code)
{
	errors << "EXITING WITH ERROR[" << code << "] = " << reason << endl;
	err_log << "EXITING WITH ERROR[" << code << "] = " << reason << endl;
	quit(code);
}
void FRAME::quit(int code)
{
	err_log << "Quitting." << endl;
	err_log.close();
	errors.close();
	SDL_Quit();
	delete control;
	exit(code);
}
void FRAME::finish()
{
	err_log << "Finishing..." << endl;
	done = true;
}

//Destroy stuff in frame (but not frame object)
void FRAME::destroy()
{
	delete control;
	delete messages;
	delete bins;
	display.destroy();
	photon_list.destroy();
	config.destroy();

	for (int i=0; i<MAX_MODELS; i++)
		model[i].destroy();


	SDL_Quit();


	err_log.close();
	errors.close();
}

void FRAME::fail(const char *fmt, ...)
{
	char reason[1024];
	va_list ap;

	for (int i=0; i<1024; i++) reason[i]=0;

	if (fmt != NULL)
	{
		va_start(ap, fmt);
		vsprintf(reason, fmt, ap);
		va_end(ap);
	}

	quit(reason, -1);
}

float FRAME::abs(float a)
{
	if (a>=0)	return a;
	return -a;
}

void FRAME::wait(float target)
{
	float t1,t2,total;

	t1 = t2 = engine.get_time();
	total = 0;
	while (true)
	{
		t1 = t2;
		t2 = engine.get_time();
		total += t2-t1;
		if (total>target)
			return;
	}
}

void FRAME::shout(const char *fmt, ...)
{
	char message[1024];
	va_list ap;
	
	for (int i=0; i<1024; i++) message[i]=0;

	if (fmt != NULL)
	{
		va_start(ap, fmt);
		vsprintf(message, fmt, ap);
		va_end(ap);
	}

	display.overlay->announce.set(message);
}

void FRAME::warning(const char *fmt, ...)
{
	char reason[1024];
	va_list ap;
	for (int i=0; i<1024; i++) reason[i]=0;
	if (fmt != NULL)
	{
		va_start(ap, fmt);
		vsprintf(reason, fmt, ap);
		va_end(ap);
	}
	err_log << "WARNING: " << reason << endl;
	errors << "WARNING: " << reason << endl;
}

void FRAME::error(const char *fmt, ...)
{
	error_count++;
	char reason[1024];
	va_list ap;
	for (int i=0; i<1024; i++) reason[i]=0;
	if (fmt != NULL)
	{
		va_start(ap, fmt);
		vsprintf(reason, fmt, ap);
		va_end(ap);
	}
	err_log << "(" << error_count << ") ERROR: " << reason << endl;
	errors << "(" << error_count << ") ERROR: " << reason << endl;
}

void FRAME::pvector(char *label, float v[3])
{
	err_log << label << ": " << v[0] << ", " << v[1] << ", " << v[2] << endl;
}
void FRAME::pvector(char *label, double v[3])
{
	err_log << label << ": " << v[0] << ", " << v[1] << ", " << v[2] << endl;
}




void FRAME::mark()							{ err_log << "!" << endl; }
void FRAME::mark(char *what)				{ err_log << what << endl; }
void FRAME::markf(char *title, float value)	{ err_log << title << ": " << value << endl; }
void FRAME::marki(char *title, int value)	{ err_log << title << ": " << value << endl; }
void FRAME::marks(char *title, char *value)	{ err_log << title << ": " << value << endl; }


void FRAME::delete_list(int *display_list)
{
	if (*display_list>0)
		glDeleteLists(*display_list,1);
	*display_list = -1;
}
void FRAME::create_list(int *display_list)
{
	if (*display_list>0)
		delete_list(display_list);

	*display_list = glGenLists(1);
	glNewList(*display_list,GL_COMPILE);
}


void FRAME::init_bins()
{
	int i;

	if (bins) delete bins;


	if (config.get_bool(VAR_GIANT_BIN)) 
	{
		//If in GIANT BIN mode, then bin_width was set in there
		bins = new BIN_LIST(bin_width,bin_width,bin_width);
	}
	else
	{
		//Analyze models to find bin width (from largest edge!)
		bin_width = absolute_longest_edge;
//		bin_width = .0048f;	//.0068
		bins = new BIN_LIST(bin_width,bin_width,bin_width);
	}


	for (i = 0; i<MAX_MODELS; i++)
	{
		model[i].init_bins();
	}
}



void FRAME::bin_down(bool new_lists)
{
	//2.0 implies bins will be twice the size of the smallest subdivision
	float w_min = config.get_float(VAR_SUBDIV_TARGET) * 3.0;
	float w = bins->get_wx();

	if (config.get_bool(VAR_INCREMENTAL_SUBDIV))
		w *= .6f;
	else
		w = w_min;

	if (w<w_min) w = w_min;
	delete bins;
	bins = new BIN_LIST(w,w,w);

	if (new_lists)
		for (int j=0; j<MAX_MODELS; j++)
			model[j].init_bins();

}




