//graphic_box.cpp

#include "main.h"

extern FRAME frame;
extern LOGGER err_log;

GRAPHIC_BOX::GRAPHIC_BOX()
{
	x = 10;
	y = 10;
	width = 50;
	height = 50;

	color.r = 1;
	color.g = 1;
	color.b = 1;
	color.a = 1;

	color_background.r = 1;
	color_background.g = 1;
	color_background.b = 1;
	color_background.a = 1;

	type = GBT_SOLID;

	passive_alpha = true;
}


void GRAPHIC_BOX::set_passive_alpha(bool new_passive_alpha) { passive_alpha = new_passive_alpha; }
void GRAPHIC_BOX::set_dims(int new_width, int new_height)
{
	width = new_width;
	height = new_height;
}
void GRAPHIC_BOX::set_color(float r, float g, float b, float a)
{ color.r = r; color.g = g; color.b = b; color.a = a; }
void GRAPHIC_BOX::set_color_background(float r, float g, float b, float a)
{ color_background.r = r; color_background.g = g; color_background.b = b; color_background.a = a; }

void GRAPHIC_BOX::set_type(int new_type) { set_type(new_type,0); }
void GRAPHIC_BOX::set_type(int new_type, int new_tex_background)
{
	type = new_type;
//	err_log << "TYPE: " << type << endl;
	if (type==GBT_TEXTURED)
	{
		tex_background = new_tex_background;
		set_textured_background_uv();
	}
}
void GRAPHIC_BOX::set_height(int new_height)	{ height = new_height; }
int GRAPHIC_BOX::get_height()					{ return height; }

// You know that you love this function name.
void GRAPHIC_BOX::set_textured_background_uv()
{
	if (tex_background==0 || !frame.display.texture_data[tex_background].active)
	{
		err_log << "Error: Trying to set invalid texture to graphic_box background: " << tex_background << endl;
		return;
	}

	err_log << "WIDTH:  " << width<<endl;
	err_log << "HEIGHT: " << height<<endl;

	u[0][0] = 0;
	u[0][1] = (float)width / (float)frame.display.texture_data[tex_background].width;
	v[0][0] = 0;
	v[0][1] = (float)height / (float)frame.display.texture_data[tex_background].height;
}

/*
void GRAPHIC_BOX::set_pos(int nx, int ny, int new_width, int new_height)
{
	width = new_width;
	height = new_height;
	set_pos(nx,ny);
}
*/
void GRAPHIC_BOX::set_pos(int nx, int ny)
{
	x = nx;
	y = ny;

	//middle
	vx[0][0] = x + w;
	vx[0][1] = x + width - w;
	vy[0][0] = y - w;
	vy[0][1] = y - height + w;

	//left
	vx[1][0] = x;
	vx[1][1] = x + w;
	vy[1][0] = y - w;
	vy[1][1] = y - height + w;

	//right
	vx[2][0] = (x + width) - w;
	vx[2][1] = (x + width);
	vy[2][0] = y - w;
	vy[2][1] = y - height + w;

	//top
	vx[3][0] = x + w;
	vx[3][1] = x + width - w;
	vy[3][0] = y - w;
	vy[3][1] = y;

	//bottom
	vx[4][0] = x + w;
	vx[4][1] = x + width - w;
	vy[4][0] = (y - height);
	vy[4][1] = (y - height) + w;

	//TL
	vx[5][0] = x;
	vx[5][1] = x + w;
	vy[5][0] = (y - height);
	vy[5][1] = (y - height) + w;

	//BL
	vx[6][0] = x;
	vx[6][1] = x + w;
	vy[6][0] = y - w;
	vy[6][1] = y;

	//BR
	vx[7][0] = (x + width) - w;
	vx[7][1] = (x + width);
	vy[7][0] = y - w;
	vy[7][1] = y;

	//TR
	vx[8][0] = (x + width) - w;
	vx[8][1] = (x + width);
	vy[8][0] = (y - height);
	vy[8][1] = (y - height) + w;

}

void GRAPHIC_BOX::set_tex(int new_tex, int nw)
{
	tex = new_tex;
	w = nw;

  //This is just too ugly.  I will just have to generalize the graphicbox type format, and
  // have it work for anything.  Text_box'es will just have a graphicbox in them, which
  // will handle drawing this outline.

	//Code to draw text box..  may be good to generalize
	// and allow for ANY box type
	float d = 8;	//dimensions of textbox texture

	//If middle is going to be just a color, leave this as is, so we're not swapping
	// textures at all.
	//However, if we want to use some textured background, then you will need to
	// modify the uv and looping parameters and such.  Not worth it?

	if (type==GBT_TEXTURED) set_textured_background_uv();
	else
	{
		//middle
		u[0][0] = w / d;
		u[0][1] = (w + 2.0) / d;
		v[0][0] = w / d;
		v[0][1] = (w + 2.0) / d;
	}


	//texture coords for left side
	u[1][0] = 0;
	u[1][1] = w / d;
	v[1][0] = w / d;
	v[1][1] = (w + 2.0) / d;

	//right side
	u[2][0] = (d - w) / d;
	u[2][1] = 1;
	v[2][0] = w / d;
	v[2][1] = (w + 2.0) / d;

	//bottom
	u[3][0] = w / d;
	u[3][1] = (w + 2.0) / d;
	v[3][0] = (d - w) / d;
	v[3][1] = 1;

	//top
	u[4][0] = w / d;
	u[4][1] = (w + 2.0) / d;
	v[4][0] = 0;
	v[4][1] = w / d;

	//bottom-left
	u[5][0] = 0;
	u[5][1] = w / d;
	v[5][0] = 0;
	v[5][1] = w / d;

	//top-left
	u[6][0] = 0;
	u[6][1] = w / d;
	v[6][0] = (d - w) / d;
	v[6][1] = 1;

	//top-right
	u[7][0] = (d - w) / d;
	u[7][1] = 1;
	v[7][0] = (d - w) / d;
	v[7][1] = 1;

	//bottom-right
	u[8][0] = (d - w) / d;
	u[8][1] = 1;
	v[8][0] = 0;
	v[8][1] = w / d;
}


void GRAPHIC_BOX::draw()
{
	int i;

//	err_log << "GB POS: " << x << ", " << y << endl;


	frame.magic.try_end();
	glDisable(GL_CULL_FACE);


	if (type!=GBT_CLEAR)
	{
		if (type==GBT_SOLID)			frame.magic.try_begin(tex);
		else if (type==GBT_TEXTURED)	frame.magic.try_begin(tex_background);

		if (passive_alpha)	frame.magic.try_color(color_background.r,color_background.g,color_background.b);
		else				glColor4f(color.r,color.g,color.b,color.a);

		glTexCoord2f(u[0][0],v[0][0]); glVertex3f(vx[0][0], vy[0][0], 11);
		glTexCoord2f(u[0][0],v[0][1]); glVertex3f(vx[0][0], vy[0][1], 11);
		glTexCoord2f(u[0][1],v[0][1]); glVertex3f(vx[0][1], vy[0][1], 11);
		glTexCoord2f(u[0][1],v[0][0]); glVertex3f(vx[0][1], vy[0][0], 11);
	}


	frame.magic.try_begin(tex);
	for (i=1; i<9; i++)
	{
		if (passive_alpha)	frame.magic.try_color(color.r,color.g,color.b);
		else				glColor4f(color.r,color.g,color.b,color.a);
		glTexCoord2f(u[i][0],v[i][0]); glVertex3f(vx[i][0], vy[i][0], 11);
		glTexCoord2f(u[i][0],v[i][1]); glVertex3f(vx[i][0], vy[i][1], 11);
		glTexCoord2f(u[i][1],v[i][1]); glVertex3f(vx[i][1], vy[i][1], 11);
		glTexCoord2f(u[i][1],v[i][0]); glVertex3f(vx[i][1], vy[i][0], 11);
	}

//	glTexCoord2f(0,0);		glVertex3f(x,			y,			11);
//	glTexCoord2f(0,1);		glVertex3f(x,			y+height,	11);
//	glTexCoord2f(1,1);		glVertex3f(x+width,		y+height,	11);
//	glTexCoord2f(1,0);		glVertex3f(x+width,		y,			11);

}



