//grid.cpp

#include "main.h"

extern FRAME frame;
extern LOGGER err_log;

GRID::GRID()
{
	ex = .5f;
	ey = .5f;
	ez = .5f;
	spacing_x = .02f;
	spacing_y = .02f;
	spacing_z = .02f;
	which = -1;
	display_list = -1;
}

void GRID::cycle_which()
{
	which++;
	if (which>2) which = -1;
	frame.delete_list(&display_list);
}

void GRID::draw()
{
	draw_grid();
	draw_extents();
	draw_axes();
}

void GRID::draw_grid()
{
	if (which<0)
		return;

	if (display_list<=0)
		create_grid();

	glCallList(display_list);
}

void GRID::create_grid()
{
	err_log << "CREATING GRID: " << which << endl;
	float x,y,z;

	float x0,y0,z0,x1,y1,z1;
	x0 = frame.extents.x0;y0 = frame.extents.y0;z0 = frame.extents.z0;
	x1 = frame.extents.x1;y1 = frame.extents.y1;z1 = frame.extents.z1;

	frame.create_list(&display_list);

//	glDisable(GL_DEPTH_TEST);
//	glDepthMask(false);
	glEnable(GL_FOG);
//	glFogf(GL_FOG_DENSITY,25);


	glFogi(GL_FOG_MODE,GL_LINEAR);
	glFogf(GL_FOG_START,0);
	glFogf(GL_FOG_END,.10f);
//	glFogfv(GL_FOG_COLOR,frame.display.background_color);

	glBegin(GL_LINES);

	switch (which)
	{
	default: break;
	case 0:
		{
			glColor4f(0,1,0,1);
			for (z=z0;z<=z1;z+=spacing_z)
			{
				for (x=x0;x<=x1;x+=spacing_x)
				{
					glVertex3f(x,y0,z);
					glVertex3f(x,y1,z);
				}
				for (y=y0;y<=y1;y+=spacing_y)
				{
					glVertex3f(x0,y,z);
					glVertex3f(x1,y,z);
				}
			}
			break;
		}
	case 1:
		{
			glColor4f(1,0,0,1);
			for (y=y0;y<=y1;y+=spacing_y)
			{
				for (x=x0;x<=x1;x+=spacing_x)
				{
					glVertex3f(x,y,z0);
					glVertex3f(x,y,z1);
				}
				for (z=z0;z<=z1;z+=spacing_z)
				{
					glVertex3f(x0,y,z);
					glVertex3f(x1,y,z);
				}
			}
			break;
		}

	case 2:
		{
			glColor4f(0,0,1,1);
			for (x=x0;x<=x1;x+=spacing_x)
			{
				for (y=y0;y<=y1;y+=spacing_y)
				{
					glVertex3f(x,y,z0);
					glVertex3f(x,y,z1);
				}
				for (z=z0;z<=z1;z+=spacing_z)
				{
					glVertex3f(x,y0,z);
					glVertex3f(x,y1,z);
				}
			}
			break;
		}
	}

	glEnd();


//	glDepthMask(true);
//	glEnable(GL_DEPTH_TEST);
	glDisable(GL_FOG);


	glEndList();

}

void GRID::draw_extents()
{
	if (!frame.config.get_bool(VAR_DRAW_EXTENTS_MODE))
		return;

	int dem = frame.config.get_float(VAR_DRAW_EXTENTS_MODE);

	float x0,y0,z0,x1,y1,z1;
	x0 = frame.extents.x0;y0 = frame.extents.y0;z0 = frame.extents.z0;
	x1 = frame.extents.x1;y1 = frame.extents.y1;z1 = frame.extents.z1;

	if (dem == DEM_WIRE)
	{
		glLineWidth(4);
		glColor4f(
			frame.display.background_color[0]*2.0f,
			frame.display.background_color[1]*2.0f,
			frame.display.background_color[2]*2.0f,
			1);

		glBegin(GL_LINE_LOOP);
			glVertex3f(x0,y0,z0);glVertex3f(x0,y1,z0);glVertex3f(x1,y1,z0);glVertex3f(x1,y0,z0);
		glEnd();
		glBegin(GL_LINE_LOOP);
			glVertex3f(x0,y0,0);glVertex3f(x0,y1,0);glVertex3f(x1,y1,0);glVertex3f(x1,y0,0);
		glEnd();
		glBegin(GL_LINE_LOOP);
			glVertex3f(x0,y0,z1);glVertex3f(x0,y1,z1);glVertex3f(x1,y1,z1);glVertex3f(x1,y0,z1);
		glEnd();
		glBegin(GL_LINE_LOOP);
			glVertex3f(0,y0,z1);glVertex3f(0,y0,z0);glVertex3f(0,y1,z0);glVertex3f(0,y1,z1);
		glEnd();
		glBegin(GL_LINE_LOOP);
			glVertex3f(x0,0,z1);glVertex3f(x0,0,z0);glVertex3f(x1,0,z0);glVertex3f(x1,0,z1);
		glEnd();
		glBegin(GL_LINES);
			glVertex3f(x0,y0,z0);	glVertex3f(x0,y0,z1);
			glVertex3f(x1,y0,z0);	glVertex3f(x1,y0,z1);
			glVertex3f(x0,y1,z0);	glVertex3f(x0,y1,z1);
			glVertex3f(x1,y1,z0);	glVertex3f(x1,y1,z1);
		glEnd();

		glLineWidth(1);
	}
	if (dem == DEM_SOLID)
	{
		/*
		142	92	17	.555	.360	.066
		111	141	18	.434	.551	.070
		34	125	36	.133	.488	.141
		21	138	106	.082	.539	.414
		36	95	123	.141	.371	.480
		98	7	152	.383	.027	.594
		139	20	32	.543	.078	.125
		*/

		glBegin(GL_QUADS);
			glColor4f(.555f,	.360f,	.066f,	1);	glVertex3f(x0,y1,z0);glVertex3f(x0,y0,z0);glVertex3f(x1,y0,z0);glVertex3f(x1,y1,z0);
			glColor4f(.434f,	.551f,	.070f,	1);	glVertex3f(x1,y1,z1);glVertex3f(x1,y0,z1);glVertex3f(x0,y0,z1);glVertex3f(x0,y1,z1);

			glColor4f(.133f,	.488f,	.141f,	1);	glVertex3f(x0,y0,z1);glVertex3f(x0,y0,z0);glVertex3f(x0,y1,z0);glVertex3f(x0,y1,z1);
			glColor4f(.082f,	.539f,	.414f,	1);	glVertex3f(x1,y1,z1);glVertex3f(x1,y1,z0);glVertex3f(x1,y0,z0);glVertex3f(x1,y0,z1);

			glColor4f(.141f,	.371f,	.480f,	1);	glVertex3f(x1,y0,z1);glVertex3f(x1,y0,z0);glVertex3f(x0,y0,z0);glVertex3f(x0,y0,z1);
			glColor4f(.543f,	.078f,	.125f,	1);	glVertex3f(x0,y1,z1);glVertex3f(x0,y1,z0);glVertex3f(x1,y1,z0);glVertex3f(x1,y1,z1);
		glEnd();
	}
}

void GRID::draw_axes()
{
	float x0,y0,z0,x1,y1,z1;
	x0 = frame.extents.x0;y0 = frame.extents.y0;z0 = frame.extents.z0;
	x1 = frame.extents.x1;y1 = frame.extents.y1;z1 = frame.extents.z1;

	if (!frame.config.get_bool(VAR_DRAW_AXES))
		return;


	glLineWidth(10);
	glBegin(GL_LINES);
		glColor4f(1,0,0,1);	glVertex3f(0,0,0);
		glColor4f(1,0,0,0);	glVertex3f(x1,0,0);
		glColor4f(0,1,0,1);	glVertex3f(0,0,0);
		glColor4f(0,1,0,0);	glVertex3f(0,y1,0);
		glColor4f(0,0,1,1);	glVertex3f(0,0,0);
		glColor4f(0,0,1,0);	glVertex3f(0,0,z1);
	glEnd();
	glLineWidth(1);

/*
	glColor4f(1,0,0,1);
	glBegin(GL_POLYGON);
		glVertex3f(0,0,0);
		glVertex3f(x1,0,0);
		glVertex3f(x1,y1,0);
		glVertex3f(0,y1,0);
	glEnd();
*/

/*
	glBegin(GL_LINE_LOOP);
		glColor4f(1,0,0,1);	glVertex3f(0,0,0);
		glColor4f(1,0,0,0);	glVertex3f(x1,0,0);
		glColor4f(0,1,0,1);	glVertex3f(0,0,0);
		glColor4f(0,1,0,0);	glVertex3f(0,y1,0);
		glColor4f(0,0,1,1);	glVertex3f(0,0,0);
		glColor4f(0,0,1,0);	glVertex3f(0,0,z1);
	glEnd();
*/
}


GRID2D::GRID2D()
{
	minor = .01;
	major = .1;
}

//W must be window width (in world coords!)
void GRID2D::draw(double x, double y, double w, double h)
{
	int screen_x = frame.display.screen_x;
	int screen_y = frame.display.screen_y;
	double sx,sy,d,p,pix;

	glDisable(GL_TEXTURE_2D);

	d = minor;
	p = d / w;
	pix = screen_x * p;
	sx = (1 - ((x-(floor(x / d)*d)) / d)) * pix;
	sy = (1 - ((y-((floor(y / d))*d)) / d)) * pix;
	glColor4f(.5,.5,.5,.5);
	draw_aux(sx,sy,pix);


	d = major;
	p = d / w;
	pix = screen_x * p;
	sx = (1 - ((x-(floor(x / d)*d)) / d)) * pix;
	sy = (1 - ((y-((floor(y / d))*d)) / d)) * pix;
	glColor4f(.75,.75,.75,.75);
	draw_aux(sx,sy,pix);
}

void GRID2D::draw_aux(double sx, double sy, double pix)
{
	glBegin(GL_LINES);
	for (double i=sx; i<=frame.display.screen_x; i+=pix)
	{
		glVertex2f(i,0);
		glVertex2f(i,frame.display.screen_y);
	}
	for (double j=sy; j<frame.display.screen_y; j+=pix)
	{
		glVertex2f(0,j);
		glVertex2f(frame.display.screen_x,j);
	}
	glEnd();
}

void GRID2D::resize(int factor, double w)
{
	//If you ZOOM OUT, grid gets finer and finer.
	// We will concede that this can happen, but discourage
	// doing so.  It is not critical.

	double min = .001f;
	double mod = 2;
	int major_factor = 10;

	double p_min = .1;
	double p = major / w;

	if (p < p_min && factor < 0)
		return;


	if (factor>0)	minor *= mod;
	if (factor<0)	minor /= mod;

	major = minor * major_factor;

}






