

/*----------------------------------------------------*/
/* ibfv_sample.c - Image Based Flow Visualization     */
/*                                                    */
/* Jarke J. van Wijk, 2002                            */
/* Technische Universiteit Eindhoven                  */
/*----------------------------------------------------*/

/*******************************************************
   Modified from original to represent vector field
    in three dimensions.
   Doug Kavendek, 2006
*******************************************************/

#include "main.h"
extern FRAME frame;

extern double rnd(float range);
extern double rnd();


#define NMESH	100
#define	NPN		64
#define DM		((float) (1.0/(NMESH-1.0)))
#define NPIX	512
#define SCALE	2.0


IBFV::IBFV()
{
	iframe = 0;
	Npat = 32;
	alpha = (0.05*255);
	tmax = NPIX/(SCALE*NPN);
	dmax = SCALE/NPIX;

	ox = oy = oz = 0;
	wx = wy = wz = 1;

	face[0].slice = 0;
}

void IBFV::init(int new_num_faces)
{
	if (num_faces >= MAX_IBFV_FACES)
	{
		frame.warning("Too many faces specified for IBFV.");
		num_faces = MAX_IBFV_FACES - 1;
	}

	num_faces = new_num_faces;
	set_faces();

	tex = (GLuint*)calloc(num_faces,sizeof(GLuint));
	glGenTextures(num_faces, tex);

	make_patterns();

	timer = 0;
	timer_max = .01f;
	dye = 0;
}

IBFV::~IBFV()
{
	int i,j,k;
	if (face[0].slice != NULL)
	{
		for (k=0; k<num_faces; k++)
		{
			for (i=0; i<NMESH; i++)
			{
				for (j=0; j<NMESH; j++)
					free(face[k].slice[i][j]);
				free(face[k].slice[i]);
			}
			free(face[k].slice);
		}
	}
}

void IBFV::gl_start()
{
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glEnable(GL_TEXTURE_2D);
//	glDisable(GL_ALPHA_TEST);
}

void IBFV::gl_stop()
{
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
//	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
//	glDisable(GL_TEXTURE_2D);
//	glShadeModel(GL_SMOOTH);
//	glEnable(GL_ALPHA_TEST);
	glEnable(GL_BLEND);
}

//Generate a few phases of noise to be reused
void IBFV::make_patterns()
{
	int lut[256];
	int phase[NPN][NPN];
	GLubyte pat[NPN][NPN][4];
	int i, j, k, t;

	for (i = 0; i < 256; i++)
		lut[i] = i < 127 ? 0 : 255;

	//Enable this for static images
//	for (i = 0; i < 256; i++) lut[i] = 255;


	for (i = 0; i < NPN; i++)
		for (j = 0; j < NPN; j++) phase[i][j] = rand() % 256; 

	list_start = glGenLists(Npat);
	for (k = 0; k < Npat; k++)
	{
		t = k*256/Npat;
		for (i = 0; i < NPN; i++) 
		for (j = 0; j < NPN; j++)
		{
			pat[i][j][0] =
			pat[i][j][1] =
			pat[i][j][2] = lut[(t + phase[i][j]) % 255];
			pat[i][j][3] = alpha;
		}

		glNewList(k + list_start, GL_COMPILE);
		glTexImage2D(GL_TEXTURE_2D, 0, 4, NPN, NPN, 0, GL_RGBA, GL_UNSIGNED_BYTE, pat);
		glEndList();
	}
}

void IBFV::getDP(int i, int j, float *px, float *py, int g)
{
	*px = face[g].slice[i][j][0];
	*py = face[g].slice[i][j][1];
}

void IBFV::draw_all()
{
	int dvm = frame.config.get_float(VAR_DRAW_VECTORS_MODE2);
	if (dvm <= 0) return;

	glDisable(GL_CULL_FACE);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glDepthMask(false);
	glColor4f(1,1,1,.5f);

	if (dvm == DVM_ALL)
		for (int i=0; i<num_faces; i++)
		{
			draw(i);
		}
	else
	{
		switch (dvm)
		{
			default:
			case DVM_X: draw(0); break;
			case DVM_Y: draw(1); break;
			case DVM_Z: draw(2); break;
		}
	}

	glEnable(GL_CULL_FACE);
	glDepthMask(true);

	frame.magic.try_end();
	glDisable(GL_TEXTURE_2D);
}

void IBFV::draw(int g)
{
	glBindTexture(GL_TEXTURE_2D, tex[g]);
	glBegin(GL_POLYGON);


	//ugly prototype
	if (g==0)
	{
		glTexCoord2f(0,0);	glVertex3f(dx0,dy0,dz2);
		glTexCoord2f(0,1);	glVertex3f(dx0,dy1,dz2);
		glTexCoord2f(1,1);	glVertex3f(dx1,dy1,dz2);
		glTexCoord2f(1,0);	glVertex3f(dx1,dy0,dz2);
	}
	else if (g==1)
	{
		glTexCoord2f(0,0);	glVertex3f(dx0,dy2,dz0);
		glTexCoord2f(0,1);	glVertex3f(dx0,dy2,dz1);
		glTexCoord2f(1,1);	glVertex3f(dx1,dy2,dz1);
		glTexCoord2f(1,0);	glVertex3f(dx1,dy2,dz0);
	}
	else if (g==2)
	{
		glTexCoord2f(0,0);	glVertex3f(dx2,dy0,dz0);
		glTexCoord2f(0,1);	glVertex3f(dx2,dy0,dz1);
		glTexCoord2f(1,1);	glVertex3f(dx2,dy1,dz1);
		glTexCoord2f(1,0);	glVertex3f(dx2,dy1,dz0);
	}


	glEnd();
}

//Generate textures for all faces for next frame
void IBFV::generate_all()
{
	int start, end;
	int dvm = frame.config.get_float(VAR_DRAW_VECTORS_MODE2);
	if (dvm <= 0) return;

	timer += frame.engine.get_time_difference();
	if (timer < timer_max) return;
	timer = 0;

//We want same randomness here
if (dye>0) srand(24);

glDisable(GL_DEPTH_TEST);

glPushMatrix();
frame.camera->set_overlay_matrices();

	if (dvm == DVM_ALL)
	{
		start = 0;
		end = num_faces-1;
	}
	else
	{
		switch (dvm)
		{
			default:
			case DVM_X: start=end=0; break;
			case DVM_Y: start=end=1; break;
			case DVM_Z: start=end=2; break;
		}
	}
	//start/end just decides which to use, or to use all
	for (int i=start; i<=end; i++)
	{
		//We're drawing into an AUX buffer, instead of to the screen
		glDrawBuffer(GL_AUX0 + i);
		glReadBuffer(GL_AUX0 + i);
		generate(i);
	}

glPopMatrix();

glEnable(GL_DEPTH_TEST);


	//We must revert to the normal color buffer to draw into.
	//Apparently, we're using the LEFT/RIGHT buffers of the back buffer?
	// That is inane and crazy, what the hell people
	glDrawBuffer(GL_BACK);
	glReadBuffer(GL_BACK);
	gl_stop();
}

//Generate the texture for the next frame for face g
void IBFV::generate(int g)
{ 
	int   i, j; 
	float x1, x2, y, px, py;
	int color = 0;
	float c[3];

	int skip = 9;

	int rx = -1;
	int ry = -1;

	if (dye>0)
	{
		rx = NMESH*rnd();
		ry = NMESH*rnd();
 	}

gl_start();
glBindTexture(GL_TEXTURE_2D, tex[g]);


//glColor4f(1,0,0,1);

	//Draws offset of previous frame onto mesh
	for (i = 0; i < NMESH-1; i++)
	{
		x1 = DM*i; x2 = x1 + DM;
		glBegin(GL_QUAD_STRIP);
		for (j = 0; j < NMESH; j++)
		{
//			if (rnd()<.015) glColor4f(1,0,0,1);
//			else glColor4f(1,1,1,1);

//			if (i==rx && (j==ry || j+1==ry))	glColor4f(1,0,0,1);
//			else				glColor4f(1,1,1,1);



			if (color>0)
			{
				color--;
				glColor4f(c[0],c[1],c[2],1);
			}
			else if (dye>0 && rnd() < .03)
			{
				c[0] = rnd(); c[1] = rnd(); c[2] = rnd();
				color = 0;
				glColor4f(c[0],c[1],c[2],1);
			}
			else glColor4f(1,1,1,1);

//glColor4f(1,1,1,1);

			//Regular grid that shifts -- looks terrible!
//			if ((i+face[g].iframe/32)%skip==0 && (j+face[g].iframe/32)%skip==0)
//				glColor4f(rnd(),rnd(),rnd(),1);
//			else glColor4f(1,1,1,1);


			y = DM*j;
			glTexCoord2f(x1, y);
			getDP(i, j, &px, &py, g);
			glVertex3f((x1+px)*NPIX, (y+py)*NPIX, 0);

			glTexCoord2f(x2, y); 
			getDP(i+1, j, &px, &py, g); 
			glVertex3f((x2+px)*NPIX, (y+py)*NPIX, 0);
		}
		glEnd();
	}
	face[g].iframe = face[g].iframe + 1;

	//Draws next layer of noise over this mesh
	glColor4f(1,1,1,1);

	glEnable(GL_BLEND);
	glCallList(face[g].iframe % Npat + list_start);
	glBegin(GL_QUAD_STRIP);
		glTexCoord2f(0.0,  0.0);  glVertex3f(0.0, 0.0, 0);
		glTexCoord2f(0.0,  tmax); glVertex3f(NPIX, 0.0, 0);
		glTexCoord2f(tmax, 0.0);  glVertex3f(0.0, NPIX, 0);
		glTexCoord2f(tmax, tmax); glVertex3f(NPIX, NPIX, 0);
	glEnd();
	glDisable(GL_BLEND);


	//At this point, it will read from screen what should be drawn in the NEXT FRAME --
	// This next frame will be drawn slightly offset, using the getDP or whatnot.  This
	// is then blended with another frame of noise (not offset), and the result is then
	// (SURPRISE) back here, being taken off the screen and put into texture space
	// -- to be used next frame with offset again.
	//So, if I can get the previous two parts to draw to a buffer, then I can steal
	// those pixels, and bind them to a texture.  Then, I can draw that texture anywhere
	// I want, and ALSO, bind that texture when I want to draw the next offset frame.
	//***Note from the future!: That is exactly what I did!

	glBindTexture(GL_TEXTURE_2D, tex[g]);
	glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 0, 0, NPIX, NPIX, 0);


gl_stop();
}


void IBFV::set_faces()
{
	if (num_faces<3)
	{frame.error("Faces should be <=3 in IBFV, but they are: %i",num_faces); return;}

	int i, j, k;
	double x,y;
	double r[3];
	float vx, vy;
	float d;
	double sx,sy,sz;
	int dimx,dimy,dimz;
	float hx,hy,hz;

	float v, max_v;

	//zomg!
	for (k=0; k<num_faces; k++)
	{
		face[k].slice = (float***)calloc(NMESH,sizeof(float**));
		for (i=0; i<NMESH; i++)
		{
			face[k].slice[i] = (float**)calloc(NMESH,sizeof(float*));
			for (j=0; j<NMESH; j++)
				face[k].slice[i][j] = (float*)calloc(2,sizeof(float));
		}
	}

	//Get various data!
	frame.vector_field->get_origin(&ox,&oy,&oz);
	frame.vector_field->get_widths(&wx,&wy,&wz);
	frame.vector_field->get_dims(&dimx,&dimy,&dimz);
	wx *= dimx; wy *= dimy; wz *= dimz;

	//Fill dimensions to use in drawing
	dx0 = ox;	dx1 = ox + wx;	dx2 = ox + wx/2.0f;
	dy0 = oy;	dy1 = oy + wy;	dy2 = oy + wy/2.0f;
	dz0 = oz;	dz1 = oz + wz;	dz2 = oz + wz/2.0f;

	//We assume a cube for this
	d = wx / NMESH;

	//We need to offset the VF returns by half the size of a VF space
	hx = wx/(dimx*2.0);
	hy = wy/(dimy*2.0);
	hz = wz/(dimz*2.0);

	x = 0;
	y = 0;

	for (k = 0; k < num_faces; k++)
	{
		face[k].iframe = 0;
		max_v = 0;

		for (i = 0; i < NMESH; i++)
		{
			x = d*i;
			for (j = 0; j < NMESH; j++)
			{
				y = d*j;

				if (k==0)
				{
					sx = ox + x - hx;
					sy = oy + y - hy;
					sz = dz2;
					frame.vector_field->get_vector(sx,sy,sz, &r[0], &r[1], &r[2]);
					vx = r[0];
					vy = r[1];
				}
				else if (k==1)
				{
					sx = ox + x - hx;
					sy = dy2;
					sz = oz + y - hz;
					frame.vector_field->get_vector(sx,sy,sz, &r[0], &r[1], &r[2]);
					vx = r[0];
					vy = r[2];
				}
				else if (k==2)
				{
					sx = dx2;
					sy = oy + x - hy;
					sz = oz + y - hz;
					frame.vector_field->get_vector(sx,sy,sz, &r[0], &r[1], &r[2]);
					vx = r[1];
					vy = r[2];
				}
				else
				{
					vx = vy = 0;
				}


				//Record the maximum vector
				v = sqrt(vx*vx + vy*vy);
				if (v > max_v) max_v = v;

				face[k].slice[i][j][0] = vx;
				face[k].slice[i][j][1] = vy;
			}
		}


		if (frame.config.get_bool(VAR_NORMALIZE_IBFV))
		{
			//Now, normalize by the maximum vector.  The effectiveness of this visualization
			// depends on having the vector not be too high, because then it just looks like noise.
			//A downside is that a slow field will look the same as a faster field, but the relative
			// differences of flow within it will be articulated correctly, which is more important.
			// Besides, without normalizing this, very many fields will contain values too great, thus
			// making the visualization worthless for most, without annoying tweaking beforehand.
			for (i = 0; i < NMESH; i++)
			{
				for (j = 0; j < NMESH; j++)
				{
					vx = face[k].slice[i][j][0];
					vy = face[k].slice[i][j][1];
					if (vx!=0 || vy!=0)
					{
						vx /= max_v;
						vy /= max_v;
					}

					//Scale it back!  .03 seems a good value.  This means that the maximum speed
					// that any vector will have is .03.
					vx *= .03;
					vy *= .03;

					face[k].slice[i][j][0] = vx;
					face[k].slice[i][j][1] = vy;
				}
			}
		}


	}


}


void IBFV::dye_test()
{
	if (dye<=0)	dye = 1;
	else		dye = 0;
}
