//list.h

#ifndef _LIST_
#define _LIST_

#include "main.h"

class FRAME;
class LOGGER;

extern FRAME frame;
extern LOGGER err_log;

template <class T>
class LIST
{
public:
	LIST();

	void create(int new_max_length);
	void destroy();
	void clear();

	T *find_next();
	void step();
	void draw();

virtual 	void zero_vars() = 0;

	int get_total()		{return total;}
	void reset_total()	{total=0;}

protected:

//	virtual void clear_func(T *e) = 0;

	int num_elements;
	int max_length;
	T *list;

	int next_index;

	//Total number of elements ever created
	int total;
};




template <class T>
LIST<T>::LIST()
{
	max_length = 0;
	list = NULL;
	num_elements = 0;
	next_index = 0;
	total = 0;
}

template <class T>
void LIST<T>::create(int new_max_length)
{
	max_length = new_max_length;
	list = new T[max_length];
	num_elements = 0;
	next_index = 0;
	zero_vars();
}


template <class T>
void LIST<T>::destroy()
{
	delete [] list;
}

template <class T>
void LIST<T>::clear()
{
	for (int i=0; i<num_elements; i++)
	{
		//clear element
//		clear_func(&list[i]);

		list[i].zero_vars();
	}
	next_index = 0;
}


template <class T>
T *LIST<T>::find_next()
{
//	err_log << "FIND NEXT from: "<< next_index << endl;

	//If we move next_index, we'd have to move an index for draw and step
	// to start from.  Otherwise, performance will severely degrade over time.
	// Until I plan that out right, let's just start insertions from zero.
	next_index = 0;

	T *r = NULL;
	for (int i=0; i<max_length; i++)
	{
		int I = next_index + i;
//		err_log << "   check " << I << endl;
		while (I>=max_length) I -= max_length;

		if (!list[I].is_active())
		{
//			err_log << "FOUND at " << I << endl;
			r = &list[I];
//			err_log << "Found blank at " << I << endl;

			//Update list stuff
			num_elements++;
		    next_index = I+1;
			while (next_index>=max_length) next_index -= max_length;

			break;
		}
	}

	total++;
	return r;
}

template <class T>
void LIST<T>::step()
{
	if (num_elements<=0)
		return;

	int sub = 0;
	int i, n =0;
	for (i=0; i<max_length; i++)
	{
		if (list[i].is_active())
		{
			list[i].step();
			n++;

			//When something is killed, we need to remove its count from num_elements
			// HOWEVER, this must be done AFTER this has iterated through or else we
			// will interfere with the current num_elements (which is still accurate)
			if (!list[i].is_active())
				sub++;

			if (n>=num_elements)
			{
				num_elements -= sub;
				return;
			}
		}
	}
	num_elements -= sub;
}

template <class T>
void LIST<T>::draw()
{
	if (num_elements<=0)
		return;

	int i, n =0;
	if (num_elements<=0) return;
	for (i=0; i<max_length; i++)
	{
		if (list[i].is_active())
		{
			list[i].draw();
			n++;

			if (n>=num_elements)
				return;
		}
	}
}






#endif

