//magic_begin.cpp

#include "main.h"

extern FRAME frame;
extern LOGGER err_log;

MAGIC::MAGIC()
{
	in_block = false;
	current_tex = -1;
};

bool MAGIC::bind_texture(int new_tex)
{
	if (new_tex==current_tex) return false;
	glBindTexture(GL_TEXTURE_2D, new_tex);
	current_tex = new_tex;
	return true;
}

void MAGIC::try_begin()
{
	try_begin(current_tex);
}
void MAGIC::try_begin(int tex)
{
	if (tex!=current_tex)
	{
		if (in_block)
		{
			//Texture change, but in a block.  End, apply new tex, begin
			glEnd();
			in_block = false;
		}
		else
		{
			//Texture change, but not in a block.  Apply nex tex, begin
		}
		bind_texture(tex);
		glBegin(GLBEGIN_DRAWTYPE);
		in_block = true;
	}
	else
	{
		if (!in_block)
		{
			//No texture change, but we're not in a block, so start
			glBegin(GLBEGIN_DRAWTYPE);
			in_block = true;
		}
		else
		{
			//No texture change, and we're already in a block.  Continue.
		}
	}
}

void MAGIC::try_end()
{
	if (in_block)
	{
		glEnd();
		in_block = false;
	}
}



void MAGIC::set_color(float r, float g, float b, float a)
{
	color.r = r;
	color.g = g;
	color.b = b;
	color.a = a;
}
void MAGIC::set_alpha(float a)
{
	color.a = a;
}
void MAGIC::try_color()								{ glColor4f(color.r,color.g,color.b,color.a); }
void MAGIC::try_color(float r, float g, float b)	{ glColor4f(r,g,b,color.a); }
void MAGIC::try_color(float a)						{ glColor4f(color.r,color.g,color.b,a); }
