//model.h

#ifndef _MODEL_
#define _MODEL_

#include "main.h"


//This is the maximum number of adjacent polys/verts for each vertex.
//The average max may be around 15 or so, but in interim states it needs many more
// 100 may be too high, but if it ever is too low, it will crash
#define MAX_ADJ					100

//Scale of normals to be rendered
#define NORMAL_DISPLAY_SCALE	.002f

struct POLY;
struct EDGE;
struct VERT;
class BIN;
class MODEL;


//I have taken out any mention of VERTEX ARRAYS because the models change too often
// to make them worthwhile.  I wanted to leave the code intact though.

/*
struct VERTEX_ARRAY
{
	bool ready;

	VERTEX_ARRAY();
	void clear();
	void create();
	void draw();

	GLdouble *verts;
	GLfloat *color;
	GLuint *indices;
	GLdouble *normal;

	int vc;
	int pc;

	int display_list;
};
*/


struct VERT
{
	VERT();
	VERT(double *p);

	//Clear vertex parameters
	void zero_vars();

	double pos[3];
	float color[4];
	double normal[3];
	double dnormal[3];

	//Calculate normal from neighboring polygons
	void init_normal();

	//Add a poly to its adjacency list
	void add_adj(POLY *poly);
	//Remove a poly from its adjacency list
	void remove_adj(POLY *poly);
	//List of adjacent polys
	POLY *adj[MAX_ADJ];
	//Number of adjacent polys
	int num_adj;

	//Add a vertex to its adjacency list
	void add_adj_p(VERT *p);
	//List of adjacent vertices
	VERT *adj_p[MAX_ADJ];
	//Number of adjacent vertices
	int num_adj_p;

	//Curvature at this vertex
	double curvature;

	//Move the vertex
	void move(double mx, double my, double mz);
	void move(double *m);

	//Used while parsing edges
	bool edge_explored;
	bool vert_explored;

	//Link for vert list from model
	VERT *next;

	//Links for bin list from bin_src
	VERT *bin_next;
	VERT *bin_prev;
	BIN *bin_src;	//Bin in which vertex resides (not used yet)

	//Set if it is an orphan (is not contained in a vertex)
	bool orphan;

	//ID / index
	int id;

	//Model in which vertex is contained
	MODEL *src_model;

	//Used for testing
	bool marked;

	//Resiude at this vertex
	RESIDUE residue;

	//Laplacian calculation intermediate values
	double laplace[3];
};

struct EDGE
{
	EDGE();

	//The endpoints of the edge
	VERT *end[2];

	//Link for the edge list from model
	EDGE *next;
};

struct POLY
{
	POLY(int new_vert_count);

	//Calculate normal of face
	void init_normal(bool inverted);

	//Add a vertex to the polygon
	void add_vert(int i, VERT *p);

	void draw();

	//Subdivide the given polygon
	void subdivide();

	//Check neighbor
	bool is_neighbor(POLY *n);

	//Calculate and store the longest edge
	void set_longest_edge();

	//Links for poly list from model
	POLY *next;
	POLY *prev;

	//Normal of face
	double normal[3];

	//Links for bin list from bin_src
	POLY *bin_next;
	POLY *bin_prev;
	BIN *bin_src;	//Bin that currently contains this poly

	//Verts that comprise polygon
	VERT *vert[3];		//Assume triangles!
	int vert_count;

	//Neighbors of polygon
	POLY *neighbor[3];	//Assuming three neighbors!

	//Edges
	EDGE *edge[3];

	//Marking, for testing
	void mark() {marked = true;}
	bool marked;

	//ID, or index
	long id;

	//Link to model in which it is contained
	MODEL *src_model;

	//Subdivision is needed, so figure out how many levels are needed
	void set_subdiv_needed(double r, double rad_limit);

	//When subdivision is performed, this is set
	bool subdiv_performed;

	//When subdivision is needed, this is set
	bool subdiv_needed;

	//The new polygons that it is being subdivided into
	POLY *sub_poly[4];
	//The new vertices that it is being subdivided into
	VERT *sub_vert[3];

	//Number of subdivision levels needed
	int subdiv_levels;

	//The longest edge in that poly
	double longest_edge;

};


class MODEL
{
public:
	MODEL();

	//Empty model structure but keep model object
	void destroy();

	//Create dispaly lists
	void create_all_lists();
	void create_list();
	void create_list(int vp);
//	void create_va();

	void draw();
	void draw(int vp);
	void draw_polys(int vp);
	void draw_edges();
	void draw_normals();
	void draw_points();

	//Perform subdivision where indicated
	void subdivide();
	//Perform subdivision on every face
	void full_subdivision();

	//Invert the model (reverse order of each poly's vertices)
	void do_invert();

	//Recalculate model stuff
	void recalculate();
	void recalculate(bool new_bins);

	//Curvature calculations -- Only laplace is implemented
	void calculate_curvature_laplace();
	void calculate_curvature_desbrun();

	//Test vector field by applying it to the model
	void vector_field_test();

	//Translate all vertices along normal direction by scale
	void translate_by_normals(float scale);

	//Load the model from a file
	bool load_ply(char *str, char *new_name);

	//Initialize bin (only after creating it)
	void init();

//	void contribute_va(GLdouble *vertex, GLfloat *color, GLdouble *normal, GLuint *indices, int *vi, int *pi);

	bool is_inverted()		{return inverted;}
	bool is_ready()			{return ready;}
	int get_vert_count()	{return vert_count;}
	int get_poly_count()	{return poly_count;}

	//Given a list of vertices, calculate a new vertex position between.
	VERT *calc_new_vert(VERT **vert, int i);

	//Add a vertex to the model with the given index
	VERT *add_vert(int index);
	//Add a vertex at the given position
	VERT *add_vert(double *p);
	//Add a vertex at the given position with the given index
	VERT *add_vert(int index, double *p);
	//Retrieve a vertex given its index
	VERT *get_vert(int index);


	//Add a poly to the model, with the given vertices
	POLY *add_poly(int vert_count, VERT *one, ...);

	//Remove a poly from the model
	void remove_poly(POLY *poly);

	//Display name of model
	char name[256];

	//Used for surface deformation from particles
	float durability;

	//Try to put each poly into the right bin.
	void init_bins();

protected:

	void init_poly_normals();
	void init_vert_normals();
	void parse_edges();
	void parse_polys();
	void add_edge(VERT *p0, VERT *p1);

	p_ply ply;

	bool inverted;
	bool ready;

	//List of polys
	POLY *first_poly;
	POLY *last_poly;
	int poly_count;

	//List of verts
	VERT *first_vert;
	VERT *last_vert;
	int vert_count;

	//List of edges
	EDGE *first_edge;
	EDGE *last_edge;
	int edge_count;

	//4 display lists, one for each VIEWPORT
	int display_list[4];
};
















#endif