//photon.cpp

#include "main.h"

extern FRAME frame;
extern LOGGER err_log;
extern double rnd();
extern double rnd(float range);



PHOTON::PHOTON()
{
	zero_vars();
}

void PHOTON::zero_vars()
{
	PHYS_OBJ::zero_vars();

	//Clear residues to zero
	residue.reset();
	//Set residue to contain moss
	residue.moss = 1;

	//Choose a random color
	color[0] = rnd();
	color[1] = rnd();
	color[2] = rnd();
	color[4] = 1;
	color[(int)(rnd()*3)] = rnd()*.5f + .5f;
	trail.zero_vars();
	trail.set_color(color);

	//Number of collisions before photon is killed.  0 => no max
	collision_max = 0;

	steps = 0;

	//Three weights.
	// - elasticity is multiplied with reflected velocity
	// - gravity is multiplied by gravity's absolute strength
	// - vf_bias is multiplied with the vector field at the particle's location, before being turned into a force
	elasticity = frame.config.get_float(VAR_PHOTON_ELASTICITY);
	gravity = 0;		//This forces them to not adhere to gravity until they collide something
	vf_bias = frame.config.get_float(VAR_PHOTON_VF_BIAS);
	conserve_momentum = frame.config.get_bool(VAR_PHOTON_CONSERVE);

	//If fire straight is not set, then start with gravity
	if (!frame.config.get_bool(VAR_FIRE_STRAIGHT))
		gravity = frame.config.get_float(VAR_PHOTON_GRAVITY);


/*
	elasticity = frame.config.get_float(VAR_PHOTON_ELASTICITY);
	gravity = frame.config.get_float(VAR_PHOTON_GRAVITY);
	vf_bias = frame.config.get_float(VAR_PHOTON_VF_BIAS);
	conserve_momentum = frame.config.get_bool(VAR_PHOTON_CONSERVE);
*/
	check_vars();
}

void PHOTON::check_vars()
{
	PHYS_OBJ::check_vars();

	//any photon vars to check
}

void PHOTON::draw()
{
	glPointSize(5);

	glColor3fv(color);

	glBegin(GL_POINTS);
	glVertex3f(pos[0],pos[1],pos[2]);
	glEnd();

	glPointSize(1);

	trail.draw();
}

void PHOTON::step()
{
	steps++;
//	frame.messages->add("STEPS: %i",steps);



//	err_log << endl << endl << "Starting photon step" << endl;
//	if (marked) return;

/*
	double t[3];
	frame.vector_field->get_vector(pos[0],pos[1],pos[2],&t[0],&t[1],&t[2]);
//	float scale = .0001f;
//	float scale2 = 50;
	float scale = .00005f;
	scale *= vf_bias;
//	t[0] *= scale;t[1] *= scale;t[2] *= scale;
	t[0] *= scale;t[1] *= scale;t[2] *= scale;

	add_force(t[0],t[1],t[2]);
//	force[0]=force[1]=force[2] = 0;
*/


//	vel[0] = t[0] * 50;
//	vel[1] = t[1] * 50;
//	vel[2] = t[2] * 50;

	convert_force();

/*
	vel[0] = vel[0]*(1-vf_bias) + t[0]*vf_bias;
	vel[1] = vel[1]*(1-vf_bias) + t[1]*vf_bias;
	vel[2] = vel[2]*(1-vf_bias) + t[2]*vf_bias;
*/

	//Check if collision has occurred
	if (!move())
	{
		//Turn on gravity after collision
		gravity = frame.config.get_float(VAR_PHOTON_GRAVITY);

//		err_log << "LAST POLY: " << &last_poly_collision << endl;
//		err_log << "LAST POLY: " << last_poly_collision.poly << endl;

//		err_log << "last_poly_collision: " << last_poly_collision << endl;
//		err_log << "last_poly_collision.poly: " << last_poly_collision.poly << endl;
//		err_log << "last_poly_collision->poly->marked: " << last_poly_collision->poly->marked << endl;
//		err_log << "last_poly_collision->poly->vert[0]: " << last_poly_collision->poly->vert[0] << endl;

		if (last_poly_collision.poly)
		{
//			last_poly_collision.poly->vert[0]->residue.moss = 1;

			//.05 - .001
//			d = (residue.moss * 100.0 * rnd()) + (residue.moss * 3.0);
//			f = .1f;
//			d *= 0.0005f;

//			err_log << "LAST: "<< last_poly_collision.speed << endl;
//			err_log << " D: " << d << endl;


//			if (!frame.config.get_bool(VAR_SUBDIV_ONLY))
			{
				double q[3];
				float u,v;
				u = last_poly_collision.u;
				v = last_poly_collision.v;
				q[0] = 1 - (u+v);	q[1] = u;	q[2] = v;

				residue.mark(last_poly_collision.poly,q);

				if (rnd()<TERMINATE_CHANCE)
				{
					//Make particle come to rest at current point.
					// Dump entire residue there!
	//				frame.messages->add("LAND");
					residue.land(last_poly_collision.poly,q);
					kill();
					return;
				}
				else
				{
					//Pick up residue from face
					residue.receive(last_poly_collision.poly,q);
				}
			}

		}

	}

	//Kill if collision max is enabled and has been reached
	if (collision_max && collision_count>=collision_max)	kill();

//	err_log << "Photon done stepping" << endl;
}

void PHOTON::activate()
{
	//Any photon-specific activation

	PHYS_OBJ::activate();
}

void PHOTON_LIST::add(double nx, double ny, double nz)
{
	add(nx,ny,nz,0,0,0);
}
PHOTON *PHOTON_LIST::add(double nx, double ny, double nz, double vx, double vy, double vz)
{
	PHOTON *p = find_next();
//	err_log << "ADDED PHOTON: " << p << endl;
	if (p)
	{
		p->zero_vars();

		p->set_pos(nx,ny,nz);

//		err_log << "NEW VEL: " << vx << ", " << vy << ", " << vz << endl;

		double scale = 500;
		vx/=scale;vy/=scale;vz/=scale;

		p->vel[0] = vx;
		p->vel[1] = vy;
		p->vel[2] = vz;

		p->add_force(rnd(.0001f),rnd(.0001f),rnd(.0001f));
		p->activate();
	}

	return p;
}

//run a simulation!  Wooo
void PHOTON_LIST::run_simulation()
{
	err_log << "Running simulation... " << endl;
	frame.config.set(VAR_IN_SIMULATION,1);
	char msg[32];

	int max_waves = frame.config.get_float(VAR_SIM_WAVES);
	int this_total = 0;
	double q;
	int n;

	//Spread type determines how photons are spread about
	// HOWEVER - when using ST_RANDOM, a given target count is just sent
	// within a given extent.  This causes different results based on model size!
	// Using ST_GRID or ST_JITTER uses, rather, a density, which provides similar
	// results for all models.
	int spread_type = frame.config.get_float(VAR_SPREAD_TYPE);
//	spread_type = ST_RANDOM;

	int seed;

	double x0,y0,z0,x1,y1,z1;
	x0 = frame.config.get_float(VAR_SIM_X0);
	y0 = frame.config.get_float(VAR_SIM_Y0);
	z0 = frame.config.get_float(VAR_SIM_Z0);
	x1 = frame.config.get_float(VAR_SIM_X1);
	y1 = frame.config.get_float(VAR_SIM_Y1);
	z1 = frame.config.get_float(VAR_SIM_Z1);

	for (int wave=0; wave<max_waves; wave++)
	{
		if (!frame.config.get_bool(VAR_INCREMENTAL_SUBDIV))
		{
			frame.config.set(VAR_SUBDIV_ONLY,false);

			if (wave==0 && first_subdiv)
			{
				seed = frame.engine.get_time();
				max_waves++;
				frame.config.set(VAR_SUBDIV_ONLY,true);
				first_subdiv = false;
			}
			else if (wave!=1)
				seed = frame.engine.get_time();

			srand(seed);
		}


		sprintf(msg,"Photon Wave %i/%i",wave+1, max_waves);
		frame.display.overlay->set_wait_reason(msg);

		err_log << "* Wave " << wave <<", " << frame.config.get_bool(VAR_SUBDIV_ONLY) << endl;

		n = 0;
		if (spread_type == ST_RANDOM)
		{
			int photon_target = 100;
			double rx,ry,rz;
			double d = .002f;

			for (int i=0; i<photon_target; i++)
			{
				rx = rnd()*(x1-x0) + x0;
				ry = rnd()*(y1-y0) + y0;
				rz = rnd()*(z1-z0) + z0;
				//Photons are given a downward force!
				frame.photon_list.add_and_run(rx,ry,rz, 0,0,-1);
				q = (float)i / (float)(photon_target-1);

				if (n%10==0)
					frame.display.overlay->post_progress((q+wave)/(float)max_waves);

				n++;
			}
		}
		else
		{
			double dx,dy,dz;
			dx=dy=dz = .003f;		//.004f

			double eqx = (fabs(x1-x0) / fabs(dx)) + 1;
			double eqy = (fabs(y1-y0) / fabs(dy)) + 1;
			double eq = eqx * eqy;
			float jitter_amount = 1;	//1 implies no overlap

			for (double qz=z0; qz<=z1; qz+=dz)
			{
				for (double qx=x0; qx<=x1; qx+=dx)
				{
	//				frame.display.overlay->post_progress(1.0 - fabs(qx+x0)/fabs(x1-x0));
					for (double qy=y0; qy<=y1; qy+=dy)
					{
						//Photons are given a downward force!
						if (spread_type==ST_JITTER)
								frame.photon_list.add_and_run(qx+rnd(jitter_amount*dx), qy+rnd(jitter_amount*dy), qz+rnd(jitter_amount*dz), 0,0,-1);
						else	frame.photon_list.add_and_run(qx, qy, qz, 0,0,-1);

						q = total / eq;
						if (n%10==0) frame.display.overlay->post_progress((q+wave)/(float)max_waves);
						n++;
					}
				}
			}
		}

		this_total += frame.photon_list.get_total();
		frame.photon_list.reset_total();

		//Reduce bin size!
//		float w_min = .0018f;

		//Analysis of models WILL NOT determine the smallest bin size.  If a model has
		// a single large triangle that is NEVER HIT, then its max will corrupt the
		// calculations.  And if it subdivides a lot, those small triangles do not say much.
		//RATHER -- we will simply base it off the target subdivision size.
		//ALSO:  If we are incrementally subdividing, then halve it each time.  OTHERWISE,
		// just set it to the min!

		frame.bin_down(false);

		for (int i=0; i<MAX_MODELS; i++)
		{
//			sprintf(msg,"Subdividing Model %i",i);
//			frame.display.overlay->post_progress(1.0f,msg);
			frame.model[i].subdivide();
		}

		//Remove everything from bins.
		//Reduce bin size
		//Add all triangles back to new bins


		for (int j=0; j<MAX_MODELS; j++)
		{
			frame.model[j].recalculate(true);
//			frame.model[j].init_bins();
		}
	}

	frame.messages->add("Photon sim run with %i photons",this_total);

	//Recalc after all waves
//	for (int j=0; j<MAX_MODELS; j++)
//		frame.model[j].recalculate(false);

	err_log << "Simulation complete." << endl;

	frame.config.set(VAR_IN_SIMULATION,0);
}

//Add a photon.  Run it for some time.  Kill it at end.
void PHOTON_LIST::add_and_run(double nx, double ny, double nz, double vx, double vy, double vz)
{
//	err_log << "Adding and running photon... " << endl;
	int max = 10000;	//10000
	PHOTON *p = add(nx,ny,nz,vx,vy,vz);
	if (!p) return;

	for (int i=0; i<max; i++)
	{
		p->step();
		if (!p->active)
		{
			return;
		}
	}
	p->kill();
}


void PHOTON_LIST::zero_vars()
{
	first_subdiv = true;

//	LIST<PHOTON_LIST>::zero_vars();
}

