//phys_obj.cpp

#include "main.h"


extern FRAME frame;
extern double rnd();
extern double rnd(float range);


static float lifetime_average = 0;
static int lifetime_samples = 0;

PHYS_OBJ::PHYS_OBJ()
{
	zero_vars();
	trail.create(128,this);
}

void PHYS_OBJ::activate()
{
	//Choose a new rad limit for subdivision
	rad_limit = frame.config.get_float(VAR_SUBDIV_TARGET);
	//Randomize slightly - may want more control of this later
	rad_limit += rnd()*(rad_limit/2.0);

	birth_time = frame.engine.get_time();
	active = true;
}

void PHYS_OBJ::zero_vars()
{
	active = false;

	collision_count = 0;
	collision_max = 0;
	marked = false;
	pos[0] = pos[1] = pos[2] = 0;
	vel[0] = vel[1] = vel[2] = 0;
	acc[0] = acc[1] = acc[2] = 0;
	force[0] = force[1] = force[2] = 0;
	mass = 1;
	extent_bound = true;
	just_collided = false;
	t = 1;
	v = 0;
	bin_src = NULL;

	gravity = 0;
	elasticity = 1;

	vf_bias = 0;//There is no need for this ZERO crap
	vf_bias = frame.config.get_float(VAR_PHOTON_VF_BIAS);

	conserve_momentum = frame.config.get_bool(VAR_PHOTON_CONSERVE);

	rad_limit = 1000000;

	last_poly_collision.reset();
}

void PHYS_OBJ::check_vars()
{
	//Elasticity should never be zero.  It can be really small though.
	if (elasticity==0) elasticity = .000000001;
}

void PHYS_OBJ::add_force(double n[3])
{
	force[0] += n[0];
	force[1] += n[1];
	force[2] += n[2];
}
void PHYS_OBJ::add_force(double nx, double ny, double nz)
{
	force[0] += nx;
	force[1] += ny;
	force[2] += nz;
}

void PHYS_OBJ::convert_force()
{
	double dt = 1.0f;

	add_force(0,0,-gravity*GRAVITY);



	double t[3];
	float scale = .00005f;

    frame.vector_field->get_vector(pos[0],pos[1],pos[2],&t[0],&t[1],&t[2]);
	scale *= vf_bias;
	t[0] *= scale;t[1] *= scale;t[2] *= scale;

	if (conserve_momentum)
		add_force(t[0],t[1],t[2]);
	else
	{
		double scale2 = 100;
		vel[0] = t[0] * scale2;
		vel[1] = t[1] * scale2;
		vel[2] = t[2] * scale2;

		if (vel[0]==vel[1] && vel[1]==vel[2] && vel[2] == 0)
		{
			//May as well kill photon?  If we're not conserving velocity, then the ONLY thing that affects it is
			// the vector field.  Since it is not in the vector field, or the field is returning zero,
			// it can't move, and so can't ever go to a point that has a non-zero field.  Therefore, it won't
			// EVER move, and so won't affect anything.  Save time, and kill it.

			kill();
		}
	}






	//Old damping ideas
/*
	double max_vel = .002f;
	double damp = .9f;
	if (v>max_vel)
	{
		//Concept1 - just crop to max
//		vel[0] = (vel[0]/v) * max_vel;	vel[1] = (vel[1]/v) * max_vel;	vel[2] = (vel[2]/v) * max_vel;

		//Concept 2 - damp if over max - max is suggestion
		vel[0] *= damp;	vel[1] *= damp;	vel[2] *= damp;
	}
*/

/*
	//New damping ideas
	double dfv[3];
	//obviously.  pshaw
	if (v!=0)
	{
		dfv[0] = vel[0] / v;
		dfv[1] = vel[1] / v;
		dfv[2] = vel[2] / v;

		double drag = 4;
		double df = -drag * v * v;
		dfv[0] *= df;
		dfv[1] *= df;
		dfv[2] *= df;

//		dfv[0] = -drag * vel[0] * vel[0];
//		dfv[1] = -drag * vel[1] * vel[1];
//		dfv[2] = -drag * vel[2] * vel[2];

		add_force(dfv);
	}
*/

	//Very simple.  We'll assume every movement has the same time step.
	//If this proves too simple, we'll deal with it later

	acc[0] = force[0] / mass;
	acc[1] = force[1] / mass;
	acc[2] = force[2] / mass;

	vel[0] += acc[0] * dt;
	vel[1] += acc[1] * dt;
	vel[2] += acc[2] * dt;

//	v = sqrt(vel[0]*vel[0] + vel[1]*vel[1] + vel[2]*vel[2]);






}

//Returns true if no collisons were made
bool PHYS_OBJ::move()
{
/*
	trail.add(pos[0],pos[1],pos[2]);
	move_from_velocity(1);
	return true;
*/


	bool r = false;
//	err_log << "Going to move..." << endl;
//	frame.pvector("  Current velocity",vel);

	//Collision!

	frame.bins->collide_proj2poly(this);

//	if (last_poly_collision.poly)
//		err_log << just_collided << " RIGHT AFTER COLLIDEPROJPOLY " << last_poly_collision.poly << endl;
//	else err_log << "["<< just_collided <<"]";

	if (!just_collided)
	{
		//Temp - keep the crap in the box
		if (pos[0]>frame.extents.x1) {collide(0,frame.extents.x1);return false;}
		if (pos[1]>frame.extents.y1) {collide(1,frame.extents.y1);return false;}
		if (pos[2]>frame.extents.z1) {collide(2,frame.extents.z1);return false;}
		if (pos[0]<frame.extents.x0) {collide(0,frame.extents.x0);return false;}
		if (pos[1]<frame.extents.y0) {collide(1,frame.extents.y0);return false;}
		if (pos[2]<frame.extents.z0) {collide(2,frame.extents.z0);return false;}
	}

	force[0] = force[1] = force[2] = 0;
	trail.add(pos[0],pos[1],pos[2]);

	if (!just_collided)
	{
//		err_log << "Moving regularly, no collision" << endl;
		move_from_velocity(1);
		r = true;
	}
	else
	{
//		err_log << "COLLISION ! !" << endl;
//		err_log << "Moving not being done normally, collision detected" << endl;
		just_collided = false;
		r = false;
	}

	if (!conserve_momentum)
	{
//		err_log << "ARRRRRRR" << endl;
		vel[0]=vel[1]=vel[2] = 0;
	}

	return r;
}

void PHYS_OBJ::collide(int dim, double limit)
{
//	collision_count++;

	//This happens when it leaves extents.
	kill();
	return;


	if (dim==2 && pos[dim] < 0)
	{
		float ox,oy,oz;
		float dx,dy,dz;
		ox = 0; oy = 0; oz = .1f;
		dx = dy = dz = .1f;

		oy = .1f;

		frame.pvector("VELOCITY",vel);

//		frame.photon_list.add(ox+rnd(dx),oy+rnd(dy),oz+rnd(dz));

		kill();

		return;
	}

	pos[dim] = limit;
	vel[dim] *= -elasticity;

//	if (rnd()>.5f)	gravity *= -1;
//	else			gravity = 0;

	if (dim==2)
		gravity *= -1;

	//This is really just a temp function - not to be taken seriously
//	just_collided = true;
}

//Collide a particle with a polygon.
void PHYS_OBJ::collide_poly()
{
//	err_log << "COLLISION" << endl;
	int i;
//	frame.config.toggle(VAR_FREEZE_PARTICLES);
//	frame.messages->add("Collision!");

//	last_poly_collision = closest;
	CLOSEST_COLLISION *closest = &last_poly_collision;
	collision_count++;


//	gravity = 1;
//	vf_bias = .5;

//	elasticity = .1;
//	gravity = 1;
//	vf_bias = 0;
//	conserve_momentum = true;

	elasticity = frame.config.get_float(VAR_PHOTON_ELASTICITY);
	gravity = frame.config.get_float(VAR_PHOTON_GRAVITY);
	vf_bias = frame.config.get_float(VAR_PHOTON_VF_BIAS);
	conserve_momentum = frame.config.get_bool(VAR_PHOTON_CONSERVE);

//	err_log << "ELAST: " << elasticity << endl;

	POLY *poly;
	mark();


	float u,v;
	t = closest->t;
	poly = closest->poly;
//	poly->mark();
	u = closest->u;
	v = closest->v;



	//Account for movement DIRECTLY TO POLY
	move_from_velocity(t);



//	static double rmin = 10000;
//	static double rmax = 0;
//	static double rtot = 0;
//	static int rnum = 0;

//	if (poly->neighbor[0]) poly->neighbor[0]->marked = true;
//	if (poly->neighbor[1]) poly->neighbor[1]->marked = true;
//	if (poly->neighbor[2]) poly->neighbor[2]->marked = true;

	if (!poly->subdiv_needed)
	{
		//Check if subdivsion will be needed
		//.0008 = bunny
		//.0004 = horse
		//.004 = single plane!

		//This may be more accurate if expanded to vertices in neighboring polys
		// also, to see if they should be given the same subdiv setting
		//Maybe base rad_limit off velocity

/*
		double r;
		int outside;
		outside = 0;
		for (i=0;i<3;i++)
		{
			r = sqrt(	pow(pos[0]-poly->vert[i]->pos[0],2) + 
						pow(pos[1]-poly->vert[i]->pos[1],2) + 
						pow(pos[2]-poly->vert[i]->pos[2],2) );

			if (r>rad_limit) outside++;
			if (outside>=2)
			{
				poly->set_subdiv_needed();
				break;
			}
		}
*/

		//THIS is totally wrong.  I should be doing something like
		// the inverse.  Because, vertices in adjacent triangles will be
		// even further away -- they is no good reason to be subdividing
		// them just because they are further outside the radius.  Unless
		// we use a larger radius for further triangles.  Maybe double it.
		//NO -- doubling it doesn't quite do it.  We must decide to subdivide
		// based purely on polygon size.  How about, do a calc per poly to
		// get an area estimate.  If it's over some amount, subdiv it.
		// Do this for all polys in general area of collision, if the have ANY
		// vertices within the radius.
		//Actually, with a multiplier of 4 it seems to work nicely.

		const int check_max = 50;
		POLY *check[check_max];
		POLY *p;

		check[0] = poly;
		int n=1;
		int q,Q;
		check[n] = NULL;

		for (q=0; q<3; q++)
		{
			for (Q=0; Q<poly->vert[q]->num_adj; Q++)
			{
				p = poly->vert[q]->adj[Q];
				for (i=0; i<n; i++)
					if (check[i]==p) break;
				if (i==n)
				{
					check[n] = p;
					if (n+1<check_max)
					{
						n++;
						check[n] = NULL;
					}
				}
			}
		}

		for (n--;n>=0;n--)
		{
			p = check[n];
//			err_log << "LONGEST EDGE: " << p->longest_edge << ", " << rad_tmp << endl;
			if (p->longest_edge > rad_limit)
			{
//				err_log << "SUBDIV!" << endl;
				p->set_subdiv_needed(p->longest_edge,rad_limit);
			}

			/*
			outside = 0;
			for (i=0;i<3;i++)
			{
				r = sqrt(	pow(pos[0]-p->vert[i]->pos[0],2) + 
							pow(pos[1]-p->vert[i]->pos[1],2) + 
							pow(pos[2]-p->vert[i]->pos[2],2) );

				if ((n==0 && r>rad_limit) || r>rad_limit*4) outside++;
				if (outside>=2)
				{
					p->set_subdiv_needed(r,rad_limit);
					break;
				}
			}
			*/
		}



	}

//	kill(); return;

//	err_log << "YE OLDE: " << rtot / rnum << ", " << rmax << ", " << rmin << ", " << rtot << endl;

	double min_speed = 0.0001;	//.0001
	double n[3], f[3], q[3];
	f[0] = vel[0];f[1] = vel[1];f[2] = vel[2];
//	normalize(f);	//No need to normalize - we want to preserve the speed


	//seriously - what was going on down there:
/*
//	if (closest->speed < MIN_PARTICLE_SPEED)
//	if (closest->speed < 0.000009)		//.000009
	{

		double Q = closest->poly->vert[0]->pos[0]*vel[0] +
			closest->poly->vert[0]->pos[1]*vel[1] +
			closest->poly->vert[0]->pos[2]*vel[2];

		frame.messages->add("Q: %f",Q);


		if (Q<.000000000001)	//.000000000001
			kill();

	}
*/

	//Potential thought:  If velocity is low enough,
	// find direction of plane of collided polygon.  Find
	// direction it's trying to move.  Find a new velocity
	// that sort of combines both.  In essence, we want its
	// new vel to be facint orthogonally to the poly.
	//WE WANT TO PROJECT THE VELOCITY VECTOR ONTO THE POLYGON'S
	// PLANE, NORMALIZE THIS RESULT, AND MULTIPLY IT BY THE
	// ORIGINAL VELOCITY'S SPEED (THEN APPLY ELASTICITY).
	//This may not be needed, however - just forcing the
	// velocity to match some given min, we get a similar
	// effect with half the calories, er, work.



	//REFLECT
	n[0] = poly->normal[0];	n[1] = poly->normal[1];	n[2] = poly->normal[2];
	double dot = vel[0]*n[0] + vel[1]*n[1] + vel[2]*n[2];
	vel[0] = vel[0] - 2 * (dot) * n[0];
	vel[1] = vel[1] - 2 * (dot) * n[1];
	vel[2] = vel[2] - 2 * (dot) * n[2];






	//Is there a need for a minimum collision speed?
	float speed = sqrt(f[0]*f[0] + f[1]*f[1] + f[2]*f[2]);
	float min_perturb = min_speed * 1.5f;
	float durability = poly->src_model->durability;
	closest->speed = speed;
//	if (closest->speed >= MIN_COLLISION_DAMAGE_SPEED)
//	if (closest->speed >= 0.000008)	//0.000008
	if (!frame.config.get_bool(VAR_SUBDIV_ONLY))
	{
		//Surface geometric perturbation of polygonal mesh (move dem vertexes!)

		f[0] /= speed;
		f[1] /= speed;
		f[2] /= speed;

		float diff = speed - (min_speed - min_perturb);
		if (diff<0) diff = 0;
		f[0] *= diff;
		f[1] *= diff;
		f[2] *= diff;

		//Force of distortion
		double force = .100;		//.05 or .5
		force *= (1.0 - durability);		//Durability is a parameter of the src model

		//f[] is the angle of collision (not normalized! - thus collision speed is taken into account)
		f[0] *= force; f[1] *= force; f[2] *= force;

		//Spread the love to the different vertices.  Perhaps, this should become gaussian
		// distributed instead of linear
		q[0] = 1 - (u+v);	q[1] = u;	q[2] = v;


		//Modify based on curvature here
		double c[3];
		double C[3][3];
		double cscale = -.1;	//Aligned to match with .05 force
		c[0] = poly->vert[0]->curvature;c[1] = poly->vert[1]->curvature;c[2] = poly->vert[2]->curvature;
		if (c[0]<0) c[0] = 0;if (c[1]<0) c[1] = 0;if (c[2]<0) c[2] = 0;
		c[0] *= cscale; c[1] *= cscale; c[2] *= cscale;

		//Get normal, scale by curvature
		C[0][0] = poly->vert[0]->normal[0]*c[0];C[0][1] = poly->vert[0]->normal[1]*c[0];C[0][2] = poly->vert[0]->normal[2]*c[0];
		C[1][0] = poly->vert[1]->normal[0]*c[1];C[1][1] = poly->vert[1]->normal[1]*c[1];C[1][2] = poly->vert[1]->normal[2]*c[1];
		C[2][0] = poly->vert[2]->normal[0]*c[2];C[2][1] = poly->vert[2]->normal[1]*c[2];C[2][2] = poly->vert[2]->normal[2]*c[2];

		//Scale weight of curvature effect, and push effect (direction of collision)
		float curve_weight = 0;	//This is not scaled by speed yet!  It should be!  Disaster results otherwise!
		float push_weight = 1;

		//Move it!

		double offset[3][3];

		offset[0][0] = (C[0][0]*curve_weight + f[0]*push_weight) * q[0];
		offset[0][1] = (C[0][1]*curve_weight + f[1]*push_weight) * q[0];
		offset[0][2] = (C[0][2]*curve_weight + f[2]*push_weight) * q[0];

		offset[1][0] = (C[1][0]*curve_weight + f[0]*push_weight) * q[0];
		offset[1][1] = (C[1][1]*curve_weight + f[1]*push_weight) * q[0];
		offset[1][2] = (C[1][2]*curve_weight + f[2]*push_weight) * q[0];

		offset[2][0] = (C[2][0]*curve_weight + f[0]*push_weight) * q[0];
		offset[2][1] = (C[2][1]*curve_weight + f[1]*push_weight) * q[0];
		offset[2][2] = (C[2][2]*curve_weight + f[2]*push_weight) * q[0];

//		if (!frame.config.get_bool(VAR_SUBDIV_ONLY))
//		{
			poly->vert[0]->move(offset[0]);poly->vert[1]->move(offset[1]);poly->vert[2]->move(offset[2]);

			poly->vert[0]->residue.collisions++;
			poly->vert[1]->residue.collisions++;
			poly->vert[2]->residue.collisions++;
//		}

	}
//	else
//	{
//		poly->vert[0]->move(0,0,0); poly->vert[1]->move(0,0,0); poly->vert[2]->move(0,0,0);
//	}


	if (frame.config.get_bool(VAR_AUTO_REFRESH) && !frame.config.get_bool(VAR_IN_SIMULATION))
	{
		//Hack, little temporary prototype of updating specific normals
		// WARNING: This does NOT update curvature!
		for (i=0; i<3; i++)
			for (int j=0; j<poly->vert[i]->num_adj; j++)
				poly->vert[i]->adj[j]->init_normal(poly->src_model->is_inverted());
		for (i=0; i<3; i++)
			poly->vert[i]->init_normal();

		//Refresh all display lists, as this has been updated
		poly->src_model->create_list();
	}


	//New velocity is scaled from elasticity
	double e=0;
//	e = elasticity;
	while (e==0) e = elasticity * rnd();
	vel[0] *= e; vel[1] *= e; vel[2] *= e;


	//Adjust velocity to be >= minimum speed.  This smells of HAX
	// But it makes sliding particles so much better
	//UGH have I don't enough sqrts?  I am awful today
	speed = sqrt(vel[0]*vel[0] + vel[1]*vel[1] + vel[2]*vel[2]);
	if (speed<min_speed && speed!=0)
	{
		double mvv = min_speed / speed;
		vel[0] *= mvv; vel[1] *= mvv; vel[2] *= mvv;
	}

	//AUX trail addition - otherwise, the 1-t movement makes for weird trails
	trail.add(pos[0],pos[1],pos[2]);

	//Account for movement DIRECTLY IN REFLECTION with remaining energy
	t = 1 - t;
	move_from_velocity(t);

	//Keep track of this, so that move() knows the particle should not move uninhibited
	just_collided = true;
}

/*
void PHYS_OBJ::move_from_velocity()
{
	pos[0] += vel[0];
	pos[1] += vel[1];
	pos[2] += vel[2];
}
*/
void PHYS_OBJ::move_from_velocity(float f)
{
	if (frame.config.get_bool(VAR_FREEZE_PARTICLES))
		return;

	pos[0] += vel[0] * f;
	pos[1] += vel[1] * f;
	pos[2] += vel[2] * f;
}


void PHYS_OBJ::kill()
{
	float t = frame.engine.get_time() - birth_time;
	lifetime_average += t;
	lifetime_samples++;

//	frame.messages->add("Time alive: %f    Avg: %f",t,lifetime_average/lifetime_samples);
	
	zero_vars();
}
