//residue.cpp

#include "main.h"

extern double rnd();
extern double rnd(float range);


RESIDUE::RESIDUE()
{
	reset();
}

void RESIDUE::reset()
{
	collisions = 0;
	offset[0]=offset[1]=offset[2] = 0;
	moss = 0;
	dirt = 0;
	rust = 0;
}

void RESIDUE::fix()
{
	if (moss>1) moss = 1;	if (moss<0) moss = 0;
	if (rust>1) rust = 1;	if (rust<0) rust = 0;
	if (dirt>1) dirt = 1;	if (dirt<0) dirt = 0;
}

void RESIDUE::interpolate(RESIDUE *r1, RESIDUE *r2, double f)
{
//	err_log << "Interpolate " << this << endl;
	if (f<0) f=0;
	if (f>1) f=1;
	double F = 1-f;

//	collisions =	r1->collisions + r2->collisions;
	collisions =	r1->collisions * f	+ r2->collisions * F;
	dirt =			r1->dirt * f		+ r2->dirt * F;
	moss =			r1->moss * f		+ r2->moss * F;
	rust =			r1->rust * f		+ r2->rust * F;
}


void RESIDUE::land(POLY *poly, double *q)
{
	mark(poly,1,q);
}

//Pick up residue from poly
void RESIDUE::receive(POLY *poly, double *q)
{
	if (frame.config.get_bool(VAR_SUBDIV_ONLY))
		return;

	double to_particle, from_poly;

	to_particle = 0.3f;
	from_poly = 0.1f;

	double d = rnd() * from_poly;

	double dmoss = d * moss;
	double ddirt = d * dirt;
	double drust = d * rust;

	for (int i=0; i<3; i++)
	{
		poly->vert[i]->residue.moss -= dmoss * q[i];
		poly->vert[i]->residue.rust -= drust * q[i];
		poly->vert[i]->residue.dirt -= ddirt * q[i];
		poly->vert[i]->residue.fix();
	}

	//Setup d to work for own residue
	d /= from_poly;
	d *= to_particle;

	moss += d;
	rust += d;
	dirt += d;
	fix();
}

//Spread own residue onto poly using 
void RESIDUE::mark(POLY *poly, double *q)
{
	mark(poly,rnd(),q);
}
void RESIDUE::mark(POLY *poly, double d, double *q)
{
	if (frame.config.get_bool(VAR_SUBDIV_ONLY))
		return;

	double to_poly, from_particle;

	to_poly = 0.1f;	//0.1f
	from_particle = 0.2f;

	//Setup d to work for polygons
	d *= to_poly;

	double dmoss = d * moss;
	double ddirt = d * dirt;
	double drust = d * rust;

	for (int i=0; i<3; i++)
	{
		poly->vert[i]->residue.moss += dmoss * q[i];
		poly->vert[i]->residue.rust += drust * q[i];
		poly->vert[i]->residue.dirt += ddirt * q[i];
		poly->vert[i]->residue.fix();
	}

	//Setup d to work for own residue
	d /= to_poly;
	d *= from_particle;

	moss -= d;
	rust -= d;
	dirt -= d;

	fix();

}










