//trail.cpp

#include "main.h"

TRAIL::TRAIL()
{
	zero_vars();
}

void TRAIL::zero_vars()
{
	active = false;
	pos[0]=pos[1]=pos[2] = 0;
}

void TRAIL_LIST::create(int new_max_length, PHYS_OBJ *new_source)
{
	src = new_source;
	LIST<TRAIL>::create(new_max_length);
	zero_vars();
}

void TRAIL_LIST::zero_vars()
{
	skip_detail = 0;
	current_skip = 0;
	for (int n=0; n<max_length; n++)
	{
		list[n].active = false;
	}
}

void TRAIL_LIST::add(float nx, float ny, float nz)
{
	if (!frame.config.get_bool(VAR_DTRAILS)) return;

//	err_log << "ADD TRAIL POS: " << nx << ", " << ny << ", " << nz << endl;
	if (current_skip>0)
	{
		current_skip--;
		return;
	}
	current_skip = skip_detail;
	

	TRAIL *t = &list[next_index];
	next_index++;
	while (next_index>=max_length) next_index -= max_length;

	if (t)
	{
		t->pos[0] = nx;
		t->pos[1] = ny;
		t->pos[2] = nz;
		t->active = true;
	}
}


void TRAIL_LIST::draw()
{
	if (!frame.config.get_bool(VAR_DTRAILS)) return;

	int i, I;
	//draw from next_index backwards


	glEnable(GL_BLEND);

	glLineWidth(2);

	glBegin(GL_LINE_STRIP);
	for (i=0; i<max_length; i++)
	{
		I = next_index + i;	while (I>=max_length) I -= max_length;
		if (list[I].is_active())
		{
			glColor4f(color[0],color[1],color[2],(float)i/(float)max_length);
			glVertex3fv(list[I].pos);
		}
	}
	glColor4f(color[0],color[1],color[2],1);
	glVertex3f(src->get_pos(0),src->get_pos(1),src->get_pos(2));
	glEnd();

	glLineWidth(1);

	glPointSize(4);
	glBegin(GL_POINTS);
	for (i=0; i<max_length; i++)
	{
		I = next_index + i;	while (I>=max_length) I -= max_length;
		if (list[I].is_active())
		{
			glColor4f(color[0],color[1],color[2],(float)i/(float)max_length);
			glVertex3fv(list[I].pos);
		}
	}
	glColor4f(color[0],color[1],color[2],1);
	glVertex3f(src->get_pos(0),src->get_pos(1),src->get_pos(2));
	glEnd();
	glPointSize(1);

}