//vector_field.cpp

#include "main.h"

extern FRAME frame;
extern LOGGER err_log;
extern double rnd();
extern double rnd(float range);

VECTOR_FIELD::VECTOR_FIELD()
{
	built = false;
	list = NULL;
	total_subdivisions = 0;
}
/*
VECTOR_FIELD::VECTOR_FIELD(int nx, int ny, int nz)
{
	built = false;
	build(nx,ny,nz);
}
*/
VECTOR_FIELD::~VECTOR_FIELD()
{
	destroy();
}

void VECTOR_FIELD::build(int nx, int ny, int nz)
{
	int i, j, k;
	dimx=nx; dimy=ny; dimz=nz;

//	srand(14);	//13

	//Allocate a three-dimensional array
	list = (FIELD_ELEMENT***)calloc(dimx,sizeof(FIELD_ELEMENT**));
	for (i=0; i<dimx; i++)
	{
		list[i] = (FIELD_ELEMENT**)calloc(dimy,sizeof(FIELD_ELEMENT*));
		for (j=0; j<dimy; j++)
		{
			list[i][j] = (FIELD_ELEMENT*)calloc(dimz,sizeof(FIELD_ELEMENT));

			for (k=0; k<dimz; k++)
			{

				//Some stuff in here would initialize to some default.  I have decided to just leave it at zero.

				//list[i][j][k].vx = list[i][j][k].vy = list[i][j][k].vz = 0;


				//Various ways of setting defaults to interesting values


				//Sin/cos waves
//				list[i][j][k].vx = sin(4*i/(float)dimx);list[i][j][k].vy = cos(3*j/(float)dimy);list[i][j][k].vz = tan(k/(float)dimz);


				//Random
//				list[i][j][k].vx = rnd(1);list[i][j][k].vy = rnd(1);list[i][j][k].vz = rnd(1);


				//Circular vector field
/*
				float px,py;
				px = (float)i/(float)dimx - .5;
				py = (float)j/(float)dimy - .5;

				float r = sqrt(px*px + py*py);
//				err_log << "R: " << r << endl;

				if (r==0)
				{
					list[i][j][k].vx = 0;
					list[i][j][k].vy = 0;
				}
				else
				{
					float tx,ty;
					float theta1, theta2;

					if (px!=0)	theta1 = atan(py/px);
					else
					{
						if (py>=0)	theta1 = deg90;
						else		theta1 = deg270;
					}

					if (px<0) theta1 = deg180 + theta1;
					theta2 = atan(1.0f/r);
					theta1 += theta2;	//+ => counterclockwise
					tx = cos(theta1)*r;
					ty = sin(theta1)*r;

					list[i][j][k].vx = (tx - px);
					list[i][j][k].vy = (ty - py);

//					list[i][j][k].vx *= 5;
//					list[i][j][k].vy *= 5;
				}
*/

			}
		}
	}
	oready = false;
	wready = false;
	built = true;
}

void VECTOR_FIELD::destroy()
{
	int i, j;

	//Deallocate three-dimensional array.  I cannot
	// believe it.
	for (i=0; i<dimx; i++)
	{
		for (j=0; j<dimy; j++)
			free(list[i][j]);
		free(list[i]);
	}
	free(list);
	list = NULL;
	built = false;
	oready = false;
	wready = false;
}

//Set origin of field -- everything is offset from here
void VECTOR_FIELD::set_origin(double nx, double ny, double nz)
{
	if (!built) frame.warning("Attempting to set origin of field that's unbuilt.");
	ox = nx; oy = ny; oz = nz;
	oready = true;
	propagate();
}
//Set width of each element - determines density and overall size
void VECTOR_FIELD::set_width(double nx, double ny, double nz)
{
	if (!built) frame.warning("Attempting to set width of field that's unbuilt.");
	wx = nx; wy = ny; wz = nz;
	wready = true;
	propagate();
}

//If both set_origin and set_width have been called, propogate will conclude
// the building of the field by propogating the appropriate positions for each
// element.
//Alternatively, the elements' positions could be calculated on the fly later,
// but this reduces the need to do so.
void VECTOR_FIELD::propagate()
{
	if (!oready || !wready)
		return;

	FIELD_ELEMENT *e;
	int i,j,k;
	double X,Y,Z;

	X = ox;
	for (i=0;i<dimx;i++)
	{
		Y = oy;
		for (j=0;j<dimy;j++)
		{
			Z = oz;
			for (k=0;k<dimz;k++)
			{
				e = &list[i][j][k];
				e->x = X;
				e->y = Y;
				e->z = Z;

				Z += wz;
			}
			Y += wy;
		}
		X += wx;
	}
}


void VECTOR_FIELD::set_color(int mode)
{
	FIELD_ELEMENT *e;
	for (int i=0;i<dimx;i++)
	{
		for (int j=0;j<dimy;j++)
		{
			for (int k=0;k<dimz;k++)
			{
				e = &list[i][j][k];
				

				switch (mode)
				{
				case 0:
				default:
					{
						e->cr = e->cg = e->cb = 1;
						break;
					}
				case 1:
					{
						e->cr = e->cg = e->cb = 0;
						if (e->vx>e->vy && e->vx>e->vz) e->cr = 1;
						if (e->vy>e->vx && e->vy>e->vz) e->cg = 1;
						if (e->vz>e->vx && e->vz>e->vy) e->cb = 1;
						break;
					}
				case 2:
					{
						float v;
						e->cr = e->cg = e->cb = 0;
						v = pow(e->vx,2) + pow(e->vy,2) + pow(e->vz,2);
						v /= .01;
						if (v<.3) e->cr = 1;
						else if (v<.6) e->cg = 1;
						else if (v<.85) e->cb = 1;
						else
						{
							e->cr = 1;
							e->cg = 1;
							e->cb = 1;
						}

					}
				case 3:
					{
						break;
					}
				}
			}
		}
	}
}


//Draw the field, either in all dimensions, or on just one plane at a time
void VECTOR_FIELD::draw()
{
	if (!frame.config.get_bool(VAR_DRAW_VECTORS_MODE))
		return;

	int dvm = frame.config.get_float(VAR_DRAW_VECTORS_MODE);

	glLineWidth(5);

	FIELD_ELEMENT *e;
	float c[3];
	int i,j,k;

	// dvm determines how to draw the field
	if (dvm == DVM_ALL)
	{
		//glEnable(GL_FOG);
		//glFogi(GL_FOG_MODE,GL_LINEAR);
		//glFogf(GL_FOG_START,.25f);
		//glFogf(GL_FOG_END,.80f);

		glBegin(GL_LINES);

		float s = 2.0f;

		for (i=0;i<dimx;i++)
		{
			for (j=0;j<dimy;j++)
			{
				for (k=0;k<dimz;k++)
				{
					e = &list[i][j][k];

//					glColor4f(1,0,0,1);
					glColor4f(e->cr,e->cg,e->cb,1);
					glVertex3f(e->x, e->y, e->z);
//					glVertex3f(e->x, e->y, k*.0125);
					//cout << "x & y: " << e->vx << " " << e->vy << endl;

//					glColor4f(1,1,1,.5f);
					glColor4f(e->cr,e->cg,e->cb,0);
					glVertex3f(e->x+e->vx*s, e->y+e->vy*s, e->z+e->vz*s);
//					glVertex3f(e->x+e->vx,e->y+e->vy,k*.0125);

				}
			}
		}

		glEnd();
		//glDisable(GL_FOG);
	}
	else
	{
		glBegin(GL_LINES);

		int ei, ej;
		if (dvm == DVM_X) {ei = dimy; ej = dimz;	k = dimx/2;}
		if (dvm == DVM_Y) {ei = dimx; ej = dimz;	k = dimy/2;}
		if (dvm == DVM_Z) {ei = dimx; ej = dimy;	k = dimz/2;}

		for (i=0;i<ei;i++)
		{
			for (j=0;j<ej;j++)
			{
					if (dvm == DVM_X)	e = &list[k][i][j];
					if (dvm == DVM_Y)	e = &list[i][k][j];
					if (dvm == DVM_Z)	e = &list[i][j][k];

					c[0] = c[1] = c[2] = 0;
					if (e->vx>e->vy && e->vx>e->vz) c[0] = 1;
					if (e->vy>e->vx && e->vy>e->vz) c[1] = 1;
					if (e->vz>e->vx && e->vz>e->vy) c[2] = 1;

					glColor4f(c[0],c[1],c[2],1);
					glVertex3f(e->x,e->y,e->z);

					glColor4f(c[0],c[1],c[2],0);
					glVertex3f(e->x+e->vx*VECTOR_SCALE,e->y+e->vy*VECTOR_SCALE,e->z+e->vz*VECTOR_SCALE);
			}
		}

		glEnd();

	}

	glLineWidth(1);
}


// Returns a vector for any given point, interpolated from nearby vectors in field.
void VECTOR_FIELD::get_vector(double x, double y, double z, double *rx, double *ry, double *rz)
{
	int i,j,k;

//	err_log << "GET VECTOR FOR: " << x << ", " << y << ", " << z << " => ";

	*rx = 0;
	*ry = 0;
	*rz = 0;

	x -= ox;
	y -= oy;
	z -= oz;

	i = (int)(x / wx);
	j = (int)(y / wy);
	k = (int)(z / wz);

	if (i<0 || j<0 || k<0 || i>=dimx || j>=dimy || k>=dimz)
	{
		return;
	}
	else
	{
		if (i+1>=dimx || j+1>=dimy || k+1>=dimz)
		{
			//For edge cases, no interpolation -- just less messy that way!
			*rx = list[i][j][k].vx;
			*ry = list[i][j][k].vy;
			*rz = list[i][j][k].vz;
			return;
		}
		//Please excuse the use of 'px' vs. 'Px', I know it is ugly

		double px, py, pz, Px, Py, Pz;
		px = 1 - ((x-(i*wx)) / wx);
		py = 1 - ((y-(j*wy)) / wy);
		pz = 1 - ((z-(k*wz)) / wz);
		Px = 1 - px;
		Py = 1 - py;
		Pz = 1 - pz;

//		err_log << "P: " << px << ", " << py << ", " << pz << endl;

		*rx += list[i][j][k].vx * px * py * pz;
		*rx += list[i+1][j][k].vx * Px * py * pz;
		*rx += list[i][j+1][k].vx * px * Py * pz;
		*rx += list[i][j][k+1].vx * px * py * Pz;
		*rx += list[i+1][j+1][k].vx * Px * Py * pz;
		*rx += list[i][j+1][k+1].vx * px * Py * Pz;
		*rx += list[i+1][j][k+1].vx * Px * py * Pz;
		*rx += list[i+1][j+1][k+1].vx * Px * Py * Pz;

		*ry += list[i][j][k].vy * px * py * pz;
		*ry += list[i+1][j][k].vy * Px * py * pz;
		*ry += list[i][j+1][k].vy * px * Py * pz;
		*ry += list[i][j][k+1].vy * px * py * Pz;
		*ry += list[i+1][j+1][k].vy * Px * Py * pz;
		*ry += list[i][j+1][k+1].vy * px * Py * Pz;
		*ry += list[i+1][j][k+1].vy * Px * py * Pz;
		*ry += list[i+1][j+1][k+1].vy * Px * Py * Pz;

		*rz += list[i][j][k].vz * px * py * pz;
		*rz += list[i+1][j][k].vz * Px * py * pz;
		*rz += list[i][j+1][k].vz * px * Py * pz;
		*rz += list[i][j][k+1].vz * px * py * Pz;
		*rz += list[i+1][j+1][k].vz * Px * Py * pz;
		*rz += list[i][j+1][k+1].vz * px * Py * Pz;
		*rz += list[i+1][j][k+1].vz * Px * py * Pz;
		*rz += list[i+1][j+1][k+1].vz * Px * Py * Pz;

//		err_log << "(" << *rx << ", " << *ry << ", " << *rz << ")" << endl;

//		printf("(%i,%i,%i) => (%f,%f,%f) => (%f,%f,%f)\n",i,j,k,list[i][j][k].vx, list[i][j][k].vy, list[i][j][k].vz,rx,ry,rz);

//		*rx = list[i][j][k].vx;		*ry = list[i][j][k].vy;		*rz = list[i][j][k].vz;
	}

}


/*
Fill the vector field with values.  If anything exists
in the current vector field, it is wiped and replaced with the
new data.

- input: the vector containing data for each point.  Translation
between three-dimensional array and one-dimensional array can be
done with I = i*3 + j*dim[0]*3 + k*dim[0]*dim[1]*3, where i,j,k
are indices in 3D array, and dim[0-2] are the dimensions of this
array.  Must have dim[0]*dim[1]*dim[2]*3 elements -- the '*3' is
because there is a vx,vy,vz for each element.
- o: Origin of field.  This allows you to put the field anywhere
within the simulation.  Must have three elements.
- w: Width of each entry - this determines the spacing between
vectors, which affects the density, and thus overall width of
total field.  Must have three elements.
- dim: Dimensions of field - how many vectors are in each dimension.
Must have three elements.
*/
void VECTOR_FIELD::read_field(double *input, double *o, double *w, int *dim)
{
	int I, i,j,k;
	if (built) destroy();
	build(dim[0],dim[1],dim[2]);
	for (i=0; i<dim[0]; i++)
	{
		for (j=0; j<dim[1]; j++)
		{
			for (k=0; k<dim[2]; k++)
			{
				I = 3*(i + dim[0]*(j + k*dim[1]));

				list[i][j][k].vx = input[I];
				list[i][j][k].vy = input[I+1];
				list[i][j][k].vz = input[I+2];
			}
		}
	}
	set_origin(o[0],o[1],o[2]);
	set_width(w[0],w[1],w[2]);
}

void VECTOR_FIELD::read_field(FIELD_ELEMENT ***f, int *dim)
{
	if (built) destroy();

	dimx = dim[0];
	dimy = dim[1];
	dimz = dim[2];

	list = f;

	built = true;
	oready = true;
	wready = true;
	/*
	for (int i=0; i<dim[0]; i++)
	{
		for (int j=0; j<dim[1]; j++)
		{
			for (int k=0; k<dim[2]; k++)
			{
				cout << "i j k: " << i << " " << j << " " << k << endl;
				cout << "vx vy vz: " << list[i][j][k].vx << " " << list[i][j][k].vy << " " << list[i][j][k].vz << endl;
			}
		}
	}
	*/
}

/*
void VECTOR_FIELD::build_from_file(void)
{
	ifstream file_x;
	ifstream file_y;
	ifstream file_z;
	ifstream file_u;
	ifstream file_v;

	file_x.open("data/ocean40x25_x.bin", ios::in|ios::binary);
	file_y.open("data/ocean40x25_y.bin", ios::in|ios::binary);
	file_z.open("data/ocean40x25x11_z.bin", ios::in|ios::binary);
	file_u.open("data/ocean40x25x11_u.bin", ios::in|ios::binary);
	file_v.open("data/ocean40x25x11_v.bin", ios::in|ios::binary);

	char c[10];
	char d[10];
	char e[10];
	char f[10];
	char g[10];
	
	FIELD_ELEMENT*** vector_f;
	
	vector_f = (FIELD_ELEMENT***)calloc(40,sizeof(FIELD_ELEMENT**));
	for(int i=0;i<40;i++)
	{
		vector_f[i] = (FIELD_ELEMENT**)calloc(25,sizeof(FIELD_ELEMENT*));
		
		for(int j=0;j<25;j++)
		{
			file_x.read(c, 4);
			file_y.read(d, 4);
			
			vector_f[i][j] = (FIELD_ELEMENT*)calloc(11,sizeof(FIELD_ELEMENT));

			for(int k=0;k<11;k++)
			{
				file_z.read(e, 4);
				file_u.read(f, 4);
				file_v.read(g, 4);

				vector_f[i][j][k].x = *(float *)c;
				vector_f[i][j][k].y = *(float *)d;
				vector_f[i][j][k].z = *(float *)e;

				vector_f[i][j][k].vx = *(float *)f;
				vector_f[i][j][k].vy = *(float *)g;
				vector_f[i][j][k].vz = 0;
				
			}
		}
	}

	int dim[3] = {40,25,11};

	read_field(vector_f, dim);

	file_x.close();
	file_y.close();
	file_z.close();
	file_u.close();
	file_v.close();

}
*/

/*
void VECTOR_FIELD::build_from_matlab(void)
{
	err_log << "Building from matlab." << endl;

	ifstream file_u;
	ifstream file_v;

	file_u.open("data/x_val2.txt");
	file_v.open("data/y_val2.txt");

	char c[32];
	char d[32];

	file_u >> c >> d;
	file_v >> c >> d;

	int dim1 = atoi(c);
	int dim2 = atoi(d);

	err_log << "DIMS: " << dim1 << ", " << dim2 << endl;
	
	list = (FIELD_ELEMENT***)calloc(dim1,sizeof(FIELD_ELEMENT**));
	for(int i=0;i<dim1;i++)
	{
		list[i] = (FIELD_ELEMENT**)calloc(dim2,sizeof(FIELD_ELEMENT*));
		
		for(int j=0;j<dim2;j++)
		{
			file_u >> c;
			file_v >> d;

			list[i][j] = (FIELD_ELEMENT*)calloc(10,sizeof(FIELD_ELEMENT));

			for(int k=0;k<10;k++)
			{
				list[i][j][k].x = i;
				list[i][j][k].y = j;
				list[i][j][k].z = k;

				list[i][j][k].vx = atof(c);
				list[i][j][k].vy = atof(d);
				list[i][j][k].vz = 0;

				err_log << "vx, vy: " << atof(c) << " " << atof(d) << endl;
				
			}
		}
	}

	int dim[3];
	dim[0] = dim1;
	dim[1] = dim2;
	dim[2] = 10;

	err_log << "Vector field n shit totally written yo." << endl;


	dimx = dim[0];
	dimy = dim[1];
	dimz = dim[2];
	built = true;
	oready = true;
	wready = true;

	file_u.close();
	file_v.close();
}
*/


//For a one-dimensional vector, representing a 3D
// vector that is just flattened, and consisting of triplets, 
// modify a given triplet with x,y,z
//Assume first three elements contain X,Y,Z dimensions
void grid_input(float *v, int i, float x, float y, float z)
{
	i *= 3;
	i += 3;	//offset since first three elements are dimensions
	v[i] = x;
	v[i+1] = y;
	v[i+2] = z;
}


///////////CRAZY WHAT subdivide too much gets weird vector field

//Take a field (built from build_from_brandon) and subdivide it
// some number of times
void VECTOR_FIELD::subdivide_field(int subdiv_iterations)
{
	err_log << "Subdividing vector field..." << endl;
	if (!built)
	{
		frame.error("Attempting to subdivide a field that hasn't been build.");
		return;
	}

	int i,j,k,q;

	//We are assuming cube!
	int d = dimx;

	int fine_size = mem_size(subdiv_iterations,d)+3;
	int coarse_size = mem_size(0,d)+3;
	float *fine = (float*)calloc(fine_size,sizeof(float));
	float *coarse = (float*)calloc(coarse_size,sizeof(float));

    *(coarse) = dimx;
    *(coarse+1) = dimy;
    *(coarse+2) = dimz;

	err_log << "Filling course grid from current list..." << endl;
	q = 3;
	for (i=0; i<d; i++)
	{
		for (j=0; j<d; j++)
		{
			for (k=0; k<d; k++)
			{
				coarse[q] = list[i][j][k].vx;
				coarse[q+1] = list[i][j][k].vy;
				coarse[q+2] = list[i][j][k].vz;
				q += 3;
			}
		}
	}


	err_log << "Subdividing..." << endl;
	subdivide(subdiv_iterations, 15, coarse, fine);
	total_subdivisions += subdiv_iterations;

	int dim1 = get_subdivided_dims(subdiv_iterations, d);
	int dim2 = dim1;		//uniform dimensions
	int dim3 = dim1;

	err_log << "New dimensions: " << dim1 << ", " << dim2 << ", " << dim3 << endl;

	float width_scale = width_scale_source / pow(2.0f,total_subdivisions);
	wx = wy = wz = (width_scale_source * 2.0f) / dim1;

	destroy();

	err_log << "Filling out vector field..." << endl;
	q = 3;
	list = (FIELD_ELEMENT***)calloc(dim1,sizeof(FIELD_ELEMENT**));
	for(i=0;i<dim1;i++)
	{
		list[i] = (FIELD_ELEMENT**)calloc(dim2,sizeof(FIELD_ELEMENT*));
		for(j=0;j<dim2;j++)
		{
			list[i][j] = (FIELD_ELEMENT*)calloc(dim3,sizeof(FIELD_ELEMENT));
			for(k=0;k<dim3;k++)
			{
				list[i][j][k].x = i * width_scale + ox;
				list[i][j][k].y = j * width_scale + oy;
				list[i][j][k].z = k * width_scale + oz;

				list[i][j][k].vx = fine[q];
				list[i][j][k].vy = fine[q+1];
				list[i][j][k].vz = fine[q+2];

				q += 3;				
			}
		}
	}


	int dim[3];
	dim[0] = dim1;
	dim[1] = dim2;
	dim[2] = dim3;

	dimx = dim[0];
	dimy = dim[1];
	dimz = dim[2];

	built = true;
	oready = true;
	wready = true;

    free(fine);
	free(coarse);

	err_log << "Vector field subdivision complete." << endl;

	set_color(2);

	//Update visual representation
	frame.ibfv.set_faces();
}

//Named such because it stems from brandon's work with subdivision.
// Takes as input a vector field file (.vf)
/*
VF file: 6 parameters, followed by any number of data points

Parameters:
int		Dimension (assumes cubic, so only one value needed)
float	Origin X
float	Origin Y
float	Origin Z
float	Field width (entire field)
int		Initial subdivisions to perform

The subsequent data points are just a linear amalgamation of a three-dimensional
field of three-element vectors.

The X dimension is major, followed by Y, then Z.

*/
bool VECTOR_FIELD::build_from_brandon(char *filename)
{
	int d, i, q, target;
	ifstream file;
	char c[128];
	int subdiv_iterations;
	int fine_size;
	int coarse_size;
	float *fine;
	float *coarse;
	bool set_faces = false;

	err_log << "Building vector field..." << endl;

	file.open(filename, ios::in|ios::binary);
	if (!file)
	{frame.error("Cannot open vector field file: %s",filename); return false;}

	if (list!=NULL)
	{
		err_log << "Destroying current field..." << endl;
		destroy();
		set_faces = true;
	}


	//Read in dimensions (first parameter -- assuming cube, so only needs one!)
	file >> c;
	d = atoi(c);
	err_log << "Vector field size: " << d << endl;
	if (d<0 || d%2==0) {frame.error("Invalid vector field dimension: %i",d); return false;}

	//Read in origin of field (three parameters, float)
	file >> c;	ox = atof(c);
	if (*c == NULL) {frame.error("Invalid vector field origin parameter."); return false;}
	file >> c;	oy = atof(c);
	if (*c == NULL) {frame.error("Invalid vector field origin parameter."); return false;}
	file >> c;	oz = atof(c);
	if (*c == NULL) {frame.error("Invalid vector field origin parameter."); return false;}

	//Read in vector field width (cube, so just one parameter, float)
	file >> c;	width_scale_source = atof(c);
	if (*c == NULL) {frame.error("Invalid vector field width parameter."); return false;}

	//Read last parameter (before data), the number of times to subdivide on entry
	file >> c;	subdiv_iterations = atoi(c);
	if (*c == NULL) {frame.error("Invalid vector field subdivision count."); return false;}

	//The 3 in here is for the three XYZ dimensions
	fine_size = mem_size(subdiv_iterations,d)+3;
	coarse_size = mem_size(0,d)+3;
	fine = (float*)calloc(fine_size,sizeof(float));
	coarse = (float*)calloc(coarse_size,sizeof(float));
	assert(fine!=NULL);
	assert(coarse!=NULL);

	coarse[0] = d;
	coarse[1] = d;
	coarse[2] = d;

	target = mem_size(0,d);

	err_log << "Reading in field from file..." << endl;
	//Coarse file starts at 3
	i = 3;
	while (*c != NULL)
	{
		file >> c;
		coarse[i] = atof(c);
		i++;
	}
	//If not enough elements, fill rest with zero, warn
	if (i!=target+1) frame.warning("Vector field has invalid number of values: %i should be %i",i,target);
	while (i<target)
	{
		coarse[i] = 0;
		i++;
	}



	//Original hard-coded vector field reading
/*
	//Set origin
	ox = oy = oz = -.5f;

	//Set the width of the entire vector field
	width_scale_source = .5f;


	err_log << "Filling course grid..." << endl;
	//Zero it all
    for (int i = 0; i < mem_size(0,d); i+=3)
    {
		coarse[3 + i] = 0.0;
		coarse[4 + i] = 0.0;
		coarse[5 + i] = 0.0;
    }

	//At this point, we could read in the vector field from a file.
	// Just fill coarse[] with the data directly, or you could use
	// grid_input
	//Also, you would need to adjust d if the vector is not 3x3.
	// In that case, this would have to happen earlier, before the allocations.


//vortex:
	grid_input(coarse,3,0,1,0);
	grid_input(coarse,15,1,0,0);
	grid_input(coarse,21,0,-1,0);
	grid_input(coarse,9,-1,0,0);

	grid_input(coarse,4,0,1,0);
	grid_input(coarse,16,1,0,0);
	grid_input(coarse,22,0,-1,0);
	grid_input(coarse,10,-1,0,0);

	grid_input(coarse,5,0,1,0);
	grid_input(coarse,17,1,0,0);
	grid_input(coarse,23,0,-1,0);
	grid_input(coarse,11,-1,0,0);

	grid_input(coarse,14,0,0,-1);
*/




	err_log << "Subdividing..." << endl;
	subdivide(subdiv_iterations, 15, coarse, fine);

	total_subdivisions = subdiv_iterations;
    

	//Figure new dimensions, from old and number of iterations to perform
	int dim1 = get_subdivided_dims(subdiv_iterations, d);
	int dim2 = dim1;		//uniform dimensions
	int dim3 = dim1;
	err_log << "New dimensions: " << dim1 << ", " << dim2 << ", " << dim3 << endl;


	//Figure width of each element
	wx = wy = wz = (width_scale_source * 2.0f) / dim1;
	float width_scale = width_scale_source / pow(2.0f,subdiv_iterations);

	err_log << "Filling out vector field..." << endl;
	q = 3;
	list = (FIELD_ELEMENT***)calloc(dim1,sizeof(FIELD_ELEMENT**));
	for(i=0;i<dim1;i++)
	{
		list[i] = (FIELD_ELEMENT**)calloc(dim2,sizeof(FIELD_ELEMENT*));
		for(int j=0;j<dim2;j++)
		{
			list[i][j] = (FIELD_ELEMENT*)calloc(dim3,sizeof(FIELD_ELEMENT));
			for(int k=0;k<dim3;k++)
			{
				list[i][j][k].x = i * width_scale + ox;
				list[i][j][k].y = j * width_scale + oy;
				list[i][j][k].z = k * width_scale + oz;

				list[i][j][k].vx = fine[q];
				list[i][j][k].vy = fine[q+1];
				list[i][j][k].vz = fine[q+2];

				q += 3;				
			}
		}
	}


	int dim[3];
	dim[0] = dim1;
	dim[1] = dim2;
	dim[2] = dim3;

	dimx = dim[0];
	dimy = dim[1];
	dimz = dim[2];

	built = true;
	oready = true;
	wready = true;

    free(fine);
	free(coarse);

	err_log << "Vector field complete." << endl;

	set_color(2);

	if (set_faces)
		frame.ibfv.set_faces();

	return true;
}


int VECTOR_FIELD::get_subdivided_dims(int subdiv_iterations, int current)
{
	int t = pow(2.0f,subdiv_iterations);
	return (t*current - (t-1));
}

