//vector_field.h

#ifndef _VECTOR_FIELD_
#define _VECTOR_FIELD_


#define DVM_NULL		0
#define DVM_ALL			1
#define DVM_X			2
#define DVM_Y			3
#define DVM_Z			4

#define VECTOR_SCALE	.1f

//Stores the position (x,y,z) of this element, and 
// the vector (vx,vy,vz) associated with it
struct FIELD_ELEMENT
{
	double x,y,z;
	double vx,vy,vz;
	double cr, cg, cb;
};

class VECTOR_FIELD
{
public:
	VECTOR_FIELD();
	~VECTOR_FIELD();

	//Initialize and create memory for vector field of the given dimensions
	void build(int nx, int ny, int nz);

	//Destroy the memory for the given field (but not the object itself!)
	void destroy();

	//Read in a field with the given parameters:
	void read_field(double *input, double *o, double *w, int *dim);

	void read_field(FIELD_ELEMENT ***f, int *dim);

//	void build_from_file(void);
//	void build_from_matlab(void);
	bool build_from_brandon(char *filename);

	int get_subdivided_dims(int subdiv_iterations, int current);

	//Get a vector, interpolated from the field
	void get_vector(double x, double y, double z, double *rx, double *ry, double *rz);

	void get_origin(double *x, double *y, double *z) {*x=ox; *y=oy; *z=oz;}
	void get_dims(int *x, int *y, int *z) {*x=dimx; *y=dimy; *z=dimz;}
	void get_widths(double *x, double *y, double *z) {*x=wx; *y=wy; *z=wz;}

	//Draw the field
	void draw();

	//Set origin/width
	void set_origin(double nx, double ny, double nz);
	void set_width(double nx, double ny, double nz);
	void set_color(int mode);

	void subdivide_field(int subdiv_iterations);


protected:
	int dimx, dimy, dimz;
	double wx, wy, wz;
	double ox, oy, oz;

	int total_subdivisions;

	FIELD_ELEMENT ***list;

	void propagate();
	bool oready;
	bool wready;
	bool built;

	float width_scale_source;
};

















#endif




