//viewport.cpp

#include "main.h"



VIEWPORT::VIEWPORT()
{
	which_corner = -1;

	poly = 1;
	wire = 1;
	normal = 0;
	points = 0;
	hidden = 0;
	lighting = 1;

	ox = 0;
	oy = 0;
	dx = 100;
	dy = 100;
	projection = PROJECTION_PERSPECTIVE;

	camera.set_viewport(this);
}


void VIEWPORT::draw()
{
	if (frame.config.get_bool(VAR_SINGLE_VIEW))
		glViewport(0,0,frame.display.screen_x,frame.display.screen_y);
	else
		glViewport(ox,oy,dx,dy);
}

void VIEWPORT::cycle_poly()			{	poly++;		do_force_new_list();	if (poly>1) poly = 0; }
void VIEWPORT::cycle_wire()			{	wire++;		do_force_new_list();	if (wire>1) wire = 0; }
void VIEWPORT::cycle_normal()		{	normal++;	do_force_new_list();	if (normal>1) normal = 0; }
void VIEWPORT::cycle_points()		{	points++;	do_force_new_list();	if (points>1) points = 0; }
void VIEWPORT::cycle_lighting()		{	lighting++;	do_force_new_list();	if (lighting>1) lighting = 0; }
void VIEWPORT::cycle_spawn_box()	{	spawn_box++;						if (spawn_box>1) spawn_box = 0; }
void VIEWPORT::cycle_visualization(){	visualization++;do_force_new_list();if (visualization>5) visualization = 0; }

void VIEWPORT::cycle_hidden()
{
	hidden++;
	if (hidden && !poly) cycle_poly();
	do_force_new_list();
	if (hidden>1) hidden = 0;
}
void VIEWPORT::cycle_grid()
{
	grid++;
	if (projection==PROJECTION_PERSPECTIVE)
	{
		grid3D.cycle_which();
	}
	else if (grid>1) grid = 0;

	err_log << "GRID SET TO: " << grid << endl;
}

void VIEWPORT::set_ul()
{
	which_corner = 0;
	ox = 0;
	oy = frame.display.screen_half_y;
	dx = frame.display.screen_half_x;
	dy = frame.display.screen_half_y;
	poly = 1;
	wire = 1;
	normal = 0;
	points = 0;
	hidden = 0;
	grid = 0;
	spawn_box = 0;
	visualization = 0;
	projection = PROJECTION_PERSPECTIVE;
}
void VIEWPORT::set_ur()
{
	which_corner = 1;
	ox = frame.display.screen_half_x;
	oy = frame.display.screen_half_y;
	dx = frame.display.screen_half_x;
	dy = frame.display.screen_half_y;
	poly = 0;
	wire = 1;
	normal = 0;
	points = 0;
	hidden = 0;
	grid = 1;
	spawn_box = 1;
	visualization = 0;
	projection = PROJECTION_ORTHO;
	camera.set_XY();
}
void VIEWPORT::set_ll()
{
	which_corner = 2;
	ox = 0;
	oy = 0;
	dx = frame.display.screen_half_x;
	dy = frame.display.screen_half_y;
	poly = 0;
	wire = 1;
	normal = 0;
	points = 0;
	hidden = 0;
	grid = 1;
	spawn_box = 1;
	visualization = 0;
	projection = PROJECTION_ORTHO;
	camera.set_XZ();
}
void VIEWPORT::set_lr()
{
	which_corner = 3;
	ox = frame.display.screen_half_x;
	oy = 0;
	dx = frame.display.screen_half_x;
	dy = frame.display.screen_half_y;
	poly = 0;
	wire = 1;
	normal = 0;
	points = 0;
	hidden = 0;
	grid = 1;
	spawn_box = 1;
	visualization = 0;
	projection = PROJECTION_ORTHO;
	camera.set_YZ();
}


void VIEWPORT::set_camera()
{
//	err_log << "CAMERA set to: " << which_corner << endl;
	frame.camera = &camera;

	camera.draw_grid();
}

void VIEWPORT::do_force_new_list()
{
	force_new_list = true;
}
bool VIEWPORT::will_force_new_list()
{
	if (force_new_list)
	{
		//Also clear for this viewport - side-effect
		force_new_list = false;
		return true;
	}
	return false;
}

void VIEWPORT::set_matrices()
{
	if (projection == PROJECTION_ORTHO)			camera.set_matrices_ortho();
	if (projection == PROJECTION_PERSPECTIVE)	camera.set_matrices_perspective();
}

VIEWPORTS::VIEWPORTS()
{
	divider_padding = 0;
}

void VIEWPORTS::init()
{
	vp[0].set_ul();
	vp[1].set_ur();
	vp[2].set_ll();
	vp[3].set_lr();
	vp[0].set_camera();
	divider_x = frame.display.screen_half_x;
	divider_y = frame.display.screen_half_y;
}


int VIEWPORTS::get_viewport(int x, int y)
{	
	if (frame.config.get_bool(VAR_SINGLE_VIEW))
		return -1;

	if (x>divider_x+divider_padding && y>divider_y+divider_padding) return 3;
	if (x<divider_x-divider_padding && y>divider_y+divider_padding) return 2;
	if (x>divider_x+divider_padding && y<divider_y-divider_padding) return 1;
	if (x<divider_x-divider_padding && y<divider_y-divider_padding) return 0;

	return -1;
}




